#include "config.h"
#include <utility>
#include <cstdlib>
#include <cstdio>
#include <cassert>
#include <cstring>
#include <climits>
#include <cfloat>
#include <stdlib.h>
#include <list>
#include "vnoc_topology.h"
#include "hetero.h"
#include "hungarian.h"
#include "vnoc_app.h"
#include "fp_btree.h"
#include "fp_sa.h"
#include "mcf.h"

using namespace std;


////////////////////////////////////////////////////////////////////////////////
//
// ROUTERS_DISTRIBUTION
//
////////////////////////////////////////////////////////////////////////////////

ROUTERS_DISTRIBUTION::ROUTERS_DISTRIBUTION(long x, long y, long x0, long y0, double ll) 
{
	_nx = x; 
	_ny = y;
	_sx = x0; 
	_sy = y0;
	_link_length = ll;
}

////////////////////////////////////////////////////////////////////////////////
//
// ROUTER_ASSIGNMENT
//
////////////////////////////////////////////////////////////////////////////////

ROUTER_ASSIGNMENT::ROUTER_ASSIGNMENT( int ary) 
{
	_n_rows = ary; 
	_n_columns = ary;
	_array_ids = (int**) malloc(_n_rows * sizeof(int *));
	_array_corners = (CORNER_NAME**) malloc(_n_rows * sizeof(int *));
	_array_extralinks = (double**) malloc(_n_rows * sizeof(double *));
	for ( int i = 0; i < _n_rows; i++) {
		_array_ids[i] = (int *) malloc(_n_columns * sizeof(int));
		_array_corners[i] = (CORNER_NAME *) malloc(_n_columns * sizeof(int));
		_array_extralinks[i] = (double *) malloc(_n_columns * sizeof(double));
	}
	clear_assignments();
}

void ROUTER_ASSIGNMENT::assign(int x,int y, int id, 
	CORNER_NAME corner, double extralink) 
{
	_array_ids[x][y] = id;
	_array_corners[x][y] = corner;
	_array_extralinks[x][y] = extralink;
	_total_extralinks += extralink;
}

int ROUTER_ASSIGNMENT::assigned(int x, int y)
{
	if ( x >= _n_columns || x < 0 || y >= _n_rows || y < 0) {
		printf("ERROR: trying to verify assignment of router %d.%d->%d (_n_columns=%d,_n_rows=%d)\n",
			x, y, _array_ids[x][y], _n_columns, _n_rows);
		exit(1);
	}
	return _array_ids[x][y];
}

CORNER_NAME ROUTER_ASSIGNMENT::corner_used(int x, int y) 
{
	if ( x >= _n_columns || x < 0 || y >= _n_rows || y < 0) {
		printf("ERROR: trying to verify corner used by router %d.%d (_n_columns=%d,_n_rows=%d)\n",
			x, y, _n_columns, _n_rows);
		exit(1);
	}
	return _array_corners[x][y];
}

pair<int, int> ROUTER_ASSIGNMENT::core_id_to_router_xy(int id) 
{
	int found_x = -1;
	int found_y = -1;
	pair<int, int> temp_xy;

	for ( int j = 0; j< _n_rows; j++) {
		for( int i = 0; i < _n_columns; i++) {
			if ( _array_ids[i][j] == id) {
				found_x = i; 
				found_y = j;
			}
		}
	}
	assert( found_x != -1 && found_y != -1);
	temp_xy.first = found_x; 
	temp_xy.second = found_y;
	return temp_xy;
}

CORNER_NAME ROUTER_ASSIGNMENT::corner_used_by_id(int id) 
{
	int found_x = -1;
	int found_y = -1;
	CORNER_NAME temp_corner;

	for ( int j = 0; j < _n_rows; j++) {
		for(int i = 0; i < _n_columns; i++) {
			if ( _array_ids[i][j] == id) {
				found_x = i;
				found_y = j;
			}
		}
	}
	assert( found_x != -1 && found_y != -1);
	temp_corner = _array_corners[found_x][found_y];
	return temp_corner;
}

double ROUTER_ASSIGNMENT::extralink_by_id(int id)
{
	int found_x = -1;
	int found_y = -1;

	for ( int j = 0; j < _n_rows; j++) {
		for(int i = 0;i < _n_columns; i++) {
			if ( _array_ids[i][j] == id){
				found_x = i; 
				found_y = j;
			}
		}
	}
	assert(found_x != -1 && found_y != -1);
	return _array_extralinks[found_x][found_y];
}

void ROUTER_ASSIGNMENT::clear_assignments()
{
	// clear all assignments;
	for ( int y = 0; y < _n_rows; y++) {
		for ( int x = 0; x < _n_columns; x++) {
			_array_ids[x][y] = -1;
		}
	}
	_total_extralinks = 0.0;
}

////////////////////////////////////////////////////////////////////////////////
//
// HETEROGENEOUS_NOC
//
////////////////////////////////////////////////////////////////////////////////

bool HETEROGENEOUS_NOC::parse_command_arguments( int argc, char *argv[]) 
{
	bool result = true;
	int i = 0;

	// topology and trace files are mandatory;
	if (argc < 3 || argv[1]=="--help" || argv[1]=="-h" || argv[1]=="-?" || argv[1]=="/?") {
		printf("\nUsage: hnoc file: (test file) cycles: (int) warmup: (int) [Options...]\n\n");
		printf(" Option\t\tDescription - default internal value\n");
		printf("FLOORPLAN ANNEALING:\n");
		printf(" [times:]\tUsed to multiply num of cores to obtain num of permutations in one temperature - 160\n");
		printf(" [local:]\tNumber of local search iterations - 6\n");
		printf(" [term_temp:]\tAnnealing final/terminating temperature (starts at 1) - 0.1\n");
		printf(" [avg_ratio:]\tUsed in annealing cost calculations - 30\n");
		printf(" [lambda:]\tUsed in the exponential for probability calculation - 1.3\n");
		printf(" [alpha:]\tAnnealing cost ratio (from 0-wirelength to 1-used area) - 0.25\n");
		printf(" [scale:]\tScale multiplier - disabled by default\n");
		printf("LOOP:\n");
		printf(" [seed:]\tRandom number generator seed. Used only once for annealing and\n\t\tre-applied for every simulation - using internal clock\n");
		printf(" [n_fps:]\tNumber of floorplans to try (attempts) - 10\n");
		printf(" [n_best:]\tNumber of best floorplans to be simulated - 3\n");
		printf(" [fp_criteria:]\tCriteria for \"best\" fp's (W-wirelengh, A-area) - A\n");
		printf(" [x_ary:]\tExtra routers per dimension - 0\n");
		printf("DESIGN FLOW:\n");
		printf(" [rp:]\t\tRouting paths calculation, mcf or shortest - mcf\n");
		printf(" [verbose:]\tPrint debugging info, 0 or 1 - 1\n");
		printf("NOC SIMULATION:\n");
		printf(" [mode:]\tSimulation mode, 2D or 3D - 2D\n");
		printf(" [inp_buf:]\tRouters input buffers size (in flits) - 12\n");
		printf(" [out_buf:]\tRouters output buffers size (in flits) - 12\n");
		printf(" [vc_n:]\tNumber of virtual channels - 3\n");
		printf(" [flit_size:]\tFlit size - 4\n");
		printf(" [link_bw:]\tLink bandwidth - 64\n");
		printf(" [pipe_link:]\tNumber of pipeline stages per link - 0\n");
		printf(" [routing_a:]\tRouting algorithm. Must be LUT for hetero NOC's - LUT\n");
		printf(" [extra:]\tExtra-links timing factor. Use 0 to disable them - 1\n");
		printf(" [gui]\t\tUse a graphical interface - disabled by default\n");
		printf(" [p:]\t\tGUI pausing behavior per simulation. 0:step-by-step, 1:once at\n\t\tthe end, 2:never pauses - 1\n");
		printf(" [name:]\tOutputs the average latency of all simulations to a file\n\t\tnamed results\\results.txt with a \"simulation name\" tag -\n\t\tdisabled by default\n");
		printf(" [load:]\tAdjust the injection load by 10,20,30...100 percent (use only\n\t\tmultiples of 10) - 100\n");
		printf(" [loadsweep]\tSweeps the injection load from 10-100%% - disabled\n");
		printf(" [bufsweep]\tSweeps buffers size from 1x to 5x - disabled\n");
		printf(" [excel]\tIf a load or buffer size sweep is used, outputs an excel file - disabled\n\n");
		exit(1);
	}

	i = 1;
	while ( i < argc) {
		if ( !strcmp(argv[i], "file:")) {
			strcpy(_inputfile, argv[i+1]);
			i += 2; // skip the next parameter;
			continue;
		}
		if ( !strcmp(argv[i], "times:")) {
			_times = atoi(argv[i+1]);
			if (_times < 1 || _times > LONG_MAX) { 
				printf("Error:\ttimes value must be between [1 %ld].\n", LONG_MAX);
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "local:")) {
			_local = atoi(argv[i+1]);
			if (_local < 0 || _local > LONG_MAX) { 
				printf("Error:\tlocal value must be between [0 %ld].\n", LONG_MAX);
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "avg_ratio:")) {
			_avg_ratio = atoi(argv[i+1]);
			if (_avg_ratio < 1 || _avg_ratio > LONG_MAX) { 
				printf("Error:\tavg_ratio value must be between [1 %ld].\n", LONG_MAX);
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "lambda:")) {
			_lambda = atof(argv[i+1]);
			if (_lambda < 0 || _lambda > 10) { 
				printf("Error:\tlambda value must be between [0 10].\n");
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "term_temp:")) {
			_term_temp = atof(argv[i+1]);
			if (_term_temp < 0 || _term_temp > 1) { 
				printf("Error:\tterm_temp value must be between [0 1].\n");
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp( argv[i], "alpha:")) {
			_alpha = atof(argv[i+1]);
			if (_alpha < 0 || _alpha > 1) { 
				printf("Error:\talpha value must be between [0 1].\n");
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "scale:")) {
			_fp_scale = atof(argv[i+1]);
			if (_fp_scale < 0.0005 || _fp_scale > 5000) { 
				printf("Error:\tscale value must be between [0.0005 5000].\n");
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "seed:")) {
			_seed = atoi(argv[i+1]);
			if (_seed < 1 || _seed > LONG_MAX) { 
				printf("Error:\tseed value must be between [1 %ld].\n", LONG_MAX);
				exit(1); 
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "n_fps:")) {
			_n_fps = atoi(argv[i+1]);
			if (_n_fps < 1 || _n_fps > LONG_MAX) { 
				printf("Error:\tn_fps value must be between [1 %ld].\n", LONG_MAX);
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "n_best:")) {
			_n_best = atoi(argv[i+1]);
			if (_n_best < 1 || _n_best > LONG_MAX) { 
				printf("Error:\tn_best value must be between [1 %ld].\n", LONG_MAX);
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "fp_criteria:")) {
			_fp_criteria = argv[i+1][0];
			if (_fp_criteria != 'W' && _fp_criteria != 'A') { 
				printf("Error:\tfp_criteria must be between W or A.\n", LONG_MAX);
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "x_ary:")) {
			_x_ary = atoi(argv[i+1]);
			if (_x_ary < 0 || _x_ary > INT_MAX) { 
				printf("Error:\tx_ary value must be between [0 %d].\n", INT_MAX);
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp( argv[i], "rp:")) { // routing paths calculation mode;
			if (strcmp(argv[i+1],"mcf") && strcmp(argv[i+1],"shortest")) { 
				printf("Error:\tmode must be mcf or shortest.\n", LONG_MAX);
				exit(1); 
			}
			if ( !strcmp(argv[i+1], "mcf")){
				_routing_paths_calc = ROUTING_PATHS_MCF;
			} else {
				_routing_paths_calc = ROUTING_PATHS_SHORTEST;
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "verbose:")) {
			int v2i = atoi(argv[i+1]);
			if (v2i != 0 && v2i != 1) { 
				printf("Error:\tverbose value must be 0 or 1.\n");
				exit(1); 
			}
			if (v2i == 0) _verbose = false;
			else if (v2i == 1) _verbose = true;
			i += 2; 
			continue;
		}
		if ( !strcmp( argv[i], "mode:")) {
			if (strcmp(argv[i+1],"2D") && strcmp(argv[i+1],"3D")) { 
				printf("Error:\tmode must be 2D or 3D.\n", LONG_MAX);
				exit(1); 
			}
			if ( !strcmp(argv[i+1], "2D")){
				_sim_mode = SIMULATED_ARCH_2D;
			} else {
				_sim_mode = SIMULATED_ARCH_3D;
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "inp_buf:")) {
			_inp_buf = atoi(argv[i+1]); 
			if (_inp_buf < 1 || _inp_buf > 1024) { 
				printf("Error:\tinp_buf value must be between [1 1024].\n");
				exit(1); 
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "out_buf:")) {
			_out_buf = atoi(argv[i+1]);
			if (_out_buf < 1 || _out_buf > 1024) { 
				printf("Error:\tout_buf value must be between [1 1024].\n");
				exit(1); 
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "vc_n:")) {
			_vc_n = atoi(argv[i+1]); 
			if (_vc_n < 1 || _vc_n > 128) { 
				printf("Error:\tvc_n value must be between [1 128].\n");
				exit(1); 
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "flit_size:")) {
			_flit_size = atoi(argv[i+1]);
			if (_flit_size < 1 || _flit_size > 128) { 
				printf("Error:\tflit_size value must be between [1 128].\n");
				exit(1); 
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "link_bw:")) {
			_link_bw = atoi(argv[i+1]);
			if (_link_bw < 1 || _link_bw > 128) { 
				printf("Error:\tlink_bw value must be between [1 128].\n");
				exit(1); 
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "pipe_link:")) {
			_pipeline_in_link = atoi(argv[i+1]);
			if (_pipeline_in_link < 0 || _pipeline_in_link > 32) { 
				printf("Error:\tpipe_link value must be between [0 32].\n");
				exit(1); 
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "routing_a:")) {
			if (argc <= i+1) {
				printf ("Error:\trouting_a option requires a string parameter.\n");
				exit (1);
			} 
			if (strcmp(argv[i+1], "LUT") == 0) {
				_routing_a = ROUTING_LUT;
			} else {
				printf("Error:\trouting_a must be LUT for heterogeneous NOC's.\n");
				exit (1);
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "cycles:")) {
			_cycles = atoi(argv[i+1]);
			if (_cycles <= 0 || _cycles > 500000) { 
				printf("Error:\tcycles value must be between [1 500000].\n");
				exit(1); 
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "warmup:")) {
			_warmup_cycles = atoi(argv[i+1]);
			if (_warmup_cycles <= 0 || _warmup_cycles > 500000) { 
				printf("Error:\twarmup value must be between [1 500000].\n");
				exit(1);
			}
			i += 2; 
			continue;
		}
		if ( !strcmp(argv[i], "extra:")) {
			_extra_links_timing_factor = atof(argv[i+1]);
			if (_extra_links_timing_factor < 0 || 
				_extra_links_timing_factor > 10) { 
				printf("Error:\textra value must be between [0 10].\n");
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "gui")) {
			_use_gui = true;
			i++; 
			continue;
		}
		if ( !strcmp(argv[i], "p:")) {
			_gui_pauses = atoi(argv[i+1]);
			if (_gui_pauses < 0 || _gui_pauses > 2) { 
				printf("Error:\tp value must be between 0, 1 or 2.\n");
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "name:")) {
			strcpy(_test_name, argv[i+1]);
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "load:")) {
			_inj_load = atoi(argv[i+1]);
			if (_inj_load < 10 || _inj_load > 100 ||
				( ceil( _inj_load/10) != _inj_load/10) ) { 
				printf("Error:\tload value must be between [10 100] in steps of 10.\n");
				exit(1); 
			}
			i += 2;
			continue;
		}
		if ( !strcmp(argv[i], "loadsweep")) {
			_load_sweep = true;
			i++; 
			continue;
		}
		if ( !strcmp(argv[i], "bufsweep")) {
			_buffer_sweep = true;
			i++; 
			continue;
		}
		if ( !strcmp(argv[i], "excel")) {
			_use_excel = true;
			i++; 
			continue;
		}
		if ( !strcmp(argv[i], "TESTCASE:")) {
			printf("\n\nTHIS MUST BE DONE ONLY IN LINUX\n\n\n");
			_testcase_creation = true;
			_testcase_multiplier = atof(argv[i+1]);
			i += 2;
			continue;
		}

		printf("Error:\tParameter #%d '%s' not recognized.\n", i, argv[i]);
		exit(1);
	} // while;
	
	return result;
}

bool HETEROGENEOUS_NOC::search_n_fps_floorplans( vector <ONE_OF_THE_BEST_FLOOR_PLANS> &best_fps) 
{
	// here we basically run the floorplanner _n_fps times and record the
	// _n_best ones only; the best ones will be used outside this function to
	// simulate them using the vNOC simulator;
	// also create _application_graph, a one time action;

	bool result = true;

	//#########################################
	//# Try a 'n_fps' number of floorplanings #
	//#########################################

	B_Tree *fp_p;
	vector<ONE_OF_THE_BEST_FLOOR_PLANS>::iterator pos_Iterator;
	ONE_OF_THE_BEST_FLOOR_PLANS temp_best_floor_plan;
	Module_Info temp_module;
	double fp_aspect_ratio = 1.0;
	int num_rejected_fps = 0;
	int THRESHOLD_TRIALS_COUNT_PER_ATTEMPT = 6;

	// loop _n_fps times and run the BTree floorplanner; record the best ones;
	for ( int i = 1; i <= _n_fps; i++) {
		
		//#######################################
		//# B*Tree representation and Annealing #
		//#######################################

		printf("\n\n\n\n\n\nFloorplan annealing attempt number: %d\n\n", i);

		// (1) load the original floorplanning and apply the annealing;
		fp_p = new B_Tree( _alpha);
		fp_p->set_hetero_host( this); // place a copy of this host in the fp object;
		fp_p->read( _inputfile, _fp_scale);
		_modules_N = fp_p->modules_N;


		// (2) if 2D mode: expand the cores to allocate routers in the same plane;
		// every core - regardless of its initial size - is expanded with a constant
		// amount of area dedicated to implementing the router and network interface;
		// I do not do this 
		//if ( _sim_mode == SIMULATED_ARCH_2D) {
		//	printf("\nModules are now expanded for 2D simulation...");
		//	for ( int id = 0; id < _modules_N; id++) {
		//		fp_p->inflate_core( id, 1);
		//	}
		//}
		if ( _verbose) {
			fp_p->show_modules(); // debug;
		}
		fp_p->init();

		// Note: in a one time action only create the _application_graph;
		// this should be done nicer outside this loop, clearly separated
		// and as a stand alone step;
		if ( _application_graph.nodes_count() <= 0) {
			printf("\nCreate the application_graph...");
			_application_graph.create_nodes_and_arcs( fp_p);
			if ( _verbose) {
				_application_graph.print_application_graph(); // debug;
			}
			printf("\nDone create application_graph.\n");
		}
		

		// (3) run floorplaning;
		SA_FLOORPLANING sa_floorplanning( fp_p, _times, _local, _term_temp);
		sa_floorplanning.set_avg_ratio( _avg_ratio);
		sa_floorplanning.set_lambda( _lambda);
		sa_floorplanning.set_alpha( _alpha);
		sa_floorplanning.set_fp_scale( _fp_scale);
		sa_floorplanning.set_verbose(_verbose); // print detailed info?
		sa_floorplanning.run_SA_Floorplaning(); // run floorplanner;


		// (4)
		fp_p->list_information();
		fp_p->show_tree();


		// (5) compute aspect ratio of this fp;
		fp_aspect_ratio = ( fp_p->Height >= fp_p->Width) ?
			( fp_p->Height / fp_p->Width) : ( fp_p->Width / fp_p->Height);
		// if the aspect ration is higher than 1.3 reject this floorplan;
		if ( num_rejected_fps <= THRESHOLD_TRIALS_COUNT_PER_ATTEMPT &&
			fp_aspect_ratio > 1.3) {
			num_rejected_fps ++;
			if ( num_rejected_fps == _n_fps) {
				// I do this because for hp testcase because cores are
				// thin and tall rects, the floorplanner cannot find
				// floorplans with aspect ratio closed to 1; hence, I 
				// relax the wirelength requirement;
				_alpha = fmin( _alpha + 0.5, 1);
			}
			// cancel this attempt and do not count it as part of _n_fps;
			// however, this should not be done too many (infinite) times;
			i --;
			delete fp_p;
			continue;
		}
		// reset num_rejected_fps so that during next "i" iteration we'll
		// explore more floorplans until we get one with good aspect ratio;
		num_rejected_fps = 0;


		//#################################################
		//# Verify if the resulting FP is one of the best #
		//#################################################

		bool this_fp_is_better = false; // should go into bests list;
		double worst_crit = 0;
		int pos_best, pos_worst, erase_pos;
		double aspect_ratio = 1.0;

		
		// start filling the list...
		if ( best_fps.size() < _n_best) {
			this_fp_is_better = true; // first _n_best fp's go into list anyway;
			pos_best = -1; // signals any position, just use push_back;
		} else {
			// if list has already a number of n_bests, find the 
			// "worst of the best" to be potentially replaced with the new one;
			for ( int pos = 0; pos < best_fps.size(); pos++) {
				aspect_ratio = 
					( best_fps[pos].Height >= best_fps[pos].Width) ?
					( best_fps[pos].Height / best_fps[pos].Width) :
					( best_fps[pos].Width / best_fps[pos].Height);
				switch ( _fp_criteria) {
				case 'W':
					if (aspect_ratio > 1.3 || best_fps[pos].WireLength > worst_crit) {
						worst_crit = best_fps[pos].WireLength;
						pos_worst = pos;
					}
					break;
				case 'A':
					if (aspect_ratio > 1.3 || best_fps[pos].Area > worst_crit) {
						worst_crit = best_fps[pos].Area;
						pos_worst = pos;
					}
					break;
				default:
					assert(false);
				}
			}
			// compute also the aspect ratio of the worst floorplan in 
			// current bests list;
			aspect_ratio = 
				( best_fps[pos_worst].Height >= best_fps[pos_worst].Width) ?
				( best_fps[pos_worst].Height / best_fps[pos_worst].Width) :
				( best_fps[pos_worst].Width / best_fps[pos_worst].Height);
			// after scaning the whole bests list, if the actual is better than
			// the "worst of the best", assign it to the list;
			switch ( _fp_criteria) {
			case 'W':
				if ( (fp_p->getWireLength() < best_fps[pos_worst].WireLength) ||
					(fp_p->getWireLength() < 1.5*best_fps[pos_worst].WireLength &&
					fp_aspect_ratio < aspect_ratio) ) {
					pos_best = pos_worst;
				 	this_fp_is_better = true;
				}
				break;
			case 'A':
				if ( (fp_p->getArea() < best_fps[pos_worst].Area) ||
					(fp_p->getArea() < 1.5*best_fps[pos_worst].Area &&
					fp_aspect_ratio < aspect_ratio)) {
					pos_best = pos_worst;
					this_fp_is_better = true;
				}
				break;
			}
		}

		if ( this_fp_is_better) {
			printf("\nAttempt number %d goes to the list of best floorplans...\n" ,i);

			// copy all the info from this floorplaning to a temporary container;
			temp_best_floor_plan.attempt_n = i;
			temp_best_floor_plan.Width = fp_p->Width;
			temp_best_floor_plan.Height = fp_p->Height;
			temp_best_floor_plan.Area = fp_p->getArea();
			temp_best_floor_plan.WireLength = fp_p->getWireLength();
			temp_best_floor_plan.modules_N = _modules_N;
			for ( int id = 0; id < _modules_N; id++) {
				// copy all info of each module to a temporary module;
				temp_module.x = fp_p->modules_info[id].x;
				temp_module.y = fp_p->modules_info[id].y;
				temp_module.rx = fp_p->modules_info[id].rx;
				temp_module.ry = fp_p->modules_info[id].ry;
				temp_module.rotate = fp_p->modules_info[id].rotate;
				temp_module.flip = fp_p->modules_info[id].flip;
				// push that temp. module to the vector inside the temp. best fp;
				temp_best_floor_plan.modules_info.push_back( temp_module);
			}

			// add this temp. to the vector of best floorplans;
			if ( pos_best == -1)
				best_fps.push_back( temp_best_floor_plan);
			else {
				printf("\nRemoving attempt number %d from the bests list...\n",
					best_fps[pos_worst].attempt_n);
				pos_Iterator = best_fps.begin(); // erase() requires an iterator;
				erase_pos = 0;
				while ( erase_pos < pos_worst) {
					erase_pos ++;
					pos_Iterator ++;
				}
				best_fps.erase( pos_Iterator);
				best_fps.push_back( temp_best_floor_plan);
			}
		}

		// prepare for the next floorplan
		temp_best_floor_plan.modules_info.clear();
		delete fp_p;
	}

	return result;
}

bool HETEROGENEOUS_NOC::calculate_final_results_statistics( int argc,char **argv,
	vector <ONE_OF_THE_BEST_FLOOR_PLANS> &best_fps)
{
	// does also printouts and prints in excel file;
	//
	// results format:
	//
	//		   Attempt (0=#1, 1=#2, ...)
	//			  |
	//			  |
	//			  |		 buffer size (0=1x, 1=2x, ...)
	//			  |			   |
	//			  |			   |
	//			  |			   |  Injection load (0=10%, 1=20%, 2=30%, ...)
	//			  |			   |	|
	//			  |			   |	|
	//			  V			   V	V
	//	best_fps[ * ].results[ * ][ * ].latency				[ns]
	//					"			   .load				(10,20,30,...100)
	//					"			   .buffer_multiplier	(1,2,...5)
	//
	bool result = true;



	// (1) print a tag ("name") to the results.txt file for easy identification;
	// we save results in "results/results.txt"; this should be controlled by user?
	FILE *p_results_file;
	char results_filename[268];

	p_results_file = fopen("results/results.txt", "a");
	fprintf(p_results_file, "%s:\n===============================\n", &_test_name);
	// print the parameters to the results.txt file for extra information and
	// future replication;
	for ( int i = 0; i < argc; i++) {
		fprintf(p_results_file, "%s ", argv[i]);
	}
	fprintf(p_results_file, "\n\n");



	// (2) spreadsheet output;
	FILE *p_excel;
	char excelname[281];
	// temp storage; will be flushed to excel file;
	vector<double> sketch_packets_per_cycle;

	if ( _use_excel) {
		sprintf( excelname, "results/%s.xls", &_test_name);
		p_excel = fopen( excelname, "w");
		if ( _load_sweep) {
			fprintf(p_excel, "Buffers\\Load[%]\t20\t30\t40\t50\t60\t70\t80\t90\t100\n");
		} else {
			fprintf(p_excel, "Buffers\\Load[%]\t%d\n", _inj_load);
		}
	}

	printf("RESULTS SUMMARY:\n================\n\n");

	// initially use the original buffers size;
	int buffer_multiplier = 1;

	// loop for several buffer sizes;
	do {

		sketch_packets_per_cycle.clear();
		// first collumn in the spreadsheet;
		if ( _use_excel) {
			fprintf( p_excel, "%dx\t", buffer_multiplier);
		}
		
		// index is 0 for a single buffer simulation; in sweep 0->1x, 1->2x, 2->3x...
		int buf_sweep_i = buffer_multiplier - 1;
		printf("BUFFERS SIZE: %dx\n\n", buffer_multiplier);
		fprintf(p_results_file, "BUFFERS SIZE: %dx\n\n", buffer_multiplier);

		// index is 0 for a single injection load; in sweep 0->10%, 1->20%, 2->30%...
		int load_sweep_i = 0;
		// get the injection load from the command line;
		int load = _inj_load;
		// if a load sweep is going to be done, start at 10%
		if ( _load_sweep == true) {
			load = 20;
		}


		// loop for several injection loads;
		do {
			double latency_sum=0.0, avg_latency=0.0, min_latency=DBL_MAX;
			double std_deviation=0.0, squared_deviations_sum=0.0;
			double a_sum=0.0, w_sum=0.0; // area, wirelength;
			double avg_a=0.0, avg_w=0.0, min_a=DBL_MAX, min_w=DBL_MAX;
			double stddev_a=0.0, squared_dev_a_sum=0.0;
			double stddev_w=0.0, squared_dev_w_sum=0.0;			
			double packets_per_cycle_of_min_latency = 0;

			packets_per_cycle_of_min_latency =
				best_fps[0].results[0][load_sweep_i].packets_per_cycle;
			printf("INJECTION LOAD: %d%%  packets/cycle: %.4lf \n",
				   load, packets_per_cycle_of_min_latency);
			fprintf(p_results_file, "INJECTION LOAD: %d%%  packets/cycle: %.4lf \n",
					load, packets_per_cycle_of_min_latency);
			// record packets_per_cycle; will be flushed to excell file too;
			sketch_packets_per_cycle.push_back( packets_per_cycle_of_min_latency);

			for (int pos = 0; pos < best_fps.size(); pos++) {

				// verify if the indexing for a load sweep and/or 
				// buffer sweep is right;
				assert( load == 
					best_fps[pos].results[buf_sweep_i][load_sweep_i].load);
				assert( buffer_multiplier == 
					best_fps[pos].results[buf_sweep_i][load_sweep_i].buffer_multiplier);

				// calculate the sum of all latencies (for this load %) 
				// and minimum;
				latency_sum += best_fps[pos].results[buf_sweep_i][load_sweep_i].latency;
				a_sum += best_fps[pos].Area;
				w_sum += double(best_fps[pos].WireLength);
				
				printf("Attempt: %d\tA: %.2f\tW: %d\tFlit latency: %.2lf\n",
					best_fps[pos].attempt_n,
					best_fps[pos].Area, best_fps[pos].WireLength,
					best_fps[pos].results[buf_sweep_i][load_sweep_i].latency);
				fprintf(p_results_file, "Attempt: %d\tA: %.2f\tW: %d\tFlit latency: %.2lf\n",
					best_fps[pos].attempt_n,
					best_fps[pos].Area, best_fps[pos].WireLength,
					best_fps[pos].results[buf_sweep_i][load_sweep_i].latency);
				min_latency = min(min_latency, 
					best_fps[pos].results[buf_sweep_i][load_sweep_i].latency);
				min_a = min( min_a, best_fps[pos].Area);
				min_w = min( min_w, double(best_fps[pos].WireLength));
			}

			// calculate the average latency and square deviation;
			avg_latency = latency_sum / best_fps.size();
			avg_a = a_sum / best_fps.size();
			avg_w = w_sum / best_fps.size();
			for (int pos=0; pos<best_fps.size(); pos++) {
				squared_deviations_sum += pow(best_fps[pos].results[buf_sweep_i][load_sweep_i].latency - avg_latency,2);
				squared_dev_a_sum += pow( best_fps[pos].Area - avg_a, 2);
				squared_dev_w_sum += pow( best_fps[pos].WireLength - avg_w, 2);
			}
			std_deviation = sqrt( squared_deviations_sum / best_fps.size());
			stddev_a =  sqrt( squared_dev_a_sum / best_fps.size());
			stddev_w =  sqrt( squared_dev_w_sum / best_fps.size());			

			printf("\nAverage flit latency: %.02lf\n", avg_latency);
			fprintf(p_results_file,"\nAverage flit latency: %.02lf\n", avg_latency);
			printf("Standard deviation: %.02lf\n", std_deviation);
			fprintf(p_results_file,"Standard deviation: %.02lf\n", std_deviation);
			printf("\nMinimum flit latency: %.02lf\n\n\n\n", min_latency);
			fprintf(p_results_file,"\nMinimum flit latency: %.02lf\n\n", min_latency);
			if ( _use_excel) {
				fprintf(p_excel,"%.02lf\t", min_latency);
			}

			// prepare for the next summary if a load sweep is being done;
			load = load + 10;
			load_sweep_i ++;

		} while ( _load_sweep && load <= 100);

		// prepare for the next round if a buffer size sweep is being done;
		buffer_multiplier ++;
		// write in the excel file also the packets/cycle values corresponding to 
		// min latencies already written in;
		if ( _use_excel) {
			fprintf(p_excel, "\n");
			for ( int i = 0; i < sketch_packets_per_cycle.size(); i++) {
				fprintf(p_excel,"%.4lf\t", sketch_packets_per_cycle[i]);
			}
		}

	} while ( _buffer_sweep && buffer_multiplier <= 5);


	// record in output file also the runtime;
	stop_chronometer();
	clock_t diff_clock = _end_clock - _start_clock;
	fprintf(p_results_file, "hnoc total cputime (processor time, clock()) = %.3f sec\n",
		(double)diff_clock/CLOCKS_PER_SEC);
	double diff_sec_usec = _end_wall.tv_sec - _start_wall.tv_sec + 
		double(_end_wall.tv_usec - _start_wall.tv_usec) / 1000000.0;
	fprintf(p_results_file, "hnoc total walltime (wall clock time) = %.3f sec\n\n\n", diff_sec_usec);

	// close files;
	fclose( p_results_file);
	if ( _use_excel) {
		fclose(p_excel);
	}

	return result;
}


bool HETEROGENEOUS_NOC::routers_assignment_routing_calculation_simulation(
	vector <ONE_OF_THE_BEST_FLOOR_PLANS> &best_fps)
{
	// 1 -- take every best floorplan and do router assignment - using
	//      the Hungarian algo
	// 2 -- do routing path calculation using MCF
	// 3 -- do buffer size refinement
	// 4 -- network simulation using the vNOC simulator
	bool result = true;


	for ( int pos = 0; pos < best_fps.size(); pos++) {
		printf("\n\n\n\n\n\nSimulating attempt number: %d.\n\n",
			best_fps[pos].attempt_n);



		// () create routers graph; this graph has nodes for all
		// routers at corners, and arcs between any two closer
		// than a threshold; the nodes of this graph will represent the
		// right-hand-nodes of the bipartite graph used later for IP/core
		// to router assignmnet;
		printf("\nCreate the routers_graph...");
		ROUTERS_GRAPH routers_graph;
		int cores_max_width_height = // Note: used for routers_graph and noc_graph;
			routers_graph.create_nodes( best_fps[pos]);
		//int threshold = cores_max_width_height;
		int threshold =	min( int(best_fps[pos].Height), int(best_fps[pos].Width));
		
		// Note: for the purpose of creating the routers graph we want
		// a smaller "cores_max_width_height" so that router-nodes with more
		// neighbors from the middle of the chip will have higher weights
		// and will be preferred during router assignmnet; otherwise if this
		// threshold is too large (say as it is calculated above: as the largest 
		// core size) then all router-nodes will have more or less the same 
		// weight and final routers assigned will cover almost entire chip;
		// even the corner ones may be selected;
		int threshold_rg = threshold / 5; // /5
		routers_graph.create_arcs( threshold_rg);
		routers_graph.compute_node_weights();
		if ( _verbose) {
			routers_graph.print_routers_graph(); // debug;
		}
		printf("\nDone create routers_graph.\n");

		

		
		//#####################
		//# Router Assignment #
		//#####################

		// () do router assignment using the Hungarian algorithm;
		// this will have as a result the fact that routers_graph will have
		// some of its routers assigned to cores;
		do_routers_assignment( routers_graph);
		printf("\nDone routers assignment.\n");


		// () create noc (routers) graph; this graph has nodes only a subset
		// of the nodes of the routers_graph; this subset is formed by
		// the routers which were assigned to a core above; arcs are also
		// a subset of the arcs of routers_graph;
		printf("\nCreate the noc_graph (subset of nodes of routers_graph)...");
		// Note: for the purpose of creating the noc graph we want
		// a relatively larger "cores_max_width_height" in order to get
		// the graph connected; however not too large either as pruning will
		// undo many arcs anyways;
		int threshold_ng = threshold / 3; // /3
		ROUTERS_GRAPH noc_graph;
		bool grow_the_noc_graph = true; // false;
		if ( grow_the_noc_graph) {
			// grow the graph; 
			// Note: results are worse; so not used currently;
			noc_graph.create_nodes_and_arcs_by_growing( &routers_graph, threshold_ng);
			while ( noc_graph.check_and_fix_full_connectivity()) {};
			noc_graph.compute_arc_weights();
			noc_graph.compute_max_manhattan_link_length(); // will determine T clock;
		} else {
			// create all arcs and then remove some of them;
			// Note: graph may result disconnected;
			noc_graph.create_nodes_and_arcs( &routers_graph, threshold_ng);
			noc_graph.compute_arc_weights();
			// remove some of the arcs because there could be too much connectivity - 
			// which would translate in many ports for each router, so area as well
			// as wiring that boils down to area too;
			noc_graph.prune_some_arcs_old(); // version1: leads to better results; why?
			//noc_graph.prune_some_arcs(); // version2;
			// if after pruning graph is disconnected fix that;
			while ( noc_graph.check_and_fix_full_connectivity()) {};
			noc_graph.compute_arc_weights();
			noc_graph.compute_max_manhattan_link_length(); // will determine T clock;
		}
		if ( _verbose) {
			noc_graph.print_routers_graph(); // debug;
			noc_graph.print_graph_statistics(); // debug;
		}
		printf("\nDone create noc_graph.\n");		



		//#############################
		//# Routing Paths Calculation #
		//#############################

		// () use MCF to find a fractional multicommodity flow solution, which will
		// be used to do randomized rounding and find the routing paths for the 
		// packets of each commodity flow;
		printf("\nRouting paths calculation based on: ");
		double edge_capacity = _application_graph.max_arcs_comm_volume();
		MCF mcf(_verbose);
		mcf.build_network_from_host_application(
			noc_graph, _application_graph, edge_capacity);
		//mcf.print_network_demands( false); // print_only_edges;
		// now either call the actual MCF solver and run the MCF algo
		// or do just a simple shortest path calculation instead;
		if ( routing_paths_calc() == ROUTING_PATHS_MCF) {
			// routing paths are calculated using MCF approach + randomization;
			printf("Fractional MCF + randomization...");
			mcf.mcf_solver();
			// use the MCF result (that gives a path for each commodity) to
			// establish the routing paths thru the NOC topology; this info 
			// will be used to construct the routing look-up tables stored 
			// inside each router;
			mcf.do_randomized_rounding();
		} else {
			// ROUTING_PATHS_SHORTEST: simply Dijkstra shortest paths; 
			// mcf is a dummy object used for this purpose;
			printf("Dijkstra shortest path...");
			mcf.do_shortest_paths_as_preferred_paths_with_inf_bandwidth(); //best case scenario;
			//mcf.do_shortest_paths_as_preferred_paths_with_limited_bandwidth();
		}
		mcf.calculate_average_path_length(); // should be gated by verbose;
		if ( _verbose) {
			mcf.print_max_accumulated_flow();
			printf("\nEdge capacities %.2f", _application_graph.max_arcs_comm_volume());
			mcf.print_network_demands( true); // print_only_edges;
			mcf.print_routing_paths();
		}

		printf("\nDone routing paths calculation.\n");
		// Note: mcf.free_topology() is called later on to clean up,
		// free memory used by mcf object;

		

		//##################
		//# NOC Simulation #
		//##################

		// () call the vNOC simulator to simulate every network topology for
		// the required injection load();
		double link_length = BASE_WIRE; // 1000 um;
		//double link_length = noc_graph.max_manhattan_link_length();
		printf("Link length: %.01f", link_length);


		_ary = (int) ceil( sqrt( _modules_N)) + _x_ary; // ???
		// initially use the original buffers size;
		int buffer_multiplier = 1;

		do {

			RESULTS dummy_list;
			best_fps[pos].results.push_back( dummy_list);
			// get the injection load from the command line;
			int load = _inj_load;
			// if a load sweep is going to be done, start at 10%;
			if ( _load_sweep == true) {
				load = 20;
			}


			do {

				printf("\n\nStarting VNOC simulation of attempt: %d at load %d%%...\n\n",
					   best_fps[pos].attempt_n, load);
				_skipped_packets = (100 - load) / 10; // with 100% load;
				_max_skip_counter = (10 - _skipped_packets);

				// ctor of VNOC_APPLICATION	does everything; calls the vNOC
				// simulator, etc.; also puts the final latency result in
				// _sketch_latency of hetero;	
				VNOC_APPLICATION vnoc_app(this, // pass the HETEROGENEOUS_NOC object as its host/owner;
										  _ary,
										  _inp_buf * buffer_multiplier,
										  _out_buf * buffer_multiplier, 
										  _vc_n, _flit_size, _link_bw, link_length, 
										  _pipeline_in_link, _inputfile, _seed,
										  _routing_a, _cycles, _warmup_cycles, _use_gui,
										  &best_fps[pos],
										  &noc_graph, &mcf, _gui_pauses);

				RESULT temp_result;
				temp_result.load = load;
				// retrieve latency as final result of the vNOC simulator;
				temp_result.latency = _sketch_latency;
				temp_result.packets_per_cycle = _sketch_packets_per_cycle;
				temp_result.buffer_multiplier = buffer_multiplier;
				// with a single injection load the latency will be stored at
				// best_fps[pos].results[0].latency
				// if a load sweep is done, we'll have results for 10% at results[0],
				// 20% at results[1], 30% at results[2], ...
				best_fps[pos].results[buffer_multiplier-1].push_back( temp_result);

				// prepare for the next simulation if a load sweep is being done;
				load = load + 10;

			} while ( _load_sweep && load <= 100);

			// prepare for the next round if a buffer size sweep is being done also;
			buffer_multiplier ++;

		} while ( _buffer_sweep && buffer_multiplier <= 5);



		// () free memory used by the MCF object;
		mcf.free_topology();
	
	} //for

	return result;
}

bool HETEROGENEOUS_NOC::do_routers_assignment( ROUTERS_GRAPH &routers_graph)
{
	// () do router assignment using the Hungarian algorithm;
	printf("\nDo routers assignment...");

	// (a) create Hungarian object;
	HUNGARIAN_ONE hungarian; // will solve the linear assignment problem;
	int router_nodes_count = routers_graph.nodes_count();
	// weight for edges that do not "exist" between two nodes in the bipartite
	// graph will have however a max weight internal to the Hungarian algo;
	hungarian.initialize( _modules_N, router_nodes_count,
						  2 * routers_graph.max_node_weight()); // max_cost;

	// (b) visit each router-node and look at its list of cores;
	// create bipartite edges with weight calculated to account for
	// the out-degree of node and its accumulated distance to all its
	// neighbor nodes;
	int max_node_weight_RG = int(routers_graph.max_node_weight());
	int max_nodes_comm_volume_AG = int(_application_graph.max_nodes_comm_volume());
	for ( int r_id = 0; r_id < router_nodes_count; r_id++) {

		ROUTERS_NODE *this_node = routers_graph.get_node( r_id);
		int bipartite_edge_cost = max_node_weight_RG -
			this_node->weight() + 1;

		for ( long k = 0; k < this_node->cores().size(); k ++) {
			int core_id = this_node->cores()[k];

			// get the communication volume of the node from the application
			// graph that has this core_id;
			int comm_volume_of_core_id = max_nodes_comm_volume_AG +
				int (_application_graph.get_node(core_id)->io_comm_volume()) + 1;
			int cost = bipartite_edge_cost + comm_volume_of_core_id;

			// add new edge to the bipartite graph;
			hungarian.set_cost( core_id, r_id, cost);
		}
	}

	
	// (c) call the actual magyar man;
	//hungarian.print_hungarian_assignment(); // debug;
	hungarian.run_hungarian();

	// (d) get assignments;
	if ( _verbose) {
		hungarian.print_hungarian_assignment(); // debug;
	}
	
	for ( int core_id = 0; core_id < _modules_N; core_id++) {
		long r_id = hungarian.report_assignment_of( core_id);
		routers_graph.get_node( r_id)->set_assigned_core( core_id);
	}
	//routers_graph.print_routers_graph(); // debug;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// ROUTERS_GRAPH
//
////////////////////////////////////////////////////////////////////////////////
//
// implements the routers graph with all routers: used to do routers
// assignment using the Hungarian algo;
// implements also the noc graph as a subset of the above routers: those
// that got assigned to IP/core and which is used as input to the MCF alo
//
////////////////////////////////////////////////////////////////////////////////

int ROUTERS_GRAPH::create_nodes( ONE_OF_THE_BEST_FLOOR_PLANS &best_fp)
{
	// create router-nodes; for each IP/core of the floorplan we can
	// create up to 4 nodes; if 2,3 or 4 cores "share" common corners (with
	// same coordinates), then only one node is created;
	// return the max width or height among all modules;
	//
	// reminder about corners identification;
	//		   NW---NE (rx,ry)
	//		   |	 |
	//		   |	 |				<- for each modules_info[]
	//	 (x,y) SW---SE
	// 
	// 0,0

	int max_width_height = -1; // to be returned;
	int modules_N = best_fp.size();
	int x = -1, y = -1, node_id = -1;
	int this_w = 0, this_h = 0;

	for ( int core_id = 0; core_id < modules_N; core_id++) {
		// attempt to create a router node for each corner of this IP/core;
		// nodes have to have unique coordinates;
		this_w = best_fp.modules_info[core_id].rx - best_fp.modules_info[core_id].x;
		this_h = best_fp.modules_info[core_id].ry - best_fp.modules_info[core_id].y;
		assert( this_w > 0 && this_h > 0);
		if ( this_w > max_width_height) {
			max_width_height = this_w;
		} else if ( this_h > max_width_height) {
			max_width_height = this_h;
		}
		// (a) SW
		x = best_fp.modules_info[core_id].x;
		y = best_fp.modules_info[core_id].y;
		node_id = node_exists_with_coords( x, y);
		assert( node_id < _nodes_count);
		if ( node_id >= 0) { // means a node already exists with coordinates x,y;
			// just add this IP/core id to the list of found node;
			_nodes[ node_id].add_cores( core_id);
		} else { // -1 means no node created so far has coordinates x,y;
			_nodes.push_back( ROUTERS_NODE( _nodes_count, x, y));
			_nodes[ _nodes_count].add_cores( core_id);
			_nodes_count ++; // prepare for next one;
		}
		// (b) SE
		x = best_fp.modules_info[core_id].rx;
		y = best_fp.modules_info[core_id].y;
		node_id = node_exists_with_coords( x, y);
		assert( node_id < _nodes_count);
		if ( node_id >= 0) {
			_nodes[ node_id].add_cores( core_id);
		} else {
			_nodes.push_back( ROUTERS_NODE( _nodes_count, x, y));
			_nodes[ _nodes_count].add_cores( core_id);
			_nodes_count ++;
		}
		// (c) NW
		x = best_fp.modules_info[core_id].x;
		y = best_fp.modules_info[core_id].ry;
		node_id = node_exists_with_coords( x, y);
		assert( node_id < _nodes_count);
		if ( node_id >= 0) {
			_nodes[ node_id].add_cores( core_id);
		} else {
			_nodes.push_back( ROUTERS_NODE( _nodes_count, x, y));
			_nodes[ _nodes_count].add_cores( core_id);
			_nodes_count ++;
		}
		// (d) NE
		x = best_fp.modules_info[core_id].rx;
		y = best_fp.modules_info[core_id].ry;
		node_id = node_exists_with_coords( x, y);
		assert( node_id < _nodes_count);
		if ( node_id >= 0) {
			_nodes[ node_id].add_cores( core_id);
		} else {
			_nodes.push_back( ROUTERS_NODE( _nodes_count, x, y));
			_nodes[ _nodes_count].add_cores( core_id);
			_nodes_count ++;
		}
	}
	// by now all nodes are created; will have to create arcs too;
	
	//printf("\n>>> max_width_height = %d\n", max_width_height);
	return ( max_width_height);
}

void ROUTERS_GRAPH::create_nodes_and_arcs( ROUTERS_GRAPH *routers_graph,
	int threshold_dist)
{
	// this is the version that is used to create a noc_graph starting
	// from a routers_graph; the noc_graph is canonical w/o redundant
	// stuff;

	// (1) nodes;
	int router_nodes_count = routers_graph->nodes_count();
	for ( int r_id = 0; r_id < router_nodes_count; r_id++) {
		ROUTERS_NODE *this_node = routers_graph->get_node( r_id);

		if ( this_node->assigned_core() >= 0) { // router is hooked to an IP/core;
			_nodes.push_back(
				ROUTERS_NODE( _nodes_count, this_node->x(), this_node->y()));
			_nodes[ _nodes_count].set_assigned_core( this_node->assigned_core());
			_nodes_count ++; // prepare for next noc graph router-node;

		} // else skip it as this router will not be used by the hetero NOC;
	}

	// (2) compute mapping _core_id_to_router_id;
	_core_id_to_router_id.resize( _nodes_count);
	for ( int i = 0; i < _nodes_count; i++) {
		_core_id_to_router_id[ _nodes[i].assigned_core() ] = i;
	}

	// (3) arcs;
	// Note: arcs of noc_graph will be created exactly the same way as arcs
	// of the routers_graph;
	create_arcs( threshold_dist);
}

void ROUTERS_GRAPH::create_arcs( int max_dist, int min_dist)
{
	// Note: by default min_dist = 0;
	// an arc is created between 2 router-nodes if the distance between
	// them is less than a user-specified distance or by default the
	// max width or height among all modules;
	// Note: size of adj lists should be controlled because that is
	// basically the number of ports routers will have to support;

	int dist = 0;
	for ( int i = 0; i < _nodes_count; i++) {
		for ( int j = i + 1; j < _nodes_count; j++) {

			// use Manhattan distance for now;
			dist = dist_euclidian( _nodes[i].x(), _nodes[i].y(),
				_nodes[j].x(), _nodes[j].y());
			if ( dist <= max_dist && dist >= min_dist) { // add this arc;
				_arcs.push_back( ROUTERS_ARC( _arcs_count, i, j));
				_nodes[ i].add_adj( _arcs_count);
				_nodes[ j].add_adj( _arcs_count);
				_arcs_count ++; // prepare for next one;
			}
		}
	}
}

void ROUTERS_GRAPH::create_nodes_and_arcs_by_growing(
	ROUTERS_GRAPH *routers_graph, int threshold_dist)
{
	// this is the version that is used to create a noc_graph starting
	// from a routers_graph; the noc_graph is canonical w/o redundant
	// stuff;
	// this is the version that grows arcs and guarantees connectivity: when
	// a new node is processed at least one arc is created to undiscovered
	// nodes;

	double max_dist = double( threshold_dist);

	// (1) nodes;
	int router_nodes_count = routers_graph->nodes_count();
	for ( int r_id = 0; r_id < router_nodes_count; r_id++) {
		ROUTERS_NODE *this_node = routers_graph->get_node( r_id);

		if ( this_node->assigned_core() >= 0) { // router is hooked to an IP/core;
			_nodes.push_back(
				ROUTERS_NODE( _nodes_count, this_node->x(), this_node->y()));
			_nodes[ _nodes_count].set_assigned_core( this_node->assigned_core());
			_nodes_count ++; // prepare for next noc graph router-node;

		} // else skip it as this router will not be used by the hetero NOC;
	}


	// (2) compute mapping _core_id_to_router_id;
	_core_id_to_router_id.resize( _nodes_count);
	for ( int i = 0; i < _nodes_count; i++) {
		_core_id_to_router_id[ _nodes[i].assigned_core() ] = i;
	}


	// (3) arcs;
	// start with the first node and create arcs to the closest ones; 
	// then take as a second one the first one to which an arc was created
	// and process it and so on;
	int MAXIMUM_ADJ_SIZE = 4;
	// Note: MINIMUM_ADJ_SIZE not used here;
	vector<int> was_discovered; // will be set as "1" to reflect that was discovered;
	for ( int i = 0; i < _nodes_count; i++) {
		was_discovered.push_back( -1); // not processed;
	}
	list<int> fifo;
	int process_i = 0;
	int count_processed = 0;
	while ( count_processed < _nodes_count) {

		// () record that was processed;
		was_discovered[process_i] = 1;
		count_processed ++;
		//printf("\n -> process: %d", process_i);

		// () check if this node should be skipped;
		if ( _nodes[process_i].adj().size() >= MAXIMUM_ADJ_SIZE &&
			 fifo.size() > 0) {
			// do not process and go to the next node in fifo queue;
			process_i = fifo.back();
			fifo.pop_back();
			continue;
		}
		
		// () otherwise, we have to add new nodes either because adj list was too
		// small or this is the last one in the fifo list and we really
		// have to connect to an undiscovered node so that the graph is not
		// left disconnected;
		vector<PAIR_TWO> distances;
		for ( int i = 0; i < _nodes_count; i++) {
			if ( i != process_i) { // no self loops;
				double dist = dist_euclidian( _nodes[i].x(), _nodes[i].y(),
					_nodes[process_i].x(), _nodes[process_i].y());
				distances.push_back( PAIR_TWO(i, double(dist)));
			}
		}
		sort( distances.begin(), distances.end());
		int nodes_count_1 = _nodes_count - 1;
		
		// () now take the closest nodes and create arcs for them; with some
		// restrictions;
		bool added_nonprocessed_adj_node = false;
		int ii = 0; // skip the closest ones;
		while ( _nodes[process_i].adj().size() < MAXIMUM_ADJ_SIZE && ii < nodes_count_1) {
			int adj_i = distances[ii].id();
			// if adj_i has been processed, then it may be already in the adj
			// list of this current process_i node; in that case it does not need 
			// to be added again;
			if ( distances[ii].value() < max_dist) {
				if ( was_discovered[adj_i] > 0) {
					// was processed; may be already in adj list; if not then add only 
					// if the neighbor does not have already too many in its adj list;
					if ( index_in_adj_of_router_id( process_i, adj_i) < 0 &&
						 _nodes[adj_i].adj().size() < MAXIMUM_ADJ_SIZE) {
						_arcs.push_back( ROUTERS_ARC( _arcs_count, process_i, adj_i)); // new arc;
						_nodes[ process_i].add_adj( _arcs_count);
						_nodes[ adj_i].add_adj( _arcs_count);
						//printf("\n add: %d", adj_i);
						_arcs_count ++; // prepare for next one;
					}				
				} else {
					added_nonprocessed_adj_node = true;
					fifo.push_front( adj_i);
					was_discovered[adj_i] = 1;
					_arcs.push_back( ROUTERS_ARC( _arcs_count, process_i, adj_i)); // new arc;
					_nodes[ process_i].add_adj( _arcs_count);
					_nodes[ adj_i].add_adj( _arcs_count);
					//printf("\n add1, also in fifo: %d", adj_i);
					_arcs_count ++; // prepare for next one;
				}
			}
			ii += 1;
		}
		// we really have to connect to an undiscovered node so that the 
		// graph is not left disconnected;

		if ( added_nonprocessed_adj_node == false &&
			 fifo.size() <= 0) {
			ii = 0;
			while ( ii < nodes_count_1) {
				int adj_i = distances[ii].id();
				if ( distances[ii].value() < max_dist) {
					if ( was_discovered[adj_i] < 0) { // not discovered so far;
						fifo.push_front( adj_i);
						was_discovered[adj_i] = 1;
						_arcs.push_back( ROUTERS_ARC( _arcs_count, process_i, adj_i)); // new arc;
						_nodes[ process_i].add_adj( _arcs_count);
						_nodes[ adj_i].add_adj( _arcs_count);
						//printf("\n add2, also in fifo: %d", adj_i);
						_arcs_count ++; // prepare for next one;
						break; // add just one more and that's it;
					}
				}
				ii ++;
			}
		}
		

		// () pick out the next node to grow from;
		if ( fifo.size() > 0) {
			process_i = fifo.back();
			fifo.pop_back();
			//if ( distances.size() > 0) distances.clear();
		}
	}
	
}

void ROUTERS_GRAPH::compute_node_weights()
{
	// for each node we go and compute a node-weight that is proportional
	// to its out-degree and cummulated distance to all its adjacent nodes;
	// record the _max_node_weight too;

	// () first get the max x,y of any node;
	int max_x = 0, max_y = 0;
	for ( int i = 0; i < _nodes_count; i++) {
		if ( _nodes[i].x() > max_x) max_x = _nodes[i].x();
		if ( _nodes[i].y() > max_y) max_y = _nodes[i].y();		
	}

	// ()
	_max_node_weight = 0;
	for ( int i = 0; i < _nodes_count; i++) {
		int x1 = _nodes[i].x();
		int y1 = _nodes[i].y();
		// () compute weight;
		int w = 0;
		for ( long k = 0; k < _nodes[i].adj().size(); k ++) { // adj contains arcs ids;
			int adj_arc_id = _nodes[i].adj()[k]; // id of adjacent arc;
			int adj_i = (_arcs[adj_arc_id].src_id() == i) ? 
				_arcs[adj_arc_id].des_id() : _arcs[adj_arc_id].src_id();
			w += dist_euclidian( x1, y1, _nodes[adj_i].x(), _nodes[adj_i].y());
		}
		int delta_x = max_x - abs( x1 - (max_x - x1) );
		int delta_y = max_y - abs( y1 - (max_y - y1) );
		w = w + ( delta_x + delta_y);
		//w *= _nodes[i].adj().size();
		
		// () record weight;
		_nodes[i].set_weight( w);
		if ( w > _max_node_weight) { 
			_max_node_weight = w; 
		}
	}
}

double ROUTERS_GRAPH::compute_max_manhattan_link_length()
{
	// compute max Manhattan distance to compute T clock;
	double max_l = 0.0;
	for ( int j = 0; j < _arcs_count; j ++) {
		double w = dist_manhattan(
			_nodes[_arcs[j].src_id()].x(), _nodes[_arcs[j].src_id()].y(),
			_nodes[_arcs[j].des_id()].x(), _nodes[_arcs[j].des_id()].y());
		if ( w > max_l) max_l = w;
	}
	_max_manhattan_link_length = int( max_l);
	return max_l;
}

void ROUTERS_GRAPH::compute_arc_weights()
{
	// edge weights are computed as the Manhattan or Euclidian distance 
	// between the src and des locations;

	for ( int j = 0; j < _arcs_count; j ++) {
		// dist_manhattan or dist_euclidian
		double w = dist_euclidian(
			_nodes[_arcs[j].src_id()].x(), _nodes[_arcs[j].src_id()].y(),
			_nodes[_arcs[j].des_id()].x(), _nodes[_arcs[j].des_id()].y());

		_arcs[j].set_weight( w);
	}
}

void ROUTERS_GRAPH::prune_some_arcs()
{
	// remove some of the arcs because there could be too much connectivity;
	int ITER_LIMIT = 16;
	int count = 0;
	while ( prune_some_arcs_one_at_a_time() && 
			count <= ITER_LIMIT) {
		count ++;
	}
}

bool ROUTERS_GRAPH::prune_some_arcs_one_at_a_time()
{
	// called only by prune_some_arcs();

	bool arc_has_been_removed = false;

	// this a soft requirement for max number of ajacent arcs for each node;
	int MAXIMUM_ADJ_SIZE = 3;
	int MINIMUM_ADJ_SIZE = 2;
	int initial_adj_size = -1;

	// initial_arcs_aids will be used to record the new ids of the new fewer 
	// arcs;
	vector<int> initial_arcs_ids;
	for ( int j = 0; j < _arcs_count; j ++) {
		initial_arcs_ids.push_back( -2);
	}


	// store in a vector all max length for each node adj list; then sort
	// the nodes by that, and later will process nodes in decreasing order 
	// of their max length among their adjacent arcs;
	vector<PAIR_TWO> representative_arcs;
	for ( int i = 0; i < _nodes_count; i++) {
		double max_adj_weight = 0;
		for ( int k = 0; k < _nodes[i].adj().size(); k++) {
			if ( _arcs[_nodes[i].adj()[k]].weight() > max_adj_weight) 
				max_adj_weight = _arcs[_nodes[i].adj()[k]].weight();
		}
		representative_arcs.push_back( PAIR_TWO(i, max_adj_weight));
	}
	sort( representative_arcs.begin(), representative_arcs.end());


	
	for ( int u = _nodes_count - 1; u >= 0; u--) {
		// u is to index representative_arcs;
		int i = representative_arcs[u].id();

		
		initial_adj_size = _nodes[i].adj().size();
		if ( initial_adj_size > MAXIMUM_ADJ_SIZE) {

			int x1 = _nodes[i].x();
			int y1 = _nodes[i].y();
			vector<PAIR_TWO> adj_arcs_sketch;

			for ( long k = 0; k < initial_adj_size; k++) {
				int adj_arc_id = _nodes[i].adj()[k]; // id of adjacent arc;
				adj_arcs_sketch.push_back( 
					PAIR_TWO(adj_arc_id, _arcs[adj_arc_id].weight()));
			}
			// sort adj arcs by their length;
			sort( adj_arcs_sketch.begin(), adj_arcs_sketch.end());


			// we will remove the neighbor that is the most remote;
			// but only if by doing that we do not make any of the neighbors
			// be left with too few adj arcs;
			int adj_arc_id = adj_arcs_sketch[initial_adj_size - 1].id();
			int adj_i = (_arcs[adj_arc_id].src_id() == i) ? 
				_arcs[adj_arc_id].des_id() : _arcs[adj_arc_id].src_id();
			// if the adjacent node (which is the most remote to "i")
			// has this node "i" and a few others and in all less than
			// MINIMUM_ADJ_SIZE, then we cannot remove any entry from
			// the adj list of "i"; however if that remote neighbor has
			// more adj entries than MINIMUM_ADJ_SIZE, then we remove it
			// from the adj list of this "i";
			if ( _nodes[adj_i].adj().size() > MINIMUM_ADJ_SIZE) {

				// remove first all adj arcs from _adj vector of this node;
				_nodes[i].remove_all_adj();
				// add back only the closest ones but the last one; notice
				// the "-1" here;
				for ( long k = 0; k < initial_adj_size - 1; k++) {
					_nodes[i].add_adj( adj_arcs_sketch[k].id());
				}

				// remove it from the adj list of the adj_i;
				_nodes[adj_i].remove_adj( adj_arc_id);
				// record that this arc is removed;
				initial_arcs_ids[ adj_arc_id] = -1;

				// record that during this call of this function at least an
				// arc has been removed;
				arc_has_been_removed = true;
			}
		}
	}


	if ( arc_has_been_removed) {
		// now we still need to remove erased arcs from the _arcs array
		// and update _arcs_count, and re-index the remaining arcs and
		// correct their values inside adj arrays of all nodes;
		int new_arcs_count = 0;
		for ( int j = 0; j < _arcs_count; j++) {
			if ( initial_arcs_ids[j] == -2) { // not removed;
				initial_arcs_ids[j] = new_arcs_count;
				// "pull" or copy it in the new j location;
				_arcs[ new_arcs_count].set( new_arcs_count,
											_arcs[j].src_id(), _arcs[j].des_id(), _arcs[j].weight());
				new_arcs_count++;
			} // else -1 has been removed;
		}
		// remove the trailing empty _arcs entries;
		for ( int j = _arcs_count - 1; j >= new_arcs_count; j--) {
			_arcs.pop_back();
		}
		// update _arcs_count too;
		_arcs_count = new_arcs_count;
		// update all adj arrays with new arcs ids;
		for ( int i = 0; i < _nodes_count; i++) {
			for ( long k = 0; k < _nodes[i].adj().size(); k++) {
				_nodes[i].adj()[k] = initial_arcs_ids[ _nodes[i].adj()[k] ];
			}
		}
	}
	

	return arc_has_been_removed;
}


void ROUTERS_GRAPH::prune_some_arcs_old()
{
	// remove some of the arcs because there could be too much connectivity;

	// this a soft requirement for max number of ajacent arcs for each node;
	int MAXIMUM_ADJ_SIZE = 4; // 5;
	int MINIMUM_ADJ_SIZE = 2; // 4;
	int initial_adj_size = -1;

	// initial_arcs_aids will be used to record the new ids of the new fewer 
	// arcs;
	vector<int> initial_arcs_ids;
	for ( int j = 0; j < _arcs_count; j ++) {
		initial_arcs_ids.push_back( -2);
	}


	// store in a vector all max length for each node adj list; then sort
	// the nodes by that, and later will process nodes in decreasing order 
	// of their max length among their adjacent arcs;
	vector<PAIR_TWO> representative_arcs;
	for ( int i = 0; i < _nodes_count; i++) {
		double max_adj_weight = 0;
		for ( int k = 0; k < _nodes[i].adj().size(); k++) {
			if ( _arcs[_nodes[i].adj()[k]].weight() > max_adj_weight) 
				max_adj_weight = _arcs[_nodes[i].adj()[k]].weight();
		}
		representative_arcs.push_back( PAIR_TWO(i, max_adj_weight));
	}
	sort( representative_arcs.begin(), representative_arcs.end());


	
	for ( int u = _nodes_count - 1; u >= 0; u--) {
		// u is to index representative_arcs;
		int i = representative_arcs[u].id();

		
		initial_adj_size = _nodes[i].adj().size();
		if ( initial_adj_size > MAXIMUM_ADJ_SIZE) {
			int x1 = _nodes[i].x();
			int y1 = _nodes[i].y();
			vector<PAIR_TWO> adj_arcs_sketch;

			for ( long k = 0; k < initial_adj_size; k++) {
				int adj_arc_id = _nodes[i].adj()[k]; // id of adjacent arc;
				adj_arcs_sketch.push_back( 
					PAIR_TWO(adj_arc_id, _arcs[adj_arc_id].weight()));
			}
			// sort adj arcs by their length;
			sort( adj_arcs_sketch.begin(), adj_arcs_sketch.end());
			// remove first all adj arcs from _sdj vector of this node;
			_nodes[i].remove_all_adj();
			// add back only the closest ones;
			for ( long k = 0; k < MAXIMUM_ADJ_SIZE; k++) {
				_nodes[i].add_adj( adj_arcs_sketch[k].id());
			}
			// then still keep those which if removed would make for their
			// neighbor be left with zero adj;
			for ( long k = MAXIMUM_ADJ_SIZE; k < initial_adj_size; k++) {
				int adj_arc_id = adj_arcs_sketch[k].id();
				int adj_i = (_arcs[adj_arc_id].src_id() == i) ? 
					_arcs[adj_arc_id].des_id() : _arcs[adj_arc_id].src_id();
				// if the adjacent node has this node "i" as the only neighbor
				// then, do not remove this arc;
				if ( _nodes[adj_i].adj().size() < MINIMUM_ADJ_SIZE) {
					// put back the arc;
					_nodes[i].add_adj( adj_arc_id);
				} else {
					// remove it from the adj list of the adj_i router too and
					// do not put it back here either;
					_nodes[adj_i].remove_adj( adj_arc_id);
					// record that this arc is removed;
					initial_arcs_ids[ adj_arc_id] = -1;
				}
			}
		}
	}
	// now we still need to remove erased arcs from the _arcs array
	// and update _arcs_count, and re-index the remaining arcs and
	// correct their values inside adj arrays of all nodes;
	int new_arcs_count = 0;
	for ( int j = 0; j < _arcs_count; j++) {
		if ( initial_arcs_ids[j] == -2) { // not removed;
			initial_arcs_ids[j] = new_arcs_count;
			// "pull" or copy it in the new j location;
			_arcs[ new_arcs_count].set( new_arcs_count,
				_arcs[j].src_id(), _arcs[j].des_id(), _arcs[j].weight());
			new_arcs_count++;
		} // else -1 has been removed;
	}
	// remove the trailing empty _arcs entries;
	for ( int j = _arcs_count - 1; j >= new_arcs_count; j--) {
		_arcs.pop_back();
	}
	// update _arcs_count too;
	_arcs_count = new_arcs_count;
	// update all adj arrays with new arcs ids;
	for ( int i = 0; i < _nodes_count; i++) {
		for ( long k = 0; k < _nodes[i].adj().size(); k++) {
			_nodes[i].adj()[k] = initial_arcs_ids[ _nodes[i].adj()[k] ];
		}
	}

}

bool ROUTERS_GRAPH::check_and_fix_full_connectivity()
{
	// used only in conjunction with the pruning routine above;
	// after pruning arcs the graph may be disconnected: add a minimum 
	// number of arcs back to make it fully connected again;

	bool result = false;
	
	// (1) start with node index 0 and discover all adjacent nodes; at the
	// end the vector was_discovered will point if there are disconnected
	// components;
	vector<int> was_discovered; // will be set as "1" to reflect that was discovered;
	for ( int i = 0; i < _nodes_count; i++) {
		was_discovered.push_back( -1); // not discovered;
	}

	list<int> fifo;
	int process_i = 0;
	fifo.push_front( process_i); // start with the first node;
	//printf("\n add in fifo: %d", process_i);
	int count_processed = 0;
	while ( fifo.size() > 0) {

		// () process new node from queue;
		process_i = fifo.back();
		fifo.pop_back();

		// () record that was processed;
		was_discovered[process_i] = 1;
		count_processed ++;

		// () discover adj nodes;
		for ( int k = 0; k < _nodes[process_i].adj().size(); k++) { // adj contains arcs ids;
			int adj_arc_id = _nodes[process_i].adj()[k]; // id of adjacent arc;
			int adj_i = (_arcs[adj_arc_id].src_id() == process_i) ? 
				_arcs[adj_arc_id].des_id() : _arcs[adj_arc_id].src_id();
			if ( was_discovered[adj_i] < 0) { // first time discovered;
				fifo.push_front( adj_i);
				was_discovered[adj_i] = 1;
				//printf("\n add in fifo: %d", adj_i);
			}
		}
	}
	//printf("\n--- _nodes_count: %d discovered: %d", _nodes_count, count_processed);

	// (2) at this time if was_discovered has "-1" entries then those are
	// nodes in a different disconnected sub-graph;
	if ( count_processed != _nodes_count) {
		result = true;
		vector<PAIR_THREE> distances;

		for ( int i = 0; i < _nodes_count; i++) {
			if ( was_discovered[i] > 0) {
				for ( int j = 0; j < _nodes_count; j++) {
					if ( was_discovered[j] < 0) {
						int dist = dist_euclidian(
							_nodes[i].x(), _nodes[i].y(), _nodes[j].x(), _nodes[j].y());
						distances.push_back( PAIR_THREE(i, j, dist));
						//printf("\n add dist: %d-%d  %d", i, j, dist);
					}
				}
			}
		}
		sort( distances.begin(), distances.end());		

		// create two arcs to connect the disconnected components; 
		// Note: for now I assume just 2 disconnected components;
		if ( distances.size() > 0) {
			// -- create a first arc;
			_arcs.push_back( ROUTERS_ARC( _arcs_count, distances[0].id1(), distances[0].id2()));
			_nodes[ distances[0].id1()].add_adj( _arcs_count);
			_nodes[ distances[0].id2()].add_adj( _arcs_count);
			//printf("\n added arc: %d-%d  %d", distances[0].id1(), distances[0].id2(), _arcs_count);
			_arcs_count ++;
			// -- go and search for next arc with source and dest. different from 
			// the first one and add it as a second added arc for better connectivity;
			for ( int k = 1; k < distances.size(); k++) {
				if ( distances[0].id1() != distances[k].id1() &&
					 distances[0].id2() != distances[k].id2()) {
					_arcs.push_back( ROUTERS_ARC( _arcs_count, distances[k].id1(), distances[k].id2()));
					_nodes[ distances[k].id1()].add_adj( _arcs_count);
					_nodes[ distances[k].id2()].add_adj( _arcs_count);
					//printf("\n added arc: %d-%d  %d", distances[k].id1(), distances[k].id2(), _arcs_count);
					_arcs_count ++;
					break;
				}
			}
		} else {
			assert (0); // impossible;
		}
	}

	return result;
}

void ROUTERS_GRAPH::print_routers_graph()
{
	// debug;
	printf("\n\nRouters graph");
	printf("\nNodes: %d", _nodes.size());
	for ( long i = 0; i < _nodes_count; i ++) {
		printf("\n Node id: %d (%d, %d) Weight: %d Assigned_core: %d",
			i, _nodes[i].x(), _nodes[i].y(), 
			_nodes[i].weight(), _nodes[i].assigned_core());
		printf("\n\tIP/cores:");
		for ( long k = 0; k < _nodes[i].cores().size(); k ++) {
			printf(" %d", _nodes[i].cores()[k]);
		}
		printf("\n\tadj arcs:");
		for ( long k = 0; k < _nodes[i].adj().size(); k ++) {
			printf(" %d", _nodes[i].adj()[k]);
		}
	}
	printf("\nArcs: %d", _arcs.size());
	for ( long j = 0; j < _arcs_count; j ++) {
		printf("\nArc id: %d %d --- %d    \t%.2lf", _arcs[j].id(),
			   _arcs[j].src_id(), _arcs[j].des_id(), _arcs[j].weight());
	}
	printf("\n");
}

void ROUTERS_GRAPH::print_graph_statistics()
{
	printf("\n Nodes: %d", _nodes_count);
	printf("\n Arcs: %d", _arcs_count);
	int max_degree = 0;
	int cumul_degree = 0;
	for ( long i = 0; i < _nodes_count; i ++) {
		if ( _nodes[i].adj().size() > max_degree) {
			max_degree = _nodes[i].adj().size();
		}
		cumul_degree += _nodes[i].adj().size();
	}
	printf("\n Max node degree: %d", max_degree);
	printf("\n Avg. node degree: %.2f", ((float)cumul_degree / (float)_nodes_count));
	printf("\n\n");
}

////////////////////////////////////////////////////////////////////////////////
//
// APPLICATION_GRAPH
//
////////////////////////////////////////////////////////////////////////////////
//
// this is the graph that models the application task graph; nodes are tasks
// and edges are communications between tasks labeled with their communication
// volume;
//
////////////////////////////////////////////////////////////////////////////////

void APPLICATION_GRAPH::print_application_graph()
{
	// debug;
	printf("\nApplication graph:");
	printf("\nmax_nodes_comm_volume: %.2lf", _max_nodes_comm_volume);
	printf("\nmax_arcs_comm_volume: %.2lf", _max_arcs_comm_volume);
	printf("\nIP/cores: %d", _nodes.size());
	for ( long i = 0; i < _nodes_count; i ++) {
		printf("\n%d  cumul comm vol: %d", i, int(_nodes[i].io_comm_volume()));
		printf("        \tfin:");
		for ( long k = 0; k < _nodes[i].fanin().size(); k ++) {
			printf(" %d", _nodes[i].fanin()[k]);
		}
		printf("\tfout:");
		for ( long k = 0; k < _nodes[i].fanout().size(); k ++) {
			printf(" %d", _nodes[i].fanout()[k]);
		}
		//printf("\tcomm_vol: %.1f", _nodes[i].io_comm_volume());
	}
	printf("\nArcs: %d", _arcs.size());
	for ( long j = 0; j < _arcs_count; j ++) {
		printf("\n%d -- %d    \t%.1f",
			_arcs[j].src_id(), _arcs[j].des_id(), _arcs[j].comm_volume());
	}
	printf("\n");
}

bool APPLICATION_GRAPH::create_nodes_and_arcs( B_Tree *fp_p)
{
	// create application communication graph;

	// (1) nodes;
	int modules_N = fp_p->size();
	for ( int core_id = 0; core_id < modules_N; core_id++) {
		_nodes.push_back( APPLICATION_NODE( _nodes_count));
		_nodes_count ++; // prepare for next one;
	}
	

	// (2) arcs;
	// Note; should be different val for diff testcase, based on how testcases
	// were created?
	double comm_multiplier = 1.0;
	for( int i = 0; i < modules_N; i++) {
		for( int j = 0; j < modules_N; j++) {
			int this_connection = fp_p->get_connection( i, j);

			if ( i != j && this_connection > 0) { // 2 is to ignore power/gnd;
				double this_comm_volume = this_connection * comm_multiplier;

				_nodes[ i].add_fanout( j);
				_nodes[ j].add_fanin( i);

				_arcs.push_back( APPLICATION_ARC( _arcs_count, i, j, this_comm_volume));
				_arcs_count ++; // prepare for next one;
			}
		}
	}
	
	// (3) misc stuff;
	compute_max_nodes_comm_volume();
	compute_max_arcs_comm_volume();

	return true;
}
