#ifndef _HETEROGENEOUS_NOC_H_
#define _HETEROGENEOUS_NOC_H_

#include <utility>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/types.h>
#include <algorithm>
#include <cstring>
#include "config.h"
#include "vnoc_topology.h"
#include "fp_btree.h"

using namespace std;


// corners of IP/cores will be "hooked up" to routers of the NoC;
enum CORNER_NAME { SW, SE, NW, NE, NOT_AFFILIATED };
enum SIM_MODE { SIMULATED_ARCH_2D, SIMULATED_ARCH_3D };
// how are the routing paths calculated: multicommodity flow based
// or Dijkstra shortest path?
enum RUOUTING_PATHS_MODE { ROUTING_PATHS_MCF, ROUTING_PATHS_SHORTEST };

// this is the minimum distance between router-nodes in the routers graph
// we need it because otherwise nodes would be too close too each other
// fore cores with corners very close but not perfectly overlapping;
#define MINIMUM_DIST_BETWEEN_ROUTERS 200

////////////////////////////////////////////////////////////////////////////////
//
// ROUTERS_DISTRIBUTION
//
////////////////////////////////////////////////////////////////////////////////

class ROUTERS_DISTRIBUTION 
{
 public:
	long _nx, _ny; // mesh configuration (right now nx = ny = ary);
	long _sx, _sy; // in the future could not start at (0,0) as usual;
	double _link_length; // distance between routers;

 public:
	ROUTERS_DISTRIBUTION( long x, long y, long x0, long y0, double ll);

	long nx() { return _nx; }
	long ny() { return _ny; }
	long sx() { return _sx; }
	long sy() { return _sy; }
	double link_length() { return _link_length; }
};

////////////////////////////////////////////////////////////////////////////////
//
// ROUTER_ASSIGNMENT
//
////////////////////////////////////////////////////////////////////////////////

class ROUTER_ASSIGNMENT {

 private:
	int **_array_ids;
	CORNER_NAME **_array_corners;
	double **_array_extralinks;
	int _n_rows;
	int _n_columns;
	double _total_extralinks;

 public:
	ROUTER_ASSIGNMENT(int ary);

	void assign(int x, int y, int id, CORNER_NAME corner, double extralink);
	int assigned(int x, int y);
	CORNER_NAME corner_used(int x, int y);
	pair<int, int> core_id_to_router_xy(int id);
	CORNER_NAME corner_used_by_id(int id);
	double extralink_by_id(int id);
	void clear_assignments();

	double total_extralinks() { return _total_extralinks; }
};

////////////////////////////////////////////////////////////////////////////////
//
// ONE_OF_THE_BEST_FLOOR_PLANS
//
////////////////////////////////////////////////////////////////////////////////

class RESULT {
 public:
	// percentage of the injection load used to be assigned after simulation;
	int load;
	double latency;
	double packets_per_cycle;
	int buffer_multiplier; // the buffers size (if multiplied);
 public:
	RESULT() {}
	~RESULT() {}	
};

typedef struct vector<RESULT> RESULTS;	// we use a list as a load sweep is done;

class ONE_OF_THE_BEST_FLOOR_PLANS 
{
 public:
	int attempt_n;
	Modules_Info modules_info;
	double Width, Height;
	double Area;
	long WireLength;
	long modules_N;
	vector<RESULTS> results; // we use a list as a buffer size sweep is done;
 public:
	int size() { return modules_N; }
};

////////////////////////////////////////////////////////////////////////////////
//
// NOC
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//
// ROUTERS_NODE
//
////////////////////////////////////////////////////////////////////////////////

class ROUTERS_NODE {
 private:
	int _id; // id of itself;
	int _x, _y; // coordinates;
	// id's of IP/cores in whose corners this router-node is located;
	vector<int> _cores;
	vector<int> _adj; // id's of adjacent arcs;
	int _weight;
	int _assigned_core; // id of core attached finally to this router-node;
		
 public:
	ROUTERS_NODE( int id, int x, int y) : _id(id), _x(x), _y(y) {
		_weight = -1; _assigned_core = -1;
	}
	~ROUTERS_NODE() {}
	
	int id() const { return _id; }
	int x() const { return _x; }
	int y() const { return _y; }
	int assigned_core() const { return _assigned_core; }
	int weight() const { return _weight; }
	vector<int> &cores() { return _cores; }
	vector<int> &adj() { return _adj; }
	void add_cores( int core_id) {
		_cores.push_back( core_id);
	}
	void add_adj( int id) { // id is index of arc this one connects to;
		_adj.push_back( id);
	}
	void remove_all_adj() { // pop_back all adj arcs;
		for ( long k = _adj.size() - 1; k >= 0; k--) _adj.pop_back();
	}
	void remove_adj( int arc_id) {
		// erase element from _adj list;
		int k = -1;
		for ( k = 0; k < _adj.size(); k++) {
			if ( _adj[k] == arc_id) break;
		}
		if ( k >= 0) {
			_adj.erase( _adj.begin()+k, _adj.begin()+k+1);
		} // else Error;
	}
	void set_weight( int weight) { _weight = weight; }
	void set_assigned_core( int r_id) { _assigned_core = r_id; }
};

////////////////////////////////////////////////////////////////////////////////
//
// ROUTERS_ARC
//
////////////////////////////////////////////////////////////////////////////////

class ROUTERS_ARC {
 private:
	int _id;
	int _src_id;
	int _des_id;
	double _weight;
		
 public:
	ROUTERS_ARC( int id, int src, int des) : _id(id), _src_id(src), _des_id(des) {
		_weight = -1;
	}
	~ROUTERS_ARC() {}
	
	int id() const { return _id; }
	int src_id() const { return _src_id; }
	int des_id() const { return _des_id; }
	double weight() const { return _weight; }
	void set_weight( double weight) { _weight = weight; }
	void set( int id, long src_id, long des_id, double weight) {
		_id = id;
		_src_id = src_id;
		_des_id = des_id;
		_weight = weight;
	}
};

////////////////////////////////////////////////////////////////////////////////
//
// ROUTERS_GRAPH
//
////////////////////////////////////////////////////////////////////////////////

class ROUTERS_GRAPH {
 private:
	long _nodes_count;
	long _arcs_count;
	vector<ROUTERS_NODE> _nodes;
	vector<ROUTERS_ARC> _arcs;
	int _max_node_weight; // max node weight for fast calculations;
	// will determine T clock;
	int _max_manhattan_link_length;
	// mapping between core_id's and router_id's; used only inside
	// the noc_graph not in all objects of this class;
	vector<int> _core_id_to_router_id;

 public:
	ROUTERS_GRAPH() {
		_nodes_count = 0; _arcs_count = 0; 
		_max_node_weight = 1;
		_max_manhattan_link_length = 1;
	}
	~ROUTERS_GRAPH() {}
	
	long nodes_count() { return _nodes_count; }
	long arcs_count() { return _arcs_count; }
	int max_node_weight() const { return _max_node_weight; }
	int max_manhattan_link_length() { return _max_manhattan_link_length; }
	vector<ROUTERS_NODE> &nodes() { return _nodes; }
	ROUTERS_NODE *get_node( long id) {
		assert(id >= 0 && id < _nodes_count);
		return &_nodes[ id]; 
	}
	ROUTERS_ARC *get_arc( long id) {
		//assert(id >= 0 && id < _arcs_count);
		return &_arcs[ id]; 
	}
	// creation done node by node; longer runtime due to mem alloc;
	int create_nodes( ONE_OF_THE_BEST_FLOOR_PLANS &best_fp);
	void create_nodes_and_arcs( ROUTERS_GRAPH *routers_graph,
		int threshold_dist); // version 1: populate and then remove;
	void create_nodes_and_arcs_by_growing( ROUTERS_GRAPH *routers_graph,
		int threshold_dist); // version 2: grow;
 
	void create_arcs( int max_dist, int min_dist = 0);
	void compute_node_weights();
	void compute_arc_weights();
	double compute_max_manhattan_link_length();

	void prune_some_arcs_old();
	void prune_some_arcs(); // used on the noc_graph only;
	bool prune_some_arcs_one_at_a_time();
	bool check_and_fix_full_connectivity();
	
	// debug;
	void print_routers_graph();
	void print_graph_statistics();
	// utils;
	int core_id_to_router_id( int core_id) {
		assert( core_id >= 0 && core_id < _nodes_count);	
		return _core_id_to_router_id[ core_id]; // router id hooked to this IP/core;
	}
	int node_exists_with_coords( int x, int y) {
		// search among all existing nodes and check if there is any
		// node with coords x,y; if not return -1;
		for ( int i = 0; i < _nodes_count; i++) {
			if ( (_nodes[i].x() == x && _nodes[i].y() == y) || // perfect overlap;
				( fabs(_nodes[i].x() - x) + fabs(_nodes[i].y() - y) <=
				  MINIMUM_DIST_BETWEEN_ROUTERS) ) { // Manhattan distance is too small;
				return ( i);
			}
		}
		return ( -1);
	}
	int index_in_adj_of_router_id( int this_r_id, int r_id) {
		// the reference is this_r_id, which together with router r_id 
		// form an arc, stored at location "k" in the adjacency list of 
		// this_r_id router; note that such arc can have either src or des 
		// registered as the this_r_id (based on construction order);
		for ( long k = 0; k < _nodes[this_r_id].adj().size(); k++) {
			int adj_arc_id = _nodes[this_r_id].adj()[k];
			if ( _arcs[adj_arc_id].src_id() == r_id ||
				 _arcs[adj_arc_id].des_id() == r_id) return k;
		}
		return -1; // error!
	}
	int dist_manhattan( int x1, int y1, int x2, int y2) {
		return ( abs(x1 - x2) + abs(y1 - y2));
	}
	int dist_euclidian( int x1, int y1, int x2, int y2) {
		return int( sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2)) );
	}
};

////////////////////////////////////////////////////////////////////////////////
//
// APPLICATION
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
//
// APPLICATION_NODE
//
////////////////////////////////////////////////////////////////////////////////

class APPLICATION_NODE {
 private:
	int _id;
	double _io_comm_volume; // sum of comm_volume of all arcs in or out;
	vector<long> _fanin; // id's of nodes with arcs as destination this node;
	vector<long> _fanout;	
		
 public:
	APPLICATION_NODE() { 
		_id = -1; 
		_io_comm_volume = 0.;
	}
	APPLICATION_NODE( int id) : _id(id) { 
		_io_comm_volume = 0.;
	}
	APPLICATION_NODE(const APPLICATION_NODE &node) :
		_id(node._id),
		_io_comm_volume(node._io_comm_volume) {}
	~APPLICATION_NODE() {}
	
	int id() const { return _id; }
	void set_io_comm_volume( double v) { _io_comm_volume = v; }
	double io_comm_volume() { return _io_comm_volume; }
	void add_to_io_comm_volume(double delta) { _io_comm_volume += delta; }
	vector<long> &fanin() { return _fanin; }
	vector<long> &fanout() { return _fanout; }
	void add_fanout( long id) { // id is index of core this one fanouts to;
		_fanout.push_back( id);
	}
	void add_fanin( long id) {
		_fanin.push_back( id);
	}
};

////////////////////////////////////////////////////////////////////////////////
//
// APPLICATION_ARC
//
////////////////////////////////////////////////////////////////////////////////

class APPLICATION_ARC {
 private:
	int _id;
	long _src_id; // id of core that is source of this arc;
	long _des_id; // id of core that is destination of this arc;
	double _comm_volume;
		
 public:
	APPLICATION_ARC( int id, int src, int des, double cv) : 
		_id(id), _src_id(src), _des_id(des), _comm_volume(cv) {
	}
	~APPLICATION_ARC() {}
	
	int id() const { return _id; }
	int src_id() const { return _src_id; }
	int des_id() const { return _des_id; }
	double comm_volume() { return _comm_volume; }
	void set_comm_volume( double comm_volume) 
		{ _comm_volume = comm_volume; }
};

////////////////////////////////////////////////////////////////////////////////
//
// APPLICATION_GRAPH
//
////////////////////////////////////////////////////////////////////////////////

class APPLICATION_GRAPH {
 private:
	long _nodes_count;
	long _arcs_count;
	vector<APPLICATION_NODE> _nodes; // IP/cores of the application;
	vector<APPLICATION_ARC> _arcs; // communication arcs of the application;
	double _max_nodes_comm_volume; // maximum communication volume among all nodes; 
	double _max_arcs_comm_volume;

 public:
	APPLICATION_GRAPH() { _nodes_count = 0; _arcs_count = 0; }
	~APPLICATION_GRAPH() {}
	
	long nodes_count() { return _nodes_count; }
	long arcs_count() { return _arcs_count; }
	double max_nodes_comm_volume() { return _max_nodes_comm_volume; }
	double max_arcs_comm_volume() { return _max_arcs_comm_volume; }
	vector<APPLICATION_NODE> &nodes() { return _nodes; }
	APPLICATION_NODE *get_node(long id) {
		//assert(id >= 0 && id < _nodes_count);	
		return &_nodes[ id]; 
	}
	APPLICATION_ARC *get_arc(long id) {
		return &_arcs[ id]; 
	}
	bool create_nodes_and_arcs( B_Tree *fp_p);
	
	void compute_max_nodes_comm_volume() {
		// this should be called once only;
		_max_nodes_comm_volume = 0.0;
		for ( long i = 0; i < _nodes_count; i++) {
			double node_comm_volume = 0.0;
			for ( long k = 0; k < _nodes[i].fanin().size(); k ++) {
				node_comm_volume += _arcs[ _nodes[i].fanin()[k] ].comm_volume();
			}
			for ( long k = 0; k < _nodes[i].fanout().size(); k ++) {
				node_comm_volume += _arcs[ _nodes[i].fanout()[k] ].comm_volume();
			}
			_nodes[i].set_io_comm_volume( node_comm_volume); // record it;
			if ( _max_nodes_comm_volume < _nodes[i].io_comm_volume()) {
				_max_nodes_comm_volume = _nodes[i].io_comm_volume();
			}
		}
	}
	void compute_max_arcs_comm_volume() {
		_max_arcs_comm_volume = 0.0;
		for ( long i = 0; i < _arcs_count; i++) {
			if ( _max_arcs_comm_volume < _arcs[i].comm_volume()) {
				_max_arcs_comm_volume = _arcs[i].comm_volume();
			}
		}
	}
	void print_application_graph();
};

////////////////////////////////////////////////////////////////////////////////
//
// HETEROGENEOUS_NOC
//
////////////////////////////////////////////////////////////////////////////////

class HETEROGENEOUS_NOC {
 private:
	// floorplaning annealing parameters;
	int _times; // used directly as argument when SA_Floorplan() is called;
	int _local; // used directly as argument when SA_Floorplan() is called;
	float _term_temp; // used directly as argument when SA_Floorplan() is called;
	float _avg_ratio;
	float _lambda;
	float _alpha;
	float _fp_scale;

	int _modules_N; // number of cores of the floorplan for a given testcase;
	int _ary; // quantity of routers in both dimensions; total number of routers;

	// big loop optimization (rather search and retain) parameters;
	char _inputfile[256];
	int _n_fps, _n_best;
	char _fp_criteria;

	// vNOC simulation parameters;
	long _inp_buf, _out_buf;
	long _vc_n, _flit_size;
	long _link_bw, _link_l, _pipeline_in_link;
	long _seed;
	ROUTING_ALGORITHM _routing_a;
	double _cycles, _warmup_cycles;
	bool _use_gui;
	int _x_ary;
	double _extra_links_timing_factor;
	int _gui_pauses;
	int _sim_mode;
	char _test_name[256];
	int _inj_load, _skipped_packets, _max_skip_counter, _skip_counter;

	// additional parameters for various sweeps and investigations;
	bool _load_sweep;
	bool _buffer_sweep;
	bool _testcase_creation;
	float _testcase_multiplier;
	bool _use_excel;

	// _sketch_latency is a temporary storage where any call of the vNOC
	// simulator will deposit the final result; needed here because the
	// current wrapper VNOC_APP of vNOC is used via its ctor, which will 
	// destroy vNOC object inside it; so, I need to pass the result;
	double _sketch_latency;
	double _sketch_packets_per_cycle; // average;

	// application graph; this is unique no matter how many floorplans
	// are investigated;
	APPLICATION_GRAPH _application_graph;

	// if _verbose is true then detailed framework run will be printed
	// by calling "print_*" functions; default is true;
	bool _verbose;
	int _routing_paths_calc;

	// runtime stuff;
	clock_t _start_clock, _end_clock;
	timeval _start_wall, _end_wall;

 public:

 public:
 HETEROGENEOUS_NOC() : _application_graph() {
		// default values;
		_times = 160;
		_local = 6;
		_avg_ratio = 30;
		_lambda = 1.3;
		_term_temp = 0.1; 
		_alpha = 0.25;
		_n_fps = 1; // default 10;
		_n_best = 1; // default 3;
		_fp_criteria = 'A';
		_inp_buf = INPUT_BUFFER_SIZE; // 5;
		_out_buf = OUTPUT_BUFFER_SIZE; // 5;
		_vc_n = VIRTUAL_CHANNEL_COUNT; // 3;
		_flit_size = FLIT_SIZE; // 1: flit is same a phit: 64 bits;
		_link_bw = LINK_BW; // 64;
		_pipeline_in_link = 0;
		_routing_a = ROUTING_LUT;
		_use_gui = false;
		_x_ary = 0;
		_extra_links_timing_factor = 1;
		_gui_pauses = 1;
		_fp_scale = 1.0;
		_sim_mode = SIMULATED_ARCH_2D;
		strcpy( _test_name, "Unnamed");
		_inj_load = 100; 
		_skipped_packets = 0; 
		_max_skip_counter = 10;
		_skip_counter = 0;
		_load_sweep = false;
		_buffer_sweep = false;
		_testcase_creation = false;
		_testcase_multiplier = 1.0;
		_use_excel = false;

		_modules_N = 0;
		_ary = 8; // default;
		_sketch_latency = 0.0;
		_sketch_packets_per_cycle = 0.0;
		// seed to be used for the C++' RNG is set randomly here; 
		// it can be overwritten via arguments by user;
		_seed = time(NULL);
		_verbose = true;
		_routing_paths_calc = ROUTING_PATHS_MCF;
	}
	~HETEROGENEOUS_NOC() {}

	bool parse_command_arguments( int argc, char *argv[]);
	void print_setup_info() {
		printf("Testcase name: \"%s\"", &_test_name);
		switch ( _sim_mode) {
		case SIMULATED_ARCH_2D:
			printf("\nWorking in 2D mode.");
			break;
		case SIMULATED_ARCH_3D:
			printf("\nWorking in 3D mode.");
			break;
		default:
			assert(0);
		}
		printf("\nSeed: %d", _seed);
	}
	void setup_seed() {
		// use the seed that potentially was asked by user via arguments;
		srand( _seed);
	}
	void start_chronometer() {
		_start_clock = clock();
		assert( _start_clock != (clock_t)(-1));
		gettimeofday( &_start_wall, 0);
	}
	void stop_chronometer() {
		_end_clock = clock();
		assert( _end_clock != (clock_t)(-1));
		gettimeofday( &_end_wall, 0);
	}
	bool search_n_fps_floorplans(
		vector <ONE_OF_THE_BEST_FLOOR_PLANS> &best_fps);
	bool routers_assignment_routing_calculation_simulation(
		vector <ONE_OF_THE_BEST_FLOOR_PLANS> &best_fps);
	bool do_routers_assignment( ROUTERS_GRAPH &routers_graph);
	
	bool calculate_final_results_statistics( int argc,char **argv,
		vector <ONE_OF_THE_BEST_FLOOR_PLANS> &best_fps);
	
	double extra_links_timing_factor() const { return _extra_links_timing_factor; };
	int sim_mode() const { return _sim_mode; };
	void set_skip_counter(int new_val) { _skip_counter = new_val; };
	int skip_counter() const { return _skip_counter; };
	int max_skip_counter() const { return _max_skip_counter; };
	bool testcase_creation() const { return _testcase_creation; };
	double testcase_multiplier() const { return _testcase_multiplier; };
	void set_sketch_latency(double val) { _sketch_latency = val; };
	double sketch_latency() const { return _sketch_latency; };
	void set_sketch_packets_per_cycle(double val) { _sketch_packets_per_cycle = val; };
	double sketch_packets_per_cycle() const { return _sketch_packets_per_cycle; };
	char *test_name() { return _test_name; };
	bool verbose() const { return _verbose; };
	int routing_paths_calc() const { return _routing_paths_calc; };
};

#endif
