#ifndef _MCF_H_
#define _MCF_H_

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <vector>
#include "hetero.h"

using namespace std;

#define MAX_DEGREE 40

////////////////////////////////////////////////////////////////////////////////
//
// MCF acts as a host to the graph and methods such as dijkstra and solver;
//
////////////////////////////////////////////////////////////////////////////////

class NODE
{
 private:

 public:
	int id; // start from 0
	int x, y; // location coordinates directly in um;
	int pre; // parent node in shortest path tree
	int pre_edge; // parent edge in shortest path tree
	double dist; // distance for shortest path algorithm
	int no_comm; // number of destinations for commodities starting from this node
	int *comms; // list of commodities starting from the node
	int no_edge; // number of edges incident to the node
	int *edges; // list of edges incident to the node
	int dij_visited; // flag for dijkstra algo
	int dij_updated; // second flag for dijkstra algo
	int min_visited; // flag for searching min c(e)
	// _preferred_path has a number of elements equal to the number of 
	// commodities; each entry stores the next, downstream, node index
	// of the preferred unique path of the flow of this commodity from 
	// src toward des; this is constructed during the randomized rounding;
	int *_preferred_path;
 public:
	NODE() {}
	~NODE() {}

};

class EDGE
{
 private:

 public:
	int id; // start from 0
	int src, dest; // source and destination node id
	double power;
	double latency;
	double length;
	double capacity; // c_e
	// accumulated total flow;
	double flow;
	// _flows has a number of elements equal to the number of
	// commodities with the index being the id of demand;
	double *_flows;

	// dual of capacity;
	double _Y_e;
	double _old_Y_e;

 public:
	EDGE() {}
	~EDGE() {}

	void set_flow_of_commodity( int id, double flow)
		{ _flows[ id] = flow; }
	void add_to_flow_of_commodity( int id, double val)
		{ _flows[ id] += val; }
	double flow_of_commodity( int id)
		{ return _flows[ id]; }
};

class CUT
{
 private:

 public:
	int id; // start from 0
	int no_edge; // number of edges this cut contains
	int *edges;
	double X; // dual variable to the cut
	double used_area; // used area
	double A; // total available area 
 public:
	CUT() {}
	~CUT() {}

};

class COMMODITY
{
 private:

 public:
	int id; // start from 0
	int src, dest; // source and destination 
	double demand;
	double left_demand;
 public:
	COMMODITY() {}
	~COMMODITY() {}

};

class PAIR_TWO
{
 private:
	int _id;
	double _value;
 public:
	PAIR_TWO() { _id = 0; _value = 0.0; }
	PAIR_TWO(int id, double value) : _id(id), _value(value) {}
	PAIR_TWO(const PAIR_TWO &pt) : 
		_id(pt._id), _value(pt._value) {}
	~PAIR_TWO() {}

	int id() const { return _id; }
	double value() const { return _value; }
	void set_id(int id) { _id = id; }
	void set_value(double value) { _value = value; }
	void operator=(const PAIR_TWO &pt) { _id = pt._id; _value = pt._value; }
	bool operator==(const PAIR_TWO &pt) const { return (_value == pt._value); }
	bool operator!=(const PAIR_TWO &pt) const { return (_value != pt._value); }
	bool operator<(const PAIR_TWO &pt)  const { return (_value  < pt._value); }
};

class PAIR_THREE
{
 private:
	int _id1, _id2;
	int _value;
 public:
	PAIR_THREE() { _id1 = 0; _id1 = 0; _value = 0; }
	PAIR_THREE(int id1, int id2, int value) : _id1(id1), _id2(id2), _value(value) {}
	PAIR_THREE(const PAIR_THREE &pt) : 
		_id1(pt._id1), _id2(pt._id2), _value(pt._value) {}
	~PAIR_THREE() {}

	int id1() const { return _id1; }
	int id2() const { return _id2; }
	int value() const { return _value; }
	void operator=(const PAIR_THREE &pt) { _id1 = pt._id1; _id2 = pt._id2; _value = pt._value; }
	bool operator==(const PAIR_THREE &pt) const { return (_value == pt._value); }
	bool operator!=(const PAIR_THREE &pt) const { return (_value != pt._value); }
	bool operator<(const PAIR_THREE &pt)  const { return (_value  < pt._value); }
};

class MCF
{
 private:
	int no_node;
	NODE *nodes;
	int no_edge;
	EDGE *edges;
	int no_commodity;
	vector<COMMODITY> _commodities;
	// this sketch array will be used to sort nodes by their maximum commodity;
	vector<PAIR_TWO> _sketch_nodes_max_commodity;

	// dual variables;
	double _phi_power; // not used;
	double _phi_latency;
	// primal variables;
	double B, LB, UB; // B is power PW; not used;
	double L, LL, UL; // L is latency LT


	// lambda_max is the first lambda that mcf() returns, with initial
	// latency_limit relaxed to 100000 (inf);
	double _lambda_max;
	double _latency_max; // associated with lambda_max;
	double _total_latency;
	double _power_max; // not used;
	double _total_power; // not used;


	// control variables;
	double _delta;
	double _epsilon1, _epsilon2, _epsilon3, _epsilon4;
	// s = [P(lambda_max) - B]/[lambda_max - dual]; see eq. 9;
	// used to implement "interval estimation";
	double _s;
	// temp_edge_flow stores how much flow is routed during an iteration;
	// it is a sketch array;
	double *_temp_edge_flow;
	int _rd;
	int _min_rd;
	// if _verbose is true then detailed debug info will be printed;
	bool _verbose;

 public:

 public:
	MCF( bool verbose) : _verbose(verbose) {
		no_node = 0;
		no_edge = 0;
		no_commodity = 0;

		_phi_latency = 0;
		_phi_power = 0;
		L=0; LL=0; UL=0;
		_lambda_max = 0;

		_latency_max = 0;
		_total_latency = 0;
		_power_max = 0; 
		_total_power = 0;

		_delta = 0;
		_epsilon1 = 0; _epsilon2 = 0; _epsilon3 = 0; _epsilon4 = 0;
		_s = 0;
		_min_rd = 0;
		_rd = 0;
		_temp_edge_flow = 0;
	}
	~MCF() {}


	// utils;
	double flow_of_commodity_thru_edge( int c_id, int e_id) {
		//assert(c_id >= 0 && c_id < no_commodity);
		//assert(e_id >= 0 && e_id < no_edge);
		return edges[ e_id].flow_of_commodity( c_id);
	}

	// host;
	void init(double delta);
	double getD();
	double getAlpha();
	void route_flow(int t, double amount, int commodity_id);
	double check_power_constraint(int t);
	double check_latency_constraint(int t);
	double min_cap( int s);
	void update_dual(int s, double epsilon);
	double minimum(double x, double y, double z);
	void scale_down_linear(float times);
	void scale_down_flows( int phase_count);

	// Dijkstra;
	void shortest_path(int s, int no_comm, int *dest_flag);
	void shortest_path_with_limited_bandwidth(
		int s, double t_commodity, int no_comm, int *dest_flag);

	// MCF solver;
	void init_param();
	double mcf(double delta, double epsilon1, double epsilon2, 
		double epsilon3, int flag);
	double mcf_solver();
	void populate_sketch_arrays_for_nodes_sorting();
	bool do_randomized_rounding();
	void calculate_average_path_length();
	bool do_shortest_paths_as_preferred_paths_with_inf_bandwidth();
	bool do_shortest_paths_as_preferred_paths_with_limited_bandwidth();
	
	// graph related utilities;
	bool build_network_from_file( char *network_filename,
		double latency_limit = 100000, double rate = 1);
	bool build_network_from_host_application(
		ROUTERS_GRAPH &noc_graph, APPLICATION_GRAPH &app_graph,
		double capacity_a);

	// more utiles;
	int get_preferred_path( int node_i, int comm_i) {
		// get downstream node id along the preffered path of
		// "comm_i" commodity, looking from "node_i";
		assert(comm_i >= 0 && comm_i < no_commodity);
		assert(node_i >= 0 && node_i < no_node);
		return ( nodes[node_i]._preferred_path[comm_i]);
	}
	int get_src_of_commodity( int comm_i) {
		return ( _commodities[comm_i].src);
	}
	int get_des_of_commodity( int comm_i) {
		return ( _commodities[comm_i].dest);
	}	
	int no_commodities() {
		return ( no_commodity);
	}

	// others;
	void read_network_topology_and_demands(FILE *fp, double rate = 1);
	void free_topology();
	void print_network_demands( bool print_only_edges = false);
	void print_backward_shortest_path( int t);
	void print_routing_paths();
	void print_max_accumulated_flow();
};

#endif
