#ifndef _VNOC_H_
#define _VNOC_H_
#include "config.h"
#include "vnoc_topology.h"
#include <assert.h>
#include <stdio.h>
#include <vector>
#include <utility>
#include <map>
#include <functional>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cstring>
#include "mcf.h"

using namespace std;


////////////////////////////////////////////////////////////////////////////////
//
// "vnoc" means versatile network on chip simulator;
//
////////////////////////////////////////////////////////////////////////////////

class EVENT;
class VNOC;
class ROUTER;
class EVENT_QUEUE;
class GUI_GRAPHICS;
class HETEROGENEOUS_NOC;

////////////////////////////////////////////////////////////////////////////////
//
// FLIT
//
////////////////////////////////////////////////////////////////////////////////

class FLIT {
	public:
		enum FLIT_TYPE { HEADER, BODY, TAIL };
	private:
		long _id;
		FLIT_TYPE _type;
		double _start_time;
		double _finish_time;
		// for hetero noc's we use router id's as addresses;
		int _src_router_id, _des_router_id;
		DATA _data; // vector<unsigned long long>;

	public:
		FLIT() : _id(0), _type(HEADER),
			_src_router_id(0), _des_router_id(0), 
			_start_time(0), _finish_time(0), _data() {}
		// heterogeneous NOC;
		FLIT( long id, FLIT_TYPE type, int src_router_id, int des_router_id,
			double start_time, const DATA &data) : 
			_id(0), _type(type), 
			_src_router_id(src_router_id), _des_router_id(des_router_id),
			_start_time(start_time), _finish_time(0), 
			_data(data) {}
 		FLIT( const FLIT &f) : _id(f.id()), _type(f.type()),
			_src_router_id(f.src_router_id()), _des_router_id(f.des_router_id()),
			_start_time(f.start_time()), _finish_time(f.finish_time()), 
			_data(f.data()) {}
		~FLIT() {}

		int id() const { return _id; }
		void set_id(int id) { _id = id; }
		FLIT_TYPE type() const { return _type; }
		double start_time() const { return _start_time; }
		double finish_time() const { return _finish_time; }
		int src_router_id() const { return _src_router_id; }
		int des_router_id() const { return _des_router_id; }
		DATA &data() { return _data; }
		const DATA &data() const { return _data; }
};

////////////////////////////////////////////////////////////////////////////////
//
// ROUTER_INPUT
//
////////////////////////////////////////////////////////////////////////////////

class ROUTER_INPUT {
	private:
		ROUTER *_router; // its owner;
		// input buffers: <physical port index<vc index<buffer>>>
		vector<vector<vector<FLIT> > > _input_buff;
		// state of each input vc;
		vector<vector<VC_STATE> > _vc_state; // INIT, ROUTING, VC_AB, SW_AB, SW_TR, HOME
		// candidate routing vcs; this stores the routing matrix/mapping
		// between inputs and outputs;
		vector<vector<vector<VC_TYPE> > > _routing; // pair<long, long>
		// the selected routing vc;
		vector<vector<VC_TYPE> > _selected_routing;
		// _vc_count stores the number of virtual channels of each port;
		// they may be different depending on how many flows go thru the
		// link comming to this input port (according to the routing tables);
		// this is in order to implement deadlock free routing;
		vector<long > _vc_count;
		// this is a flag record that the buffer of injection is full;
		// used to control the packet injection and decrease mem usage;
		bool _injection_buff_full;

	public:
		ROUTER_INPUT(long physical_ports_count, long vc_number);
		~ROUTER_INPUT() {}

	public:
		void initialize_input_buffers(long physical_ports_count);
		// _input_buff;
		vector<vector<vector<FLIT> > > &input_buff() { return _input_buff; }
		const vector<vector<vector<FLIT> > > &input_buff() const 
			{ return _input_buff; }
		vector<FLIT> &input_buff(long i, long j) { return _input_buff[i][j]; } 
		const vector<FLIT> &input_buff(long i, long j) const 
			{ return _input_buff[i][j]; }
		// # of vc for each input port;
		void incr_vc_count_of_port( long port_id) {
			_vc_count[port_id] ++;
		}
		long vc_count( long port_id) const { return _vc_count[ port_id]; }
		//_vc_state;
		vector<vector<VC_STATE> > &vc_state() { return _vc_state; }
		const vector<vector<VC_STATE> > &vc_state() const { return _vc_state; }
		VC_STATE vc_state(long i, long j) { return _vc_state[i][j]; }
		const VC_STATE vc_state(long i, long j) const { return _vc_state[i][j]; }
		void vc_state_update(long i, long j, VC_STATE state)
			{ _vc_state[i][j] = state;}
		// _routing;
		vector<vector<vector<VC_TYPE> > > &routing() 
			{ return _routing; }
		const vector<vector<vector<VC_TYPE> > > &routing() const 
			{ return _routing; }
		vector<VC_TYPE> &routing(long i, long j)
			{ return _routing[i][j]; }
		const vector<VC_TYPE> &routing(long i, long j) const 
			{ return _routing[i][j]; }
		void add_routing(long i, long j, VC_TYPE t)
			{ _routing[i][j].push_back(t); }
		void clear_routing(long i, long j)
			{ _routing[i][j].clear(); }
		// _selected_routing;
		vector<vector<VC_TYPE> > &selected_routing() { return _selected_routing; }
		const vector<vector<VC_TYPE> > & selected_routing() const 
			{ return _selected_routing; }
		VC_TYPE &selected_routing(long i, long j) { return _selected_routing[i][j]; }
		const VC_TYPE &selected_routing(long i, long j) const 
			{ return _selected_routing[i][j]; }
		void assign_selected_routing(long i, long j, VC_TYPE c) 
			{ _selected_routing[i][j] = c; }
		void clear_selected_routing(long i, long j) 
			{ _selected_routing[i][j] = VC_NULL; }

		// utils;
		void add_flit(long i, long j, const FLIT &flit) 
			{ _input_buff[i][j].push_back(flit); }
		void remove_flit(long i, long j)
			{ _input_buff[i][j].erase(_input_buff[i][j].begin()); }
		FLIT &get_flit(long i, long j) {
			assert( _input_buff[i][j].size() > 0);
			return ( _input_buff[i][j][0]); 
		}
		FLIT &get_flit(long i, long j, long k) {
			assert( _input_buff[i][j].size() > k);
			return ( _input_buff[i][j][k]); 
		}

		void set_injection_buff_full() { _injection_buff_full = true; }
		void clear_injection_buff_full() { _injection_buff_full = false; }
		bool injection_buff_full() const {return _injection_buff_full; }
};

////////////////////////////////////////////////////////////////////////////////
//
// ROUTER_OUTPUT
//
////////////////////////////////////////////////////////////////////////////////

class ROUTER_OUTPUT {
	public:
		enum VC_USAGE { USED, FREE };
	private:
		ROUTER *_router; // its owner;
		// used for input of next router;
		long _buffer_size_next_r;
		vector<vector<long> > _counter_next_r;
		vector<vector<VC_STATE> > _flit_state; // INIT, ROUTING, VC_AB, SW_AB, SW_TR, HOME;
		// assigned for the input
		vector<vector<VC_TYPE> > _assigned_to; // pair<long, long>
		vector<vector<VC_USAGE> > _vc_usage; // USED FREE
		vector<vector<FLIT> > _out_buffer; // actual local output buffers;
		vector<vector<VC_TYPE> > _out_addr; // output address;
		vector<long> _local_counter; // one for each output port;
		// _vc_count stores the number of virtual channels of each port;
		// they may be different depending on how many flows go thru the
		// link going out from this output port (according to the routing tables);
		// this is in order to implement deadlock free routing;
		vector<long > _vc_count;

	public:
		ROUTER_OUTPUT(long physical_ports_count, long vc_number,
			long input_buffer_size, long output_buffer_size);
		~ROUTER_OUTPUT() {}

	public:
		void initialize_output_buffers(long physical_ports_count);
		long buffer_size_next_r() const { return _buffer_size_next_r; }
		vector<vector<long> > &counter_next_r() { return _counter_next_r; }
		long counter_next_r(long i, long j) const { return _counter_next_r[i][j]; }
		void counter_next_r_inc(long i, long j) { _counter_next_r[i][j]++; }
		void counter_next_r_dec(long i, long j) { _counter_next_r[i][j]--; }

		VC_STATE flit_state(long i) const { return _flit_state[i][0]; }

		const VC_TYPE &assigned_to(long i, long j) const 
			{ return _assigned_to[i][j]; }
		void acquire(long i, long j, VC_TYPE pair) 
			{ _vc_usage[i][j] = USED; _assigned_to[i][j] = pair; }
		void release(long i, long j) 
			{ _vc_usage[i][j] = FREE; _assigned_to[i][j] = VC_NULL; }
		vector<vector<VC_USAGE> > &vc_usage() { return _vc_usage; }
		VC_USAGE vc_usage(long i, long j) { return _vc_usage[i][j]; }
		VC_USAGE vc_usage(long i, long j) const { return _vc_usage[i][j]; }
		
		void add_flit(long i, const FLIT &flit)
			{ _out_buffer[i].push_back(flit); _local_counter[i]--; }
		void remove_flit(long i);
		FLIT &get_flit(long i) {
			assert( _out_buffer[i].size() > 0);
		   	return _out_buffer[i][0];
		}
		vector<FLIT> &out_buffer(long i) { return _out_buffer[i]; }
		const vector<FLIT> &out_buffer(long i) const 
			{ return _out_buffer[i]; }
		// # of vc for each output port;
		void incr_vc_count_of_port( long port_id) {
			_vc_count[port_id] ++;
		}
		long vc_count( long port_id) const { return _vc_count[ port_id]; }

		vector<vector<VC_TYPE> > out_addr() { return _out_addr; }
		const vector<vector<VC_TYPE> > out_addr() const { return _out_addr; }
		vector<VC_TYPE> out_addr(long i) { return _out_addr[i];}
		const vector<VC_TYPE> out_addr(long i) const { return _out_addr[i]; }
		VC_TYPE get_addr(long i) { return _out_addr[i][0];}
		void remove_addr(long i) { 
			assert( _out_addr[i].size() > 0); 
			_out_addr[i].erase( _out_addr[i].begin());
		}
		void add_addr(long i, VC_TYPE b) { _out_addr[i].push_back(b); }

		vector<long> &local_counter() { return _local_counter; }
		long local_counter(long i) { return _local_counter[i]; }
		long local_counter(long i) const { return _local_counter[i]; }
		void local_counter_inc(long i) { _local_counter[i]++; }
		void local_counter_dec(long i) { _local_counter[i]--; }
};

////////////////////////////////////////////////////////////////////////////////
//
// ROUTER
//
////////////////////////////////////////////////////////////////////////////////

class ROUTER {
	private:
		VNOC *_vnoc; // its owner, which has EVENT_QUEUE *_event_queue as well;
		//ADDRESS _address; // vector<long>;
		long _id;
		ROUTER_INPUT _input; // input buffer module;
		ROUTER_OUTPUT _output; // output buffer module;
		// initial random number
		DATA _init_data;
		long _flit_size; // how many groups of "64 bits";
		long _physical_ports_count;
		// _vc_number is only for local port of local PE; the other ports
		// have their vc_number stored in _vc_count vectors of _input and
		// _output of each router;
		// Note: actually not used anymore; see ctor of VNOC for more info;
		long _vc_number;
		long _buffer_size; // of each virtual channel;
		long _out_buffer_size; // of each vc out buffer size;
		// accumulated total "propagation" delay of all flits with destination
		// (ie consumed) by this router;
		double _total_delay;
		ROUTING_ALGORITHM _routing_algo; // routing algorithm used;
		double _local_injection_time; // used for the next packet injection time;
		long _packet_counter;
		ifstream *_local_injection_file; // input trace file;
		// for hetero NOC's we need to keep track of the connectivity between
		// routers thru their port id's;
		vector<int> _neighbor_router_id_of_this_port;
		vector<int> _neighbor_port_of_this_port;
		// routing LUT is a "_routers_count x _routers_count" matrix that stores
		// at each router locally info about routing paths of flows between any
		// path src->des required by the application task graph;
		// TODO: use a map data structure to save memory for known paths;
		vector<vector<long> > _routing_LUT;

	public:
		int _assigned_core;

	public:
		ROUTER( long physical_ports_count, long vc_number, long buffer_size, 
				long out_buffer_size, 
				long flit_size, VNOC *owner_vnoc, long id, long assigned_core);
		~ROUTER() {}

		void set_id(long id) { _id = id; }
		long id() { return _id; }
		void set_owner(VNOC *vnoc) { _vnoc = vnoc; }
		ROUTER_INPUT &input() { return _input; }
		ROUTER_OUTPUT &output() { return _output; }	
		long physical_ports_count() { return _physical_ports_count; }
		long vc_number() { return _vc_number; }
		long buffer_size() { return _buffer_size; }

		// calculate the  accumulated total "propagation" delay of all flits;	
		void update_delay(double delta) { _total_delay += delta; }
		double total_delay() const { return _total_delay; }
		// retrieve packet from trace file associated with this;
		void receive_packet_from_trace_file();
		void inject_packet( long flit_id, int src_router_id, int des_router_id,
			double time, long packet_size);
		// flit and credit utils;
		void receive_flit_from_upstream(long port_id, long vc_id, FLIT &flit);
		void consume_flit(double t, const FLIT &flit);
		void receive_credit(long i, long j);

		// run simulation of this router;
		void simulate_one_router();
		void routing_decision_stage();
		VC_TYPE vc_selection(long i, long j);
		void vc_arbitration_stage();
		void sw_arbitration_stage();
		void send_flit_to_out_buffer();
		void send_flit_via_physical_link(long i);
		void send_flits_via_physical_link();
		void sanity_check() const;
		// more utils;
		void set_neighbor_router_id_of_this_port( int port_id, int r_id) {
			//assert( port_id >= 0 &&  port_id < _physical_ports_count);
			//assert( r_id >= 0 &&  r_id < _vnoc->routers_count());
			_neighbor_router_id_of_this_port[port_id] = r_id; 
		}
		void set_neighbor_port_of_this_port( int port_id, int neighbor_port_id) {
			_neighbor_port_of_this_port[port_id] = neighbor_port_id; 
		}
		void set_routing_LUT( int src_id, int des_id, long val) {
			_routing_LUT[src_id][des_id] = val;
		}
		// bookkeeping of # of virtual channels as the number of 
		// flows going thru links;
		void incr_flow_count_of_input_port( long port_id) {
			assert( port_id >= 0 && port_id < _physical_ports_count);
			_input.incr_vc_count_of_port( port_id);
		}
		long flow_count_of_input_port( long port_id) {
			return _input.vc_count( port_id);
		}
		void incr_flow_count_of_output_port( long port_id) {
			assert( port_id >= 0 && port_id < _physical_ports_count);
			_output.incr_vc_count_of_port( port_id);
		}
		long flow_count_of_output_port( long port_id) {
			return _output.vc_count( port_id);
		}

		int get_port_id_of_adj_router( int next_r_id) {
			// what is the port id inside this router that is connected
			// to the neighboring next_r_id?
			for ( int k = 0; k < _physical_ports_count; k++) {
				if ( _neighbor_router_id_of_this_port[k] == next_r_id)
					return k;
			}
			assert(0);
			return -1; // error!
		}

		void print_router_stats();

		ifstream &local_injection_file() { return *_local_injection_file; }
		void init_local_injection_file();
		void close_injection_file() { // needed to avoid memory leaks;
			_local_injection_file->close();
			delete _local_injection_file;
		}

		void call_current_routing_algorithm(
			const int des_r_id, const int src_r_id,
			long s_ph, long s_vc);
};

////////////////////////////////////////////////////////////////////////////////
//
// LINK
//
////////////////////////////////////////////////////////////////////////////////

class LINK {
 private:
	int _id;
	int _x;
	int _y;
	
 public:
 	LINK( int x, int y, int z) : _x(x), _y(y), _id(-1) {}
	~LINK() {}
	
	int id() const { return _id; }
	void set_id(int id) { _id = id; }
};

////////////////////////////////////////////////////////////////////////////////
//
// VNOC: THE VERSATILE NOC SIMULATOR
//
////////////////////////////////////////////////////////////////////////////////

class VNOC {
	private:
		long _routers_count;
		long _packets_count;
		ifstream _input_file_st;
		// _latency is the final result of running the vNOC simulator; introduced
		// in order to be able to read it from other classes, when vNOC is used 
		// multiple times;
		double _latency;
		// if _verbose is true then detailed debug info will be printed;
		bool _verbose;

		// here are the delay numbers extrapolated from the original
		// numbers from popnet that were derived for a link of 1mm and
		// technology of 100nm;
		// here, in the heterogeneous NOC I compute the T clock by taking
		// the maximum Manhattan distance between routers connected to form 
		// the topology and extrapolate the delay numbers using the Elmore 
		// delay expression;
		// Note: it's possible for the same testcase, for different floorplans
		// to get topologies that will have different T clock because the 
		// topologies will be different;
		// these values will be computed by calling compute_heterogeneous_t_clock(),
		// after each topology construction and before the vnoc simulator 
		// will be called;
		double _h_wire_delay;
		double _h_pipe_delay;
		double _h_credit_delay;
		
	public:
		// _routers was made public to be accessed by the gui;
		vector<ROUTER> _routers;
		TOPOLOGY *_topology;
		EVENT_QUEUE *_event_queue;
		ROUTERS_GRAPH *_noc_graph;
		GUI_GRAPHICS *_gui;
		HETEROGENEOUS_NOC *_hetero;

	public:
		VNOC( TOPOLOGY *topology, EVENT_QUEUE *event_queue,
			  ROUTERS_GRAPH *noc_graph, MCF *mcf, bool verbose=true);
		~VNOC() {
			// all files must be closed, or memory leaks can occur;
			_input_file_st.close();
			for ( int i = 0; i < _routers.size(); i++) {
				_routers[i].close_injection_file();
			}
		}

		TOPOLOGY *topology() const { return _topology; }
		EVENT_QUEUE *event_queue() { return _event_queue; }
		ROUTERS_GRAPH *noc_graph() { return _noc_graph; }
		GUI_GRAPHICS *gui() { return _gui; };
		void set_gui(GUI_GRAPHICS *gui) { _gui = gui; }
		HETEROGENEOUS_NOC *hetero() { return _hetero; }
		void set_hetero_host( HETEROGENEOUS_NOC *hetero) { _hetero = hetero; }

		vector<ROUTER> &routers() { return _routers; }
		const vector<ROUTER> &routers() const { return _routers; }
		ROUTER *router( int id) {
			assert( id >= 0 && id < _routers_count);
			return ( &_routers[ id]);
		}
		long routers_count() { return _routers_count; }
		long packets_count() { return _packets_count; }
		long packets_count() const { return _packets_count; }
		void packets_count_incr() { _packets_count ++;}	
		double latency() const { return _latency; }
		bool verbose() const { return _verbose; }

		bool receive_EVENT_PE();
		bool receive_EVENT_ROUTER( EVENT this_event);
		bool receive_EVENT_LINK( EVENT this_event);
		bool receive_EVENT_CREDIT( EVENT this_event);

		bool run_simulation();
		void check_simulation();
		void print_simulation_results();
		void print_network_routers();
		void init_input_file();

		// delay calculations;
		double Elmore_delay(double base_delay, double base_length, double actual_length);
		double Linear_delay( double base_delay, double base_length, double actual_length);
		double Via_delay(double base_delay, double layers);
		void compute_heterogeneous_t_clock( double max_link_length,
			bool square_proportionality = false); // linear not square by default;
		double h_wire_delay() { return _h_wire_delay; }
		double h_pipe_delay() { return _h_pipe_delay; }
		double h_credit_delay() { return _h_credit_delay; }
};

#endif

