#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <math.h>
#include <assert.h>
#include "mcf_pqueue.h"
#include "mcf.h"

using namespace std;

double BEST_SO_FAR = -1; // best of all simulated topologies;


////////////////////////////////////////////////////////////////////////////////
//
// MCF host
//
////////////////////////////////////////////////////////////////////////////////

void MCF::init(double delta)
{
	int i;

	//init dual variables
	_phi_latency = delta / L; // dual variable PHI_d = 1/100000
	for ( i = 0; i < no_edge; i++) {
		edges[i]._Y_e = delta / edges[i].capacity;
		edges[i]._old_Y_e = edges[i]._Y_e;
	}	

	//init edges
	for ( i = 0; i < no_edge; i++) {
		edges[i].flow = 0.0;
		for (int j = 0; j < no_commodity; j++) {
			edges[i]._flows[ j] = 0.0;
		}
	}
	//reset edge flows
	for ( i = 0; i < no_edge; i++) {
		_temp_edge_flow[i] = 0;
	}

	//init commodities
	for ( i = 0; i < no_commodity; i++) {
		_commodities[i].left_demand = _commodities[i].demand;
	}
		
	//init edge length
	for ( i = 0; i < no_edge; i++) {
		edges[i].length = 0;
		edges[i].length += edges[i]._Y_e;
		edges[i].length += edges[i].latency * _phi_latency;
	}

	// inits
	_total_latency = 0.0;

	return;
}

double MCF::getD()
{
	// this computes "D" numerator of eq. 7 or line 5 of algo 2;
	double D = 0.0;
	for ( int i = 0; i < no_edge; i++) {
		D += edges[i]._Y_e * edges[i].capacity;
		// Note1: "length" function is l + b*phi in Karakostas' paper;
		//D += edges[i].latency * _phi_latency;
	}
	// Note2: Garg's paper defines though D using directly the budget L;
	D += L * _phi_latency; // less accurate?
	return D;
}

double MCF::getAlpha()
{
	// alpha is the denuminator of dual = D/alpha (den of eq. 7);
	int i, j;
	double alpha = 0.0; // to return;
	for ( i = 0; i < no_node; i++) {
		if ( nodes[i].no_comm) {
			int *dest_flag = (int*)malloc(no_node*sizeof(int));
			if ( dest_flag == NULL) {
				printf("\nError: Unable to malloc <getAlpha>.\n"); exit (1);
			}
			memset((void*)dest_flag,0,no_node*sizeof(int));
				
			for (j = 0; j < nodes[i].no_comm; j++) {
				dest_flag[_commodities[nodes[i].comms[j]].dest] = 1;
			}
			
			shortest_path( nodes[i].id, nodes[i].no_comm, dest_flag);
			_rd++;
			free( dest_flag);

			for (j = 0; j < nodes[i].no_comm; j++) {
				alpha += _commodities[nodes[i].comms[j]].demand * 
					nodes[_commodities[nodes[i].comms[j]].dest].dist;
			}
		}
	}

	return alpha;
}

double MCF::check_power_constraint(int t) 
{
	double amount = 0.0;
	while (nodes[t].pre != -1) {
		amount += edges[nodes[t].pre_edge].power;
		t = nodes[t].pre;
	}
	return B/amount;
}

double MCF::check_latency_constraint(int t)
{
	double amount = 0.0;
	while (nodes[t].pre != -1) {
		amount += edges[nodes[t].pre_edge].latency;
		t = nodes[t].pre;
	}
	//printf("\n    check_latency: L=%.2f amnt=%.2f usable_amount=L/amnt=%.2f",L,amount,L/amount);
	// see Fleischer paper, section 4;
	return L/amount;
}

double MCF::min_cap( int s) 
{
	// find "c" as the minimum capacity of the edges on ALL
	// the paths in the shortest paths tree for this source node "s";
	int t = 0;
	double min_capacity = 1000000.0;

	_min_rd++;

	// start from all dest nodes, traverse shortest path tree;
	for ( int i = 0; i < nodes[s].no_comm; i++) {
		if ( _commodities[nodes[s].comms[i]].left_demand > 1e-6) {
			// pick up this destination and walk backward to sourse "s";
			t = _commodities[nodes[s].comms[i]].dest;
			while ((nodes[t].pre != -1) && (nodes[t].min_visited != _min_rd)) {
				int edge_index = nodes[t].pre_edge;
				nodes[t].min_visited = _min_rd;

				if ( edges[edge_index].capacity < min_capacity) {
					min_capacity = edges[edge_index].capacity;
				}
			}
		}
	}

	return min_capacity;
}

void MCF::route_flow( int t, double amount, int commodity_id)
{
	// t is destination to which we route "amount" of commodity;
	while ( nodes[t].pre != -1) {
		_temp_edge_flow[nodes[t].pre_edge] += amount;

		// record this amount for this commodity id on the 
		// corresponding edge also;
		assert(commodity_id >= 0 && commodity_id < no_commodity);
		edges[ nodes[t].pre_edge]._flows[ commodity_id] += amount;

		t = nodes[t].pre;
	}
	return;
}

void MCF::update_dual( int s, double epsilon)
{
	// update dual variables; compute l_i_j_s(e) in Karakostas' paper, where
	// "j" is jth iteration of phase "i", and "s" is the the current step;
	int i;
	int t;
	//double old_phi_power;
	//double temp_power = 0.0;
	double old_phi_latency;
	double temp_latency = 0.0;



	// () accumulate temp_latency along the shortest paths for the
	// shortest paths tree for the commodities of this source node;
	_min_rd++;
	for ( i = 0; i < nodes[s].no_comm; i++) {
		t = _commodities[nodes[s].comms[i]].dest;
		while ((nodes[t].pre != -1) && (nodes[t].min_visited != _min_rd)) {
			int edge_index = nodes[t].pre_edge;
			nodes[t].min_visited = _min_rd;
			
			//temp_power += _temp_edge_flow[edge_index] * edges[edge_index].power;
			temp_latency += _temp_edge_flow[edge_index] * edges[edge_index].latency;

			// update the dual variable Y_e;
			edges[edge_index]._old_Y_e = edges[edge_index]._Y_e;
			edges[edge_index]._Y_e *= 
				1 + epsilon * (_temp_edge_flow[edge_index]) / edges[edge_index].capacity;

			// walk upstream on shortest path;
			t = nodes[t].pre;
		}
	}
	_min_rd++;



	// () update additional dual variable PHI_d;
	//old_phi_power = _phi_power;
	//_phi_power *= 1 + epsilon * temp_power / B;
	//_total_power += temp_power;
	old_phi_latency = _phi_latency;
	_phi_latency *= 1 + epsilon * temp_latency / L; // adjust value from prev. iter;
	_total_latency += temp_latency;



	// () update the "length" function;
	for ( i = 0; i < no_edge; i++) {
		edges[i].length += 
			(edges[i]._Y_e - edges[i]._old_Y_e) +
			//edges[i].power * (_phi_power - old_phi_power) + 
			edges[i].latency * (_phi_latency - old_phi_latency);
	}



	// () add to the flow recorded for each edge the accumulated 
	// amounts (as sum of f_{i,j,s}^{c_q}) for each commodity routed during
	// this iteration, amounts which are reflected by _temp_edge_flow (which
	// has values != zero) for edges of shortest path of this iter;
	for (i = 0; i < no_edge; i++) {
		edges[i].flow += _temp_edge_flow[i];
	}


	// () reset temp storage of pushed flow during this iter; prepare it
	// for the next push/iteration;
	for (i = 0; i < no_edge; i++) {
		_temp_edge_flow[i] = 0.0;
	}


	return;
}

void MCF::scale_down_linear(float times) 
{
	for (int i = 0; i < no_edge; i++) {
		edges[i].length /= times;
		edges[i]._Y_e /= times;
	}
	//_phi_power /= times;
	_phi_latency /= times;
	return;
}

void MCF::scale_down_flows( int phase_count)
{
	// scale down final solution; see Garg's and Karakostas' papers;
	//double base = 1 + _epsilon3;
	//double factor = (1 + _epsilon3 * no_commodity) / _delta;
	//double scale = log( factor) / log( base);

	int scale = max( 1, phase_count); // this is "t";
	for ( int i = 0; i < no_edge; i ++) {
		edges[i].flow /= scale;
		for (int j = 0; j < no_commodity; j ++) {
			edges[i]._flows[ j] /= scale;
		}
	}
}

double MCF::minimum(double x, double y, double z) 
{
	double min;
	if (x < y) {
		if (x < z) min = x;
		else min = z;
	} else {
		if (y < z) min = y;
		else min = z;
	}
	return min;
}

double MCF::mcf( double delta,
	double epsilon1, double epsilon2, double epsilon3,
	int flag)
{
	// flag:
	// 0 -- concurrent flow;
	// 1 -- min-cost max concurrent; 
	// 2 -- min-cost bounded by 1;
	double lambda; // result to be returned;
	double alpha;
	double cumulated_gap=0.0, gap=0.0, old_gap=0.0, old_old_gap=0.0;
	int iter=0, seg_iter=0;
	// maximum factor is "total_latency/L" or "total_power/B"
	double max_factor = 0.0;
	double D;
	double usable_amount_latency, usable_amount_cap, routed_amount;
	struct timeval tv_start, tv_end;
	time_t duration;


	// () initialization of PHI_p, PHI_d, length(e)
	init( delta); // 1.0, 0
	_rd = 1;
	for ( int i = 0; i < no_node; i++) {
		nodes[i].dij_visited = 0;
		nodes[i].dij_updated = 0;
		nodes[i].min_visited = 0;
	}
	

	// () the big loop, each run of this loop is a phase;
	gettimeofday(&tv_start, (struct timezone*) NULL);
	while ( 1) {
		//printf("\nPHASE %d", iter); // iter counts phases;

		// () in every phase we start with the demand d_j for every commodity;
		// see Karakostas' paper;
		for ( int j = 0; j < no_commodity; j++) {
			// Note: "- 2*iter' is a technique to bias the rounding process
			// at the end of the mcf_solver; normally a commodity of say demand
			// 500 would be routed as 250 along the shortest path and 250 along 
			// another path; now with this biasing it will be routed as 250 and
			// 249; then during the rouding the path with 250 will be preferred
			// to achieve an integral flow solution;
			_commodities[j].left_demand = _commodities[j].demand - 2*iter;
		}

	

		// () next there are |S| iterations, one for each node that is a 
		// source for at least a commodity;
		//for ( int i = 0; i < no_node; i++) {
		for ( int k = no_node - 1; k >= 0; k--) {
			int i = _sketch_nodes_max_commodity[k].id();
		
			if ( nodes[i].no_comm) { // if this node is source of "r" _commodities;

				int commodities_left = nodes[i].no_comm;
				int *dest_flag = (int*)malloc(no_node*sizeof(int));
				if ( dest_flag == NULL) {
					printf("\nError: Unable to malloc <mcf>.\n"); exit (1);
				}
				memset((void*)dest_flag,0,no_node*sizeof(int));
				// dest_flag is set "1" for nodes that are destinations of _commodities;
				for ( int j = 0; j < nodes[i].no_comm; j++) {
					dest_flag[_commodities[nodes[i].comms[j]].dest] = 1;
				}


				// while there are left commodities to be routed (ie left) 
				// for this node; there are a number of steps for current
				// iteration;
				int step_count = 0;
				while ( commodities_left) {
					//printf("\nSTEP %d", step_count);
					step_count ++;

					// (a) compute shortest PATHS tree, where edges have "length(e)";
					// of all paths from this sink to all its destinations;
					shortest_path( nodes[i].id, commodities_left, dest_flag);
					

					// (b) available flow amount from bottleneck of shortest path tree;
					// this "c" represents the minimum capacity of the edges on ALL
					// the paths in the shortest paths tree;
					usable_amount_cap = min_cap( nodes[i].id);

					// (c) route "f = d(c_q)" units of flow of a given commodity and 
					// update the flow of each edge: f_e = f_e + f, along its shortest path;
					bool flow_has_been_routed = false;
					for ( int j = 0; j < nodes[i].no_comm; j++) {

						// the amount of commodity c_q that has not been routed yet
						//  at step "s";
						double left_demand = _commodities[nodes[i].comms[j]].left_demand;

						if ( left_demand > 1e-6) {
							flow_has_been_routed = true;
							// debug;
							//print_backward_shortest_path( _commodities[nodes[i].comms[j]].dest);

							// available flow amount from power,latency constraint
							// flag = 0 OR 1,2
							usable_amount_latency = 
								check_latency_constraint(_commodities[nodes[i].comms[j]].dest);
							//printf("\n    usable_amount_cap = %.2f",usable_amount_cap);
							//printf("    left_demand = %.2f",left_demand);
							//printf("    usable_amount_latency = %.2f",usable_amount_latency);

							// flow amount to be routed at step "s": f_{i,j,s}^{c_q};
							routed_amount = minimum( usable_amount_cap, left_demand, 
								usable_amount_latency);
							//routed_amount = fmin( left_demand, usable_amount_latency);
							//printf("    MIN:routed_amount = %.2f",routed_amount);

							// update every "_temp_edge_flow" - from dest backward to src
							// will be added routed_amount;
							route_flow( _commodities[nodes[i].comms[j]].dest,
										routed_amount, nodes[i].comms[j]);

							// update commodity amounts to be routed
							_commodities[nodes[i].comms[j]].left_demand -= routed_amount;

							if ( _commodities[nodes[i].comms[j]].left_demand <= 1e-6) {
								// this commodity is done, set its destination flag to 0;
								commodities_left --;
								dest_flag[_commodities[nodes[i].comms[j]].dest] = 0;
							}
						}
					}

					// (d) update dual variables;
					// including X, phi_power, phi_latency, and length; also update/increase
					// edges[].flow with the amount routed during this iter;
					update_dual( nodes[i].id, epsilon3);

					_rd ++;
					if ( !flow_has_been_routed) break;
				}//while ( comm_left)


				free(dest_flag);

			}//if ( nodes[i].no_comm)...
		}//for ( i = 0; i < no_node; i++)


	
		// () empirical stopping criterion: if this main loop takes too long 
		// than exit; this is subject to change, depends on application;
		gettimeofday(&tv_end, (struct timezone*) NULL);
		duration = tv_end.tv_sec - tv_start.tv_sec;
		if ( duration > 120) { return -1.0; }

		// () scale down if necessary;
		iter++;
		seg_iter++;
		D = getD();
		if ( D > 1.0e300) {
			scale_down_linear( 1.0e300);
			D /= 1.0e300;
		}
		alpha = getAlpha();


		// () calculate max_factor;
		max_factor = 0;
		//scale factor on latency constraint
		if (max_factor < _total_latency/L) {
			max_factor = _total_latency/L;
		}


		// () keep track of gap;
		old_old_gap = old_gap;
		old_gap = gap;
		lambda = (double) iter/max_factor;
		//printf("\n    > iter: %d / max_factor: %.3lf lambda: %.3f",iter,max_factor,lambda);
		double dual = D/alpha;
		gap = dual/lambda - 1; // if *100 would represent percentage;

		// () ce este asta aici ???
		// Note2: with the correction from Note1 above, this does not have
		// an impact on the final flows solution;
		if ( seg_iter != 1) { cumulated_gap += old_gap - gap; }
		if ( (seg_iter >= 8) && 
			(cumulated_gap/(seg_iter - 1) > (old_old_gap-gap) )) {
			epsilon3 = 0.618 * epsilon3;
			if ( epsilon3 < epsilon1) epsilon3 = epsilon1;
			seg_iter = 0;
			cumulated_gap = 0; // re-start accumulating gap;
		}
		

		 
		// () this implements the "interval estimation"; see Theorem 3 of
		// paper of Hu and Zhu; it also check for the termination condition
		// which is lambda>1 (or dual<1) when flag=2;
		if ( flag == 2) {
			// 2 -- min-cost bounded by 1;
			double UL1 = UL, LL1 = LL;
			double d = dual;
			if ( d < 1) d = 1;
			double s1 = (_latency_max - L)/( _lambda_max - d);
			if ( s1 > 0 && ( _s < 0 || _s > s1)) { 
				_s = s1; 
			}
			if ( _s > 0) {
				if ( lambda < 1) {
					double delta1 = 1 - lambda;
					UL1 = L + delta1 * _s;
					if ( UL1 < UL) UL = UL1;
				}
				if ( dual > 1) {
					double delta2 = dual - 1;
					LL1 = L - delta2 * _s;
					if ( LL1 > LL) LL = LL1;
				}
			}
			//printf("   dual: D/alpha=%.2lf/%.2lf=%.2f",D,alpha,dual);			
			//printf("   LL=%.2lf UL=%.2lf",LL,UL);
			if ( lambda > 1) { UL = L; break; }
			if ( dual < 1)   { LL = L; break; }
			if ( (UL-LL < 0) || (UL/LL - 1) < epsilon2) { break; }
		} else {
			// 0 -- concurrent flow;
			// 1 -- min-cost max concurrent; 
			if ( dual / lambda - 1 < epsilon1) { // gap between dual and primal is less 1%
				_total_latency /= max_factor;
				break;
			}
		}
		
	}//while (1)...
	


	// () scale down the final flows so that the solution is feasible 
	// (capacities are met);
	scale_down_flows( iter);
	

	lambda = (double) iter/max_factor;
	printf("\nlambda = %lf, dual = %lf, [%lf, %lf], L=%lf, iter_count=%d",lambda,D/alpha,LL,UL,L,iter);
	return lambda;
}


////////////////////////////////////////////////////////////////////////////////
//
// MCF actual solver
//
////////////////////////////////////////////////////////////////////////////////

void MCF::init_param()
{
	double w = 0.3; // 0.3
	B = 1000; // 1000
	L = 1e16;
	_epsilon1 = 0.1; // 0.1
	_epsilon2 = 0.01; // accuracy of binary search 0.01
	_epsilon3 = 0.6; // internal accuracy 0.6
	_epsilon4 = 1 - 1 / pow(w+1,(double)1/3); // accuracy of MCF solver;
	_delta = 1.0; // 1.0
}

double MCF::mcf_solver()
{
	// the actual MCF solver;
	// it is assumed that the network was already created from files
	// or host application;


	// () runtime;
	clock_t start_clock, end_clock;
	clock_t diff_clock;
	start_clock = clock();
	assert(start_clock != (clock_t)(-1));


	// (0) a one time population of the sketch array that stores nodes in
	// increasing order of their max commodity demands; will be used to
	// process nodes in a different order;
	populate_sketch_arrays_for_nodes_sorting();


	// (1) first, run of MCF solver with latency L=100000 (inf);
	printf("\n\nPART 1: MAX CONCURRENT FLOW \n---------------------------");
	_lambda_max = 
		mcf(_delta, _epsilon1,_epsilon2,_epsilon3, 0); // 1, 0.1, 0.01, 0.6

	_latency_max = _total_latency;
	//printf("\n latency max = %lf lambda max = %lf",_latency_max,_lambda_max);

	// () early exit if there is no feasible solution;
	if ( _lambda_max < 1 - _epsilon1) {
		printf("\nWarning: No feasible solution, lambda_max=%lf\n",_lambda_max);
		free_topology();
		return -1;
	}



	// (2) second, binary search to refine L
	printf("\n\nPART 2: BINARY SEARCH FOR L \n---------------------------");
	LL = 0;
	UL = _latency_max / _lambda_max;
	_s = -1;
	int mcf_runs_count = 0;
	double lambda = 0.0; // "iter/max_factor"
	while ((UL - LL) / LL > _epsilon2) { // 0.01, adica 1%;
		//printf("\n--------------------------- binary run %d",mcf_runs_count);
		mcf_runs_count ++;

		// () set Latency as the middle point between LL and UL;
		L = (LL + UL) / 2;
		// () call MCF for maximum concurrent multicommodity flow; this call
		// of MCF modifies LL and UL using the "interval estimation" technique
		// from the paper of Hu and Zhu;
		lambda = mcf(_delta, _epsilon4,_epsilon2,_epsilon3, 2);
		if ( lambda < 0) {
			lambda = mcf(_delta, _epsilon3,_epsilon2,_epsilon3/2, 2);
			if ( lambda < 0) break;
		}
	}
	
	printf("\n LL=%lf, UL=%lf", LL, UL);
	printf("\n Final L = %lf", UL);


	// () runtime;
	end_clock = clock();
	assert(end_clock != (clock_t)(-1));
	diff_clock = end_clock - start_clock;
	printf("\n mcf_solver total cputime (processor time) = %.3f sec\n", (double)diff_clock/CLOCKS_PER_SEC);

	return UL;
}

void MCF::populate_sketch_arrays_for_nodes_sorting()
{
	// a one time population of the sketch array that stores nodes in
	// increasing order of their max commodity demands; will be used to
	// process nodes in a different order;
	for ( int i = 0; i < no_node; i++) {
		double max_comm_of_this_node = 0;
		for ( int j = 0; j < nodes[i].no_comm; j++) {
			if ( _commodities[nodes[i].comms[j]].demand > max_comm_of_this_node)
				max_comm_of_this_node = _commodities[nodes[i].comms[j]].demand;
		}
		
		_sketch_nodes_max_commodity.push_back( PAIR_TWO(i, max_comm_of_this_node));
	}
	sort( _sketch_nodes_max_commodity.begin(), _sketch_nodes_max_commodity.end());
}

bool MCF::build_network_from_file(char *network_filename,
	double latency_limit, double rate)
{
	// rate is the demand coefficient (injection rate): 0.05, 0.1, 0.15, 0.2, 0.25;
	// latency_limit and rate are not used;

    FILE *fp; // file pointer for network file;

	// (1) import the network topology and the demands;
   	if (( fp = fopen(network_filename, "r"))) {
       	read_network_topology_and_demands( fp); // rate is default;
   	} else {
       	printf("Can not open file: %s\n", network_filename);
      	return false;
   	}

	// (2) initialize;
	init_param();

	// (3) cleanup;
	fclose(fp);

	return true;
}

bool MCF::build_network_from_host_application(
	ROUTERS_GRAPH &noc_graph, APPLICATION_GRAPH &app_graph,
	double capacity_a)
{
	// used from inside the host application that hosts also the floorplanner
	// and the vnoc simulator;
	// capacity is the capacity of every arc; it is set thru argument
	// in order to facillitate iterative search;


	// (1) construct the network topology and the demands using the noc_graph
	// and the application graph;


	// () arcs of the app_graph represent communication requirements between 
	// IP/cores; these represent the commodities with demands equal to the
	// communication volume;
	no_commodity = app_graph.arcs_count();


	// () nodes;
	no_node = noc_graph.nodes_count();
	nodes = (NODE*)malloc(sizeof(NODE) * no_node);
	if ( nodes == NULL) {
		printf("\nError: Unable to malloc <nodes>.\n"); exit (1);
	}

	for ( int i = 0; i < no_node; i++) {
		nodes[i].id = i;
		nodes[i].x = noc_graph.get_node(i)->x();
		nodes[i].y = noc_graph.get_node(i)->y();
		nodes[i].pre = -1;
		nodes[i].dist = DBL_MAX;
		nodes[i].no_comm = 0;
		nodes[i].comms = NULL;
		nodes[i].no_edge = 0;
		nodes[i].dij_visited = 0;
		nodes[i].dij_updated = 0;
		nodes[i].min_visited = 0;
		// here we work with a fixed pre-allocation; not optimal; we should
		// allocate only as much as needed; also in this way we have to make
		// sure rthere will be no nodes with a bigger degree than MAX_DEGREE=40;
		nodes[i].edges = (int *)malloc(sizeof(int) * MAX_DEGREE);
		if ( nodes[i].edges == NULL) {
			printf("\nError: Unable to malloc <nodes.edges>.\n"); exit (1);
		}

		nodes[i]._preferred_path = (int *)malloc(sizeof(int) * no_commodity);
		if ( nodes[i]._preferred_path == NULL) {
			printf("\nError: Unable to malloc <_preferred_path>.\n"); exit (1);
		}
		for ( int j = 0; j < no_commodity; j++) {
			nodes[i]._preferred_path[j] = -1;			
		}
	}


	// () edges;
	// Note: here the number of edges is double that of the noc_graph;
	// that is for an edge i->j I create an edge j->i also; that is 
	// because physical links in the NOC architecture are bidirectional
	// (actually they are two sets of say 64 wires, a set in one direction
	// and the other set in the other direction);
	int half_no_edge = noc_graph.arcs_count();
	no_edge = 2 * half_no_edge;
	edges = (EDGE *)malloc(sizeof(EDGE) * no_edge);
	if ( edges == NULL) {
		printf("\nError: Unable to malloc <edges>.\n"); exit (1);
	}
	_temp_edge_flow = (double*)malloc(sizeof(double) * no_edge);
	if ( _temp_edge_flow == NULL) {
		printf("\nError: Unable to malloc <_temp_edge_flow>.\n"); exit (1);
	}
	for ( int i = 0; i < half_no_edge; i++) {
		// (a) one edge corresponding to the edge from the app_graph;
		edges[i].id = i;
		edges[i].src = noc_graph.get_arc(i)->src_id();
		edges[i].dest = noc_graph.get_arc(i)->des_id();
		edges[i].power = 0.0; // we do not use power for now;
		// I use the wirelength instead of delay; I could extrapolate the
		// Elmore delay here too;
		// Note: in fact, because the NoC will be clocked at the freq. determined
		// by longest link, there is no point in using individual link-lengths,
		// because it may actually make the flows go through unnecessarily
		// more routers because of potentially shorter physical links between
		// a-c via b; using individual link-lebgth would make sense if I used
		// elastic interconnects though;
		edges[i].latency = noc_graph.max_manhattan_link_length(); // get_arc(i)->weight();
		edges[i].length = 0.0;
		edges[i].capacity = capacity_a;
		edges[i].flow = 0.0;
		// Note: here I can allocate _flows here because no_commodity is
		// already known;
		edges[i]._flows = (double *)malloc(sizeof(double) * no_commodity);
		if ( edges[i]._flows == NULL) {
			printf("\nError: Unable to malloc <edges._flows>.\n"); exit (1);
		}
		for ( int j = 0; j < no_commodity; j++) {
			edges[i]._flows[j] = 0.0;			
		}

		// (b) counterpart edge, in the opposite direction, corresponding 
		// to the edge from the app_graph; opposite edges are indexed
		// in the second half;
		edges[half_no_edge + i].id = half_no_edge + i;
		// reverse src and des here;
		edges[half_no_edge + i].src = noc_graph.get_arc(i)->des_id();
		edges[half_no_edge + i].dest = noc_graph.get_arc(i)->src_id();
		edges[half_no_edge + i].power = 0.0; // not used;
		edges[half_no_edge + i].latency = noc_graph.max_manhattan_link_length(); // get_arc(i)->weight();
		edges[half_no_edge + i].length = 0.0;
		edges[half_no_edge + i].capacity = capacity_a;
		edges[half_no_edge + i].flow = 0.0;
		edges[half_no_edge + i]._flows = (double *)malloc(sizeof(double) * no_commodity);
		if ( edges[half_no_edge + i]._flows == NULL) {
			printf("\nError: Unable to malloc <edges._flows>.\n"); exit (1);
		}
		for ( int j = 0; j < no_commodity; j++) {
			edges[half_no_edge + i]._flows[j] = 0.0;			
		}
	}


	// () record adjacent edges for each node;
	for ( int i = 0; i < no_edge; i++) {
		int index = edges[i].src;
		nodes[index].edges[nodes[index].no_edge] = edges[i].id;
		nodes[index].no_edge++;
	}


	// () read demands/commodities;
	double amount;
	int src, dest;
	_commodities.resize( no_commodity);
	for ( int i = 0; i < no_commodity; i++) {
		_commodities[i].id = i;
		// node ids of the app_graph are core ids;
		src = noc_graph.core_id_to_router_id( app_graph.get_arc(i)->src_id());
		_commodities[i].src = src;
		dest = noc_graph.core_id_to_router_id( app_graph.get_arc(i)->des_id());
		_commodities[i].dest = dest;

		// demand of this commodity is the communication volume;
		amount = app_graph.get_arc(i)->comm_volume();
		_commodities[i].demand = amount * 1; // rate = 1 by default;
		_commodities[i].left_demand = amount;
		if (nodes[src].comms == NULL) {
			nodes[src].comms = (int *)malloc(sizeof(int) * no_node);
			if ( nodes[src].comms == NULL) {
				printf("\nError: Unable to malloc <nodes.comms>.\n"); exit (1);
			}
		}
		nodes[src].comms[nodes[src].no_comm] = i;
		nodes[src].no_comm ++;
	}

	// debug;
	//print_network_demands();
	

	// (2) initialize;
	init_param();

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// MCF Dijkstra
//
////////////////////////////////////////////////////////////////////////////////

void MCF::shortest_path( int s, int no_comm, int *dest_flag) 
{
	// implements Dijkstra's all paths shortest path algorithm;

	PQDATUM wf, wf1; // WAVEFRONTS;

	PQUEUE pq;
	int *pos = (int *)malloc(sizeof(int) * no_node);
	if ( pos == NULL) {
		printf("\nError: Unable to malloc <shortest_path>.\n"); exit (1);
	}

	pq.pqinit( 400); // 400 is a constant, just dont care the memory;
	// source "s" resets;
	nodes[s].pre = -1;
	nodes[s].pre_edge = -1;
	nodes[s].dist = 0.0;

	wf.set_node( s); // sourse "s";
	wf.set_dist( 0.0);

	pq.pqinsert( wf, pos);

	while ( !pq.pqempty()) {
		int v, w;
		
		// retreive the shortest non-visited node
		pq.pqremove( &wf1, pos);
		v = wf1.node();
		if ( dest_flag[v]) no_comm--;
		// break when all shortest paths to all destinations from source "s"
		// have been found;
		if ( no_comm <= 0) break;

		//printf(" v=%d", nodes[v].id);
		nodes[v].dij_visited = _rd;
		for ( int i = 0; i < nodes[v].no_edge; i++) {
			w = edges[nodes[v].edges[i]].dest;
			if ( nodes[w].dij_visited != _rd) // not VISITED;
				if ( nodes[w].dij_updated != _rd || // not UPDATED;
					 nodes[w].dist > wf1.dist() + edges[nodes[v].edges[i]].length) {
					nodes[w].pre = v;
					nodes[w].pre_edge = nodes[v].edges[i];
					nodes[w].dist = wf1.dist() + edges[nodes[v].edges[i]].length;
					wf.set_node( w);
					wf.set_dist( nodes[w].dist);
					if (nodes[w].dij_updated != _rd) {
						pq.pqinsert( wf, pos);
					} else {
						pq.pqdeckey( wf, pos);
					}
					nodes[w].dij_updated = _rd;
				}
		}
	}
	pq.pqfree( pos); // clean up;
}

void MCF::shortest_path_with_limited_bandwidth(
	int s, double t_commodity, int no_comm, int *dest_flag) 
{
	// developed from the above but I book-keep edge capacities;
	// for each new node processed to find the shortest tree, edge capacities
	// will reflect previously routed flows;
	// Note: used only when MCF is used to compute routing paths as shortest
	// paths but under bandwidth constraints;
	// t_commodity is the commodity attempted to route now along shortest path;

	//printf("\nEnter shortest path");
	PQDATUM wf, wf1; // WAVEFRONTS;

	PQUEUE pq;
	int *pos = (int *)malloc(sizeof(int) * no_node);
	if ( pos == NULL) {
		printf("\nError: Unable to malloc <shortest_path>.\n"); exit (1);
	}

	pq.pqinit( 400); // 400 is a constant, just dont care the memory;
	// source "s" resets;
	nodes[s].pre = -1;
	nodes[s].pre_edge = -1;
	nodes[s].dist = 0.0;

	wf.set_node( s); // sourse "s";
	wf.set_dist( 0.0);

	pq.pqinsert( wf, pos);

	while ( !pq.pqempty()) {
		int v, w;
		
		// retreive the shortest non-visited node
		pq.pqremove( &wf1, pos);
		v = wf1.node();
		if ( dest_flag[v]) no_comm--;
		//printf(" v=%d", nodes[v].id);
		// break when all shortest paths to all destinations from source "s"
		// have been found;
		if ( no_comm <= 0) break;

		nodes[v].dij_visited = _rd;
		for ( int i = 0; i < nodes[v].no_edge; i++) {


			w = edges[nodes[v].edges[i]].dest;
			if ( nodes[w].dij_visited != _rd) {


				if ( nodes[w].dij_updated != _rd) {
					// if this node "w" is visited first time now, set its 
					// parent at once;
					nodes[w].pre = v;
					nodes[w].pre_edge = nodes[v].edges[i];
					nodes[w].dist = wf1.dist() + edges[nodes[v].edges[i]].length;
					wf.set_node( w);
					wf.set_dist( nodes[w].dist);
					if (nodes[w].dij_updated != _rd) {
						pq.pqinsert( wf, pos);
					} else {
						pq.pqdeckey( wf, pos);
					}
					nodes[w].dij_updated = _rd;
				}
				else if ( edges[nodes[v].edges[i]].capacity >= t_commodity) {
					// as long as this edge has sufficient capacity, then select 
					// the shortest path OR select it anyway if the prev. parent 
					// that was recorded did not have enough capacity;
					if ( edges[nodes[w].pre_edge].capacity < t_commodity ||
						 wf1.dist() + edges[nodes[v].edges[i]].length < nodes[w].dist) {
						nodes[w].pre = v;
						nodes[w].pre_edge = nodes[v].edges[i];
						nodes[w].dist = wf1.dist() + edges[nodes[v].edges[i]].length;
						wf.set_node( w);
						wf.set_dist( nodes[w].dist);
						if (nodes[w].dij_updated != _rd) {
							pq.pqinsert( wf, pos);
						} else {
							pq.pqdeckey( wf, pos);
						}
						nodes[w].dij_updated = _rd;
					}
				} 
				else if ( edges[nodes[v].edges[i]].capacity >
						  edges[nodes[w].pre_edge].capacity) {
					// if this edge to "w" has a higher capacity than the already 
					// recorded as parent for "w", then use it even though its dist
					// may be longer; we give higher priority to capacity constraints 
					// then to minimum distance;
					nodes[w].pre = v;
					nodes[w].pre_edge = nodes[v].edges[i];
					nodes[w].dist = wf1.dist() + edges[nodes[v].edges[i]].length;
					wf.set_node( w);
					wf.set_dist( nodes[w].dist);
					if (nodes[w].dij_updated != _rd) {
						pq.pqinsert( wf, pos);
					} else {
						pq.pqdeckey( wf, pos);
					}
					nodes[w].dij_updated = _rd;
				}
				else if ( edges[nodes[v].edges[i]].capacity ==
						  edges[nodes[w].pre_edge].capacity) {
					// if this edge has same capacity as the already recorded parent,
					// then use it if dist is smaller;
					if ( wf1.dist() + edges[nodes[v].edges[i]].length < nodes[w].dist) {
						nodes[w].pre = v;
						nodes[w].pre_edge = nodes[v].edges[i];
						nodes[w].dist = wf1.dist() + edges[nodes[v].edges[i]].length;
						wf.set_node( w);
						wf.set_dist( nodes[w].dist);
						if (nodes[w].dij_updated != _rd) {
							pq.pqinsert( wf, pos);
						} else {
							pq.pqdeckey( wf, pos);
						}
						nodes[w].dij_updated = _rd;
					}
				}


			}
		}
	}
	pq.pqfree( pos); // clean up;
	//printf("\nDone shortest path");
}

////////////////////////////////////////////////////////////////////////////////
//
// MCF network graph
//
////////////////////////////////////////////////////////////////////////////////

void MCF::read_network_topology_and_demands(FILE *fp, double rate) 
{
	// example of .network format:
	// 8 <-- num of nodes
	// 0 100 300 <-- node id, (x,y) location in um
	// 1 100 100
	// 2 200 300
	// 3 200 100
	// 4 300 300
	// 5 300 100
	// 6 400 300
	// 7 400 100
	// 12 <-- num of edges
	// 0 0 2 10.0 1.0 2.00 <-- id, src, des, capacity, power, delay
	// 1 2 4 10.0 1.0 2.00 
	// 2 3 2 10.0 4.0 6.00 
	// 3 1 3 10.0 1.0 2.00 
	// 4 3 5 10.0 1.0 2.00
	// 5 2 3 10.0 4.0 6.00 
	// 6 4 2 10.0 1.0 2.00
	// 7 5 3 10.0 1.0 2.00
	// 8 5 4 10.0 4.0 6.00 
	// 9 4 5 10.0 4.0 6.00 
	// 10 4 6 10.0 1.0 2.00
	// 11 5 7 10.0 1.0 2.00
	// 2 <-- num of demands (commodities)
	// 0 0 7 0.577004 <-- id src des amount
	// 1 1 6 1.777268

	int id, x, y, src, dest;
	double power, delay, capacity;


	// (1) nodes
	fscanf(fp,"%d", &no_node);
	nodes = (NODE*)malloc(sizeof(NODE) * (no_node));
	if ( nodes == NULL) {
		printf("\nError: Unable to malloc <nodes>.\n"); exit (1);
	}
	for ( int i = 0; i < no_node; i++) {
		fscanf(fp, "%d %d %d", &id, &x, &y);
		nodes[i].id = id;
		nodes[i].x = x;
		nodes[i].y = y;
		nodes[i].pre = -1;
		nodes[i].dist = DBL_MAX;
		nodes[i].no_comm = 0;
		nodes[i].comms = NULL;
		nodes[i].no_edge = 0;
		nodes[i].dij_visited = 0;
		nodes[i].dij_updated = 0;
		nodes[i].min_visited = 0;

		// here we work with a fixed pre-allocation; not optimal; we should
		// allocate only as much as needed; also in this way we have to make
		// sure rthere will be no nodes with a bigger degree than MAX_DEGREE=40;
		nodes[i].edges = (int *)malloc(sizeof(int) * MAX_DEGREE);
		if ( nodes[i].edges == NULL) {
			printf("\nError: Unable to malloc <nodes.edges>.\n"); exit (1);
		}
	}


	// (2) edges
	fscanf(fp,"%d", &no_edge);
	edges = (EDGE *)malloc(sizeof(EDGE) * (no_edge));
	if ( edges == NULL) {
		printf("\nError: Unable to malloc <edges>.\n"); exit (1);
	}
	_temp_edge_flow = (double*)malloc(sizeof(double) * (no_edge));
	if ( _temp_edge_flow == NULL) {
		printf("\nError: Unable to malloc <_temp_edge_flow>.\n"); exit (1);
	}
	for (int i = 0; i < no_edge; i++) {
		fscanf(fp, "%d %d %d %lf %lf %lf",&id,&src,&dest,&capacity,&power,&delay);
		edges[i].id = id;
		edges[i].src = src;
		edges[i].dest = dest;
		edges[i].capacity = capacity;
		edges[i].power = power;
		edges[i].latency = delay;
		edges[i].length = 0.0;

		edges[i].flow = 0.0;
		edges[i]._flows = NULL;
	}


	// (3) record adjacent edges for each node;
	for ( int i = 0; i < no_edge; i++) {
		int index = edges[i].src;
		nodes[index].edges[nodes[index].no_edge] = edges[i].id;
		nodes[index].no_edge++;
	}


	// (4) read demands/commodities;
	double amount;
	fscanf(fp,"%d", &no_commodity);
	_commodities.resize( no_commodity);
	for ( int i = 0; i < no_commodity; i++) {
		fscanf(fp,"%d %d %d %lf", &id, &src, &dest, &amount);
		_commodities[i].id = id;
		_commodities[i].src = src;
		_commodities[i].dest = dest;
		_commodities[i].demand = amount * rate; // rate = 1 by default;
		_commodities[i].left_demand = amount;

		if (nodes[src].comms == NULL) {
			nodes[src].comms = (int *)malloc(sizeof(int) * no_node);
			if ( nodes[src].comms == NULL) {
				printf("\nError: Unable to malloc <nodes[src].comms>.\n"); exit (1);
			}
		}
		nodes[src].comms[nodes[src].no_comm] = i;
		nodes[src].no_comm++;
	}


	// () reset;
	for (int i = 0; i < no_edge; i++) {
		// Note: I had to delay this allocation because no_commodity had 
		// not been set yet;
		edges[i]._flows = (double *)malloc(sizeof(double) * (no_commodity));
		if ( edges[i]._flows == NULL) {
			printf("\nError: Unable to malloc <edges._flows>.\n"); exit (1);
		}
		for (int j = 0; j < no_commodity; j++) {
			edges[i]._flows[j] = 0.0;
		}
	}

	// debug;
	//print_network_demands();
	
	return;
}

void MCF::free_topology()
{
	int i;

	//free edges
	for (i = 0; i < no_edge; i++) {
		free( edges[i]._flows);
	}
	free( edges);
	free( _temp_edge_flow);

	//free nodes
	for (i = 0; i < no_node; i++) {
		free( nodes[i].comms);
		free( nodes[i].edges);
		free( nodes[i]._preferred_path);
	}
	free( nodes);

	return;
}

void MCF::print_network_demands( bool print_only_edges)
{
	printf("\nNetwork and demands");
	printf("\nNodes %d",no_node);
	if ( print_only_edges == false) {
		for (int i = 0; i < no_node; i++) {
			printf("\n %d  (%d %d)", nodes[i].id, nodes[i].x, nodes[i].y);
			printf("  num_commodities=%d dist=%lf",nodes[i].no_comm,nodes[i].dist);
			printf("\n     ");
			for ( int k = 0; k < nodes[i].no_edge; k++) {
				printf(" %d", nodes[i].edges[k]);
			}
		}
	}
	printf("\nEdges %d",no_edge);
	for (int i = 0; i < no_edge; i++) {
		printf("\n %d  %d -> %d  cap: %.2lf  flow: %.2lf  latency: %.2lf",
			edges[i].id, edges[i].src, edges[i].dest,
			edges[i].capacity, edges[i].flow, edges[i].latency);
		printf("  breakdown: ");
		for (int j = 0; j < no_commodity; j++) {
			printf(" %.2lf", edges[i]._flows[ j]);
		}
	}
	printf("\nDemands %d",no_commodity);
	if ( print_only_edges == false) {
		for (int i = 0; i < no_commodity; i++) {
			//printf("\n %d  %d -> %d demand=%.2lf left_demand=%.2lf", _commodities[i].id,
			//	   _commodities[i].src, _commodities[i].dest,
			//	   _commodities[i].demand, // amount * rate
			//	   _commodities[i].left_demand); // amount
			printf("\n%d %d %d %.2lf", _commodities[i].id,
				   _commodities[i].src, _commodities[i].dest,
				   _commodities[i].demand);
		}
	}
	printf("\n");
}

void MCF::print_routing_paths()
{
	// call only after a call of do_randomized_rounding();
	printf("\nRandomized rounded paths:");

	for ( int i = 0; i < no_commodity; i++) {
		printf("\nCommodity %d: %d -> %d (%.1f): ", i,
			_commodities[i].src, _commodities[i].dest, _commodities[i].demand );

		int src_id = _commodities[i].src;
		while ( src_id != _commodities[i].dest) {
			printf(" %d", src_id);
			src_id = nodes[src_id]._preferred_path[i];
		}
		printf(" %d", src_id); // dest;
	}
	printf("\n");
}

void MCF::print_max_accumulated_flow()
{
	// look at all edges and see what is the one with maximum flow
	// thru it;
	double max_edge_flow = 0.0;
	int edge_id = -1;
	for ( int i = 0; i < no_edge; i++) {
		if ( edges[i].flow > max_edge_flow) {
			max_edge_flow = edges[i].flow;
			edge_id = i;
		}
	}
	printf("\nEdge < %d > has maximum flow: %.2f", edge_id, max_edge_flow);
}
 
void MCF::print_backward_shortest_path( int t)
{
	// debug only;
	printf("\n");
	while ( nodes[t].pre != -1) {
		printf(" %d ->", t);
		t = nodes[t].pre;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// MCF randomized rounding
//
////////////////////////////////////////////////////////////////////////////////

bool MCF::do_randomized_rounding()
{
	// () after mcf_solver finds a fractional flow solution, we do a 
	// randomized rounding to set only one path for each commodity;
	// otherwise the commodities would traverse multiple paths that
	// would translate in path splitting for packets which would require
	// router architecture modification too much and re-ordering of packets
	// at destination;

	for ( int i = 0; i < no_commodity; i++) {
		// () for each commodity we start from its source and traverse
		// downstream nodes; at each step we pick up the node that has
		// the largest fraction of this commodity as the preferred path;
		// record this preferred path in the preferred_path array;

		int src_id = _commodities[i].src;
		while ( src_id != _commodities[i].dest) {
			// recursively travel from src to dest searching for maximum
			// fractional flow;
				
			int id_max_flow_fraction = -1;
			double max_flow_fraction = 0.0;
			for ( int k = 0; k < nodes[ src_id].no_edge; k++) {
				// for each adjacent edge look at the commodity index "i",
				// and seek the edge id with maximum flow fraction;

				int edge_id = nodes[ src_id].edges[k];
				if ( max_flow_fraction < edges[edge_id]._flows[ i]) {
					max_flow_fraction = edges[edge_id]._flows[ i];
					id_max_flow_fraction = edge_id;
				}
			}
			assert(id_max_flow_fraction >= 0 && id_max_flow_fraction < no_edge);

			// () record the neighbor node id as the downstream node of the
			// preferred path for this commodity; that is, along the edge with
			// max fraction of flow from current node for this commodity;
			assert(edges[id_max_flow_fraction].dest >= 0 &&
				edges[id_max_flow_fraction].dest < no_node);
			nodes[src_id]._preferred_path[i] = edges[id_max_flow_fraction].dest;

			// prepare for next iter;
			src_id = nodes[src_id]._preferred_path[i];
		}
	}

	return true;
}

void MCF::calculate_average_path_length()
{
	// () calculate the average path length among all communications;
	int total_paths_length = 0;
	for ( int i = 0; i < no_commodity; i++) {
		int src_id = _commodities[i].src;
		while ( src_id != _commodities[i].dest) {
			total_paths_length ++;
			src_id = nodes[src_id]._preferred_path[i];
		}
	}
	printf("\nAvg. path length for %d communications: %.2f\n",no_commodity,
		   double(total_paths_length)/no_commodity);
}	

////////////////////////////////////////////////////////////////////////////////
//
// Disjkstra shortest path based routing paths -- for comparison purposes;
//
////////////////////////////////////////////////////////////////////////////////

bool MCF::do_shortest_paths_as_preferred_paths_with_inf_bandwidth()
{
	// Version 1: infinite bandwidth;
	// do not do any MCF solving; simply call shortest path algo for
	// each commodity; this will be used for comparison to the case when
	// MCF is used;
	// Note: this is version 1 "with infinite bandwidth"; in otehr words
	// once a flow is routed thru the shortest path we assume that the network
	// will entirely be available for the next flow and find its shortes path, etc.
	// However, when all flows will concurrently use the network the bandwidth
	// may not be available; so we should compute the shortest path of the second,
	// third flows, etc. on a network with bandwidths/capacities updated to reflect
	// the fact that we used some of it for first flow, etc.; this is done in
	// version two of this function below;

	// Note: because mcf_solver is not called we overwrite .length with 
	// .latency (which is currently the physical length) for shortest path 
	// calculation purposes; in this run mode these variables like length, 
	// latency, duals, etc. do not keep their meanings;
	init( _delta);
	_rd = 1;
	for (int i = 0; i < no_node; i++) {
		nodes[i].dij_visited = 0;
		nodes[i].dij_updated = 0;
		nodes[i].min_visited = 0;
	}
	for (int j = 0; j < no_edge; j++) {
		edges[j].length = edges[j].latency;
	}
	

	for ( int i = 0; i < no_node; i++) {

		if ( nodes[i].no_comm) { // if this node is source of "r" _commodities;
			//printf("\nNODE <%d> HAS %d COMMODITIES: ",i,nodes[i].no_comm);

			int commodities_left = nodes[i].no_comm;
			int *dest_flag = (int*)malloc(no_node*sizeof(int));
			if ( dest_flag == NULL) {
				printf("\nError: Unable to malloc <do_shortest*>.\n"); exit (1);
			}
			memset((void*)dest_flag,0,no_node*sizeof(int));
			// dest_flag is set "1" for nodes that are destinations of _commodities;
			for ( int j = 0; j < nodes[i].no_comm; j++) {
				dest_flag[_commodities[nodes[i].comms[j]].dest] = 1;
			}


			// compute shortest PATHS tree;
			shortest_path( nodes[i].id, commodities_left, dest_flag);


			for ( int j = 0; j < nodes[i].no_comm; j++) {
				int comm_id = nodes[i].comms[j];
				//printf("\nCOMMODITY [%d]: ", comm_id);

				double t_commodity = _commodities[comm_id].demand;
				// for each of the commodities that has this node as src
				// get the shortest path and stored it as preferred;
				int t = _commodities[comm_id].dest;
				int prev_t = t;
				//printf(" %d", prev_t);
				// walk back the shortest path and record preferred path;
				while ( nodes[t].pre != -1) {
					t = nodes[t].pre;
					//assert( t >= 0 && t < no_node);
					nodes[t]._preferred_path[comm_id] = prev_t;
					//printf(" <- %d", t);
				
					// update also the flow of each edge used by this path;
					//assert( nodes[prev_t].pre_edge >= 0 && nodes[prev_t].pre_edge < no_edge);
					edges[nodes[prev_t].pre_edge].flow += t_commodity;

					prev_t = t; // prepare for next iter;
				}
			}
			

			// clean up;
			_rd ++;
			free(dest_flag);
		}
	}

	return true;
}

bool MCF::do_shortest_paths_as_preferred_paths_with_limited_bandwidth()
{
	// Version 2: limited bandwidth;
	// -- sort all flows by their demand size;
	// -- route the first one; update the network to reflect only available
	// links for the next flow in terms of available capacities/bandwidths;
	// -- route second flow; update again; etc.;

	// Note1: because mcf_solver is not called we overwrite .length with 
	// .latency (which is currently the physical length) for shortest path 
	// calculation purposes; in this run mode these variables like length, 
	// latency, duals, etc. do not keep their meanings;
	// Note2: every time an edge is used by a flow we accumulate that and
	// store it inside flow variable of each edge; at the same time decrease 
	// the edge capacity with the just routed flow; shortest path for a commodity
	// then will select the edge with the highest capacity available;
	init( _delta);
	_rd = 1;
	for (int i = 0; i < no_node; i++) {
		nodes[i].dij_visited = 0;
		nodes[i].dij_updated = 0;
		nodes[i].min_visited = 0;
	}
	for (int j = 0; j < no_edge; j++) {
		edges[j].length = edges[j].latency;
	}

	// store in a vector all max commodities for each node; then sort
	// the nodes by that, and later will process nodes in decreasing order 
	// of their max commodity;
	vector<PAIR_TWO> representative_commodity;
	for ( int i = 0; i < no_node; i++) {
		double max_comm_of_this_node = 0;
		for ( int j = 0; j < nodes[i].no_comm; j++) {
			if ( _commodities[nodes[i].comms[j]].demand > max_comm_of_this_node)
				max_comm_of_this_node = _commodities[nodes[i].comms[j]].demand;
		}
		
		representative_commodity.push_back( PAIR_TWO(i, max_comm_of_this_node));
	}
	sort( representative_commodity.begin(), representative_commodity.end());
	

	// process nodes in decreasing order of their max commodity;
	for ( int k = no_node - 1; k >= 0; k--) {
		// k is to index representative_commodity;
		// get the index of actual node to be processed by shortest path;
		int i = representative_commodity[k].id();
		//printf("\n<< %d >>  ", i);
 
		if ( nodes[i].no_comm) { // if this node is source of "r" _commodities;


			for ( int j = 0; j < nodes[i].no_comm; j++) {
				int comm_id = nodes[i].comms[j];
				//printf("\nsrc:%d --> dst: %d", i, _commodities[nodes[i].comms[j]].dest);


				int commodities_left = 1;
				int *dest_flag = (int*)malloc(no_node*sizeof(int));
				if ( dest_flag == NULL) {
					printf("\nError: Unable to malloc <do_shortest*>.\n"); exit (1);
				}
				memset((void*)dest_flag,0,no_node*sizeof(int));
				// dest_flag is set "1" for THE node that is destination 
				// this _commodities;
				dest_flag[_commodities[comm_id].dest] = 1;

				// this commodity that has the same node "i" as src has to be
				// reflected as being routed by all capacities of all edges
				// that are part of the path;
				double t_commodity = _commodities[comm_id].demand;

				// compute shortest PATH for this commodity only;
				shortest_path_with_limited_bandwidth(
					nodes[i].id, t_commodity, commodities_left, dest_flag);

				
				

				// for each of the commodities that has this node as src
				// get the shortest path and stored it as preferred;
				int t = _commodities[comm_id].dest;
				int prev_t = t;
				//printf(" %d", t);
				// walk back the shortest path and record preferred path;
				while ( nodes[t].pre != -1) {
					t = nodes[t].pre;
					nodes[t]._preferred_path[comm_id] = prev_t;
					//printf(" <- %d", t);


					// book-keeping of flows thru edges; needed for stats only;
					// subtract t_commodity from (available) capacity of this edge;
					//printf("\n before cap: %.2f", edges[nodes[prev_t].pre_edge].capacity);
					edges[nodes[prev_t].pre_edge].capacity = max( 0.0, 
						edges[nodes[prev_t].pre_edge].capacity - t_commodity);
					edges[nodes[prev_t].pre_edge].flow += t_commodity;
					//printf(" after cap: %.2f", edges[nodes[prev_t].pre_edge].capacity);

					prev_t = t; // prepare for next iter;
				}



				// clean up;
				_rd ++;
				free(dest_flag);
			}
		}
	}

	return true;
}
