#include "config.h"
#include <limits.h>
#include "hetero.h"
#include "vnoc.h"
#include "vnoc_event.h"
#include <math.h>
#include <iomanip>
#include <sstream>
#include <cstdio>
#include <algorithm>

using namespace std;


// I should encapsulate this one inside _vnoc;
unsigned long VC_MASK[10] = {1, 2, 4, 8, 16, 32, 64, 128, 256, 512};


////////////////////////////////////////////////////////////////////////////////
//
// ROUTER_INPUT
//
////////////////////////////////////////////////////////////////////////////////

ROUTER_INPUT::ROUTER_INPUT(long physical_ports_count, long vc_number):
	_input_buff(),
	_vc_state(),
	_routing(),
	_selected_routing(),
	_injection_buff_full(false)
{
	// for now here only allocate memory for _vc_count and reset them
	// for each port; later after constructing the routing table we'll
	// have populated/incremented _vc_count and we'll be able to allocate
	// mem for the actual input and out buffers as well;

	// Note: the port id zero which is associated with the local PE
	// will have a fixed number of virtual channels for each router;
	// this is by default 3;
	_vc_count.resize( physical_ports_count);
	//_vc_count[ 0] = vc_number;
	for ( long i = 0; i < physical_ports_count; i++) {
		_vc_count[ i] = 0; // reset;
	}
}
void ROUTER_INPUT::initialize_input_buffers(long physical_ports_count)
{
	// called only once after the ctor has been called and _vc_count have
	// been counted;
	long i = 0;
	_input_buff.resize( physical_ports_count);
	for ( i = 0; i < physical_ports_count; i++) {
		_input_buff[i].resize( _vc_count[i]);
	}
	_vc_state.resize( physical_ports_count);
	for ( i = 0; i < physical_ports_count; i++) {
		_vc_state[i].resize( _vc_count[i], INIT);
	}
	_routing.resize( physical_ports_count);
	for ( i = 0; i < physical_ports_count; i++) {
		_routing[i].resize( _vc_count[i]);
	}
	_selected_routing.resize( physical_ports_count);
	for ( i = 0; i < physical_ports_count; i++) {
		_selected_routing[i].resize( _vc_count[i], VC_NULL);
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// ROUTER_OUTPUT
//
////////////////////////////////////////////////////////////////////////////////

ROUTER_OUTPUT::ROUTER_OUTPUT(long physical_ports_count, long vc_number,
	long input_buffer_size, long output_buffer_size):
	_buffer_size_next_r( input_buffer_size),
	_counter_next_r(),
	_flit_state(),
	_assigned_to(),
	_vc_usage(),
	_out_buffer(),
	_out_addr(),
	_local_counter()
{
	// Note: the port id zero which is associated with the local PE
	// will have a fixed number of virtual channels for each router;
	// this is by default 3;
	_vc_count.resize( physical_ports_count);
	//_vc_count[ 0] = vc_number;
	for ( long i = 0; i < physical_ports_count; i++) {
		_vc_count[ i] = 0; // reset;
	}

	_out_buffer.resize( physical_ports_count);
	_flit_state.resize( physical_ports_count);
	_out_addr.resize( physical_ports_count);
	// each of the output ports has initially output_buffer_size 
	// space available;
	_local_counter.resize( physical_ports_count, output_buffer_size);
}
void ROUTER_OUTPUT::initialize_output_buffers(long physical_ports_count)
{
	// called only once after the ctor has been called and _vc_count have
	// been counted;
	long i = 0;
	_counter_next_r.resize( physical_ports_count);
	for ( i = 0; i < physical_ports_count; i++) {
		// initialize the counter at any output port - which maintains
		// the number of available space of input buffer of downstream/next
		// router that has vc_count virtual channels - with input_buffer_size;
		_counter_next_r[i].resize( _vc_count[i], _buffer_size_next_r);
	}
	_assigned_to.resize( physical_ports_count);
	for ( i = 0; i < physical_ports_count; i++) {
		// innitially none of the output ports is assigned 
		// to any input port;
		_assigned_to[i].resize( _vc_count[i], VC_NULL);
	}
	_vc_usage.resize( physical_ports_count);
	for ( i = 0; i < physical_ports_count; i++) {
		_vc_usage[i].resize( _vc_count[i], ROUTER_OUTPUT::FREE);
	}
}

void ROUTER_OUTPUT::remove_flit( long i)
{ 
	_out_buffer[i].erase( _out_buffer[i].begin());
	if ( _flit_state[i].size() > 0) { // this fixed a bug in the original vnoc;
		_flit_state[i].erase( _flit_state[i].begin() );
	}
	_local_counter[i]++; 
}

////////////////////////////////////////////////////////////////////////////////
//
// ROUTER
//
////////////////////////////////////////////////////////////////////////////////

ROUTER::ROUTER( long physical_ports_count, long vc_number, long buffer_size,
	long out_buffer_size,
	long flit_size, VNOC *owner_vnoc, long id, long assigned_core):
	_id(id),
	_assigned_core(assigned_core),
	 // _input and _output do not use same vc_number anymore for all ports;
	// each port will have its own number of vc's depending on the number of 
	// flows thru a link; here the argument "vc_number" will be only be used 
	// for the local PE;
	_input(physical_ports_count, vc_number),
	_output(physical_ports_count, vc_number, buffer_size, out_buffer_size),
	_flit_size(flit_size),
	_physical_ports_count(physical_ports_count),
	_vc_number(vc_number),
	_buffer_size(buffer_size),
	_out_buffer_size(out_buffer_size),
	_total_delay(0),
	_local_injection_time(),
	_packet_counter(0),
	_local_injection_file()
{
	_vnoc = owner_vnoc;


	long i = 0;
	_neighbor_router_id_of_this_port.resize( physical_ports_count);
	_neighbor_port_of_this_port.resize( physical_ports_count);
	for ( i = 0; i < physical_ports_count; i++) {
		_neighbor_router_id_of_this_port[i] = -1;
	}
	for ( i = 0; i < physical_ports_count; i++) {
		_neighbor_port_of_this_port[i] = -1;
	}


	long routers_count = _vnoc->routers_count();
	_routing_LUT.resize( routers_count);
	for ( i = 0; i < routers_count; i++) {
		_routing_LUT[i].resize( routers_count);
	}


	_init_data.resize( flit_size);
	for ( long i = 0; i < flit_size; i++) {
		_init_data[i] = _vnoc->topology()->rng().flat_ull(0, ULLONG_MAX);
	}
	// get and open the trace file name for this router;
	init_local_injection_file();
	local_injection_file() >> _local_injection_time;

	// set the routing algo for this router as required thru the
	// topology file from user;
	_routing_algo = _vnoc->topology()->routing_algo();
}

void ROUTER::init_local_injection_file()
{
	// called by ctor of ROUTER;
	// Note: the way the vNOC implementation works is that every IP/core
	// hooked up to a router will "inject" packets by reading from an 
	// associated file; this has the advantage of reading in traces
	// generated using other applications or real applications; the drawback
	// is that we open a lot of files; I should also implement an internal 
	// traffic generator as an alternative;
	pair <int, int> temp_pair;
	char temp_str[5];

	string name_t = _vnoc->topology()->trace_file();
	name_t.append(".");
	sprintf(temp_str, "%d", _assigned_core);
	name_t.append(temp_str);
	name_t.append(".trs");

	_local_injection_file = new ifstream;
	_local_injection_file->open( name_t.c_str());

	if ( _local_injection_file->fail() ) {
		printf("\nError: Cannot open trace file: %s\n", name_t.c_str());
		exit(1);
	}
}

void ROUTER::receive_packet_from_trace_file()
{
	// retrieve packet from source trace file associated with this router;
	double event_time = _vnoc->event_queue()->current_time();
	long cube_size = _vnoc->topology()->cube_size();
	int src_router_id = -1, des_router_id = -1;
	int core_id = -1;
	long packet_size = -1;

	while ( ( _input.injection_buff_full() == false) && 
			( _local_injection_time <= event_time + S_ELPS)) {

		// read source;
		local_injection_file() >> core_id;
		//assert( !local_injection_file().eof());
		if ( local_injection_file().eof()) { return; }
		src_router_id = _vnoc->noc_graph()->core_id_to_router_id( core_id);
		assert( src_router_id >= 0 && src_router_id < _vnoc->routers_count());

		// read destination;
		local_injection_file() >> core_id;
		assert( !local_injection_file().eof());
		des_router_id = _vnoc->noc_graph()->core_id_to_router_id( core_id);
		assert( des_router_id >= 0 && des_router_id < _vnoc->routers_count());

		// read packet size;
		local_injection_file() >> packet_size;
		assert( !local_injection_file().eof());


		// inject this packet: src -> des;
		// injection load control; in 10 by 10 cycles some packets are NOT
		// going to be injected, but the events creation and files reading
		// will be left untouched;
		_vnoc->hetero()->set_skip_counter( (_vnoc->hetero()->skip_counter() + 1) % 10 );
		if ( _vnoc->hetero()->skip_counter() < _vnoc->hetero()->max_skip_counter()) {
			inject_packet( _packet_counter, src_router_id, des_router_id,
				_local_injection_time, packet_size);
			_packet_counter ++;
			_vnoc->packets_count_incr();
		} else {
			_vnoc->_event_queue->undo_injection();
		}

		// second, create next EVG_ event
		if ( !local_injection_file().eof()) {
			local_injection_file() >> _local_injection_time;
			if ( local_injection_file().eof()) { return; }
		}
	}
}

void ROUTER::inject_packet( long flit_id, int src_router_id, int des_router_id,
	double time, long packet_size)
{
	// version 2: heterogeneous noc's: using router_id's as addresses;
	VC_TYPE vc_t;

	for ( long l = 0; l < packet_size; l++) {

		DATA flit_data; // vector<unsigned long long>;
		for ( long i = 0; i < _flit_size; i++) {
			_init_data[i] = static_cast<DATA_ATOMIC_UNIT>( // "make up stuff";
				_init_data[i] * CORR_EFF + 
				_vnoc->topology()->rng().flat_ull(0, ULLONG_MAX));
			flit_data.push_back( _init_data[i]);
		}


		if ( l == 0) {
			vc_t = pair<long, long>(0, _input.input_buff(0,0).size());
			// if it's the HEADER flit choose the shortest waiting vc queue;
			// Note: here we inject packets into the port 0 (local PE)
			// for whcih the number of virtual channels is the same for each
			// router and stored in each router's _vc_number;
			long input_vc_number = _input.vc_count( 0);
			for ( long i = 0; i < input_vc_number; i++) {
				long t = _input.input_buff(0,i).size();
				if ( t < vc_t.second) {
					vc_t = pair<long, long>(i, t);
				}
			}
			// if the input buffer is empty, set it to be ROUTING;
			if ( _input.input_buff(0, vc_t.first).size() == 0) {
				_input.vc_state_update(0, vc_t.first, ROUTING);
			}

			// if the input buffer has more than predefined flits, then
			// add the flits and flag it;
			if ( _input.input_buff(0, vc_t.first).size() > BUFF_BOUND) {
				// this is causing problems - look for "bug1";
				_input.set_injection_buff_full();
			}
			_input.add_flit( 0, (vc_t.first),
				FLIT(flit_id, FLIT::HEADER, src_router_id,des_router_id, time, flit_data));
		}
		else if ( l < packet_size - 1) {
			_input.add_flit( 0, (vc_t.first),
				FLIT(flit_id, FLIT::BODY, src_router_id,des_router_id, time, flit_data));
		}
		else {
			_input.add_flit( 0, (vc_t.first),
				FLIT(flit_id, FLIT::TAIL, src_router_id,des_router_id, time, flit_data));
		}
		// power module writing here;
	}
}

void ROUTER::receive_flit_from_upstream(long port_id, long vc_id, FLIT &flit)
{
	// receive flit from upstream (neighboring) router;
	_input.add_flit( port_id, vc_id, flit);

	// power module writing here;
	if ( flit.type() == FLIT::HEADER) {
		if ( _input.input_buff(port_id, vc_id).size() == 1) {
			_input.vc_state_update( port_id, vc_id, ROUTING);
		}
	} else {
		if ( _input.vc_state(port_id, vc_id) == INIT) {
			_input.vc_state_update( port_id, vc_id, SW_AB);
		}
	}
}

void ROUTER::consume_flit(double time, const FLIT &flit)
{
	// receive (i.e., consume) one flit at the destination router;
	if ( flit.type() == FLIT::TAIL) {
		_vnoc->event_queue()->inc_finished_count();
		double delta_t = time - flit.start_time();

		update_delay( delta_t); // _total_delay += delta;
	}
}

void ROUTER::receive_credit(long i, long j)
{
	_output.counter_next_r_inc( i, j);
}

////////////////////////////////////////////////////////////////////////////////
//
// simulation of one router
//
////////////////////////////////////////////////////////////////////////////////

void ROUTER::simulate_one_router()
{

	// simulate all routing pipeline stages;
	// stage 5: flit traversal;
	send_flits_via_physical_link();
	// stage 4: flit output buffer;
	send_flit_to_out_buffer();
	// stage 3: switch arbitration;
	sw_arbitration_stage();
	// stage 2: vc arbitration
	vc_arbitration_stage();
	// stage 1: routing decision;
	routing_decision_stage();
}

////////////////////////////////////////////////////////////////////////////////
//
// send_flits_via_physical_link
//
////////////////////////////////////////////////////////////////////////////////

void ROUTER::send_flit_via_physical_link( long i)
{
	// if the output buffer of this port "i" is not empty,
	// take a flit and ship it out thru the physical link to 
	// the downstream router;
	//      4
	//   ------
	//  |      |  
	// 1|   R  | 2
	//   \_____|
	//  0   3
	// example: i=3, then wire_pc_t=4;

	// flit traversal through the link; is it pipelined?
	double event_time = _vnoc->event_queue()->current_time();
	if ( _output.out_buffer(i).size() > 0) {

		int wire_router_id = _neighbor_router_id_of_this_port[ i];
		long wire_port = _neighbor_port_of_this_port[ i];

		FLIT flit_t( _output.get_flit(i));
		VC_TYPE outadd_t = _output.get_addr(i);
		// power stuff here;

		_output.remove_flit(i);
		_output.remove_addr(i);
		
		// physical link delay for the flit to travel forward;
		//double delay = WIRE_DELAY;
		double delay = _vnoc->h_wire_delay();
		_vnoc->event_queue()->add_event( EVENT(EVENT::LINK, 
			event_time + delay,
			_id, // this router is the src;
			wire_router_id, // "address" as r_id to which physical link goes;
			wire_port, // pc; 
			outadd_t.second, // vc; 
			flit_t));
	}
}

void ROUTER::send_flits_via_physical_link()
{
	// send flits from output ports through links to 
	// downstream routers; flits traverse physical links here;
	for (long i = 1; i < _physical_ports_count; i++) {
		send_flit_via_physical_link(i);
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// send_flit_to_out_buffer
//
////////////////////////////////////////////////////////////////////////////////

void ROUTER::send_flit_to_out_buffer()
{
	// send flit to the output buffer - basically switch traversal,
	// after the SA_AB took place;

	for ( long i = 0; i < _physical_ports_count; i++) {
		long input_vc_number = _input.vc_count( i);
		for ( long j = 0; j < input_vc_number; j++) {

			if ( _input.vc_state(i, j) == SW_TR) {
				// from input port "i", vc index "j" -> output port "out_t.first",
				// vc index "out_t.second";
				VC_TYPE out_t = _input.selected_routing(i, j);
				// in this upstream router, at the output port with index
				// "out_t.first", we maintain for vc with index "out_t.second"
				// how many empty slots are available in the downstream router
				// "seen" from this port;
				// when sending a flit now, we know that downstream router will
				// have a slot less of space available;
				_output.counter_next_r_dec(out_t.first, out_t.second);


				double event_time = _vnoc->event_queue()->current_time();
				if ( i != 0) {
					int cre_router_id = _neighbor_router_id_of_this_port[ i];
					long cre_port = _neighbor_port_of_this_port[ i];

					// physical link delay for the credit token to travel back;
					//double delay = CREDIT_DELAY;
					double delay = _vnoc->h_credit_delay();
					_vnoc->event_queue()->add_event( EVENT(EVENT::CREDIT,
						event_time + delay,
						_id, // this router is src of credit;
						cre_router_id, // des of credit;
						cre_port, // pc;
						j)); // vc;
				}


				long in_size_t = _input.input_buff(i,j).size();
				assert(in_size_t >= 1);
				FLIT flit_t( _input.get_flit(i,j));
				_input.remove_flit(i, j);
				// power stuff here;

				// Note: output buffer is not organized by virtual channels like
				// the input buffers; all flits from various vc indices of various 
				// input ports are stored in the output buffer;
				_output.add_flit(out_t.first, flit_t);
				if ( i == 0) {
					if ( _input.injection_buff_full() == true) {
						if ( _input.input_buff(0,j).size() < BUFF_BOUND) {
							_input.clear_injection_buff_full();
							// "bug1" happens here?
							receive_packet_from_trace_file();
						}
					}
				}
				_output.add_addr(out_t.first, out_t);
				// the output buffer has just added a tail flit: that means that
				// this output port is available (if it still has empty slots)
				// to accept more flits of other packets from any other input
				// port and vc index; the already existing flits in the output
				// buffer of this port are already waiting the the pipeline to be
				// shipped thru the physical link;
				if ( flit_t.type() == FLIT::TAIL) {
					_output.release(out_t.first, out_t.second);
				}
				if ( in_size_t > 1) {
					if ( flit_t.type() == FLIT::TAIL) {
						if ( _vnoc->topology()->vc_sharing_mode() == NOT_SHARED) {
							if (i != 0){
								if (in_size_t != 1) {
									cout<<i<<":"<<in_size_t<<endl;
								}
								assert(in_size_t == 1);
							}
						}
						// if input port "i", vc index "j" had more than one flit,
						// and the one just moved to output port was a TAIL flit;
						_input.vc_state_update(i, j, ROUTING);
					} else {
						// if input port "i", vc index "j" had more than one flit,
						// and the one just moved to output port was not a TAIL,
						// then there are still flits of the same packet to be 
						// arbitrated;
						_input.vc_state_update(i, j, SW_AB);
					}
				} else {
					// input port "i", vc index "j" had only one flit, which
					// now has just been moved to the output buffer of port 
					// "out_t.first"; hence, input buffer "i", vc "j" becomes 
					// empty; so, update its status to INIT;
					_input.vc_state_update(i, j, INIT);
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// sw arbitration
//
////////////////////////////////////////////////////////////////////////////////

void ROUTER::sw_arbitration_stage() 
{
	// switch arbitration pipeline stage; processes all input ports
	// at the same time;

	// (1) build map;
	map<long, vector<VC_TYPE> > vc_o_map;

	for ( long i = 0; i < _physical_ports_count; i++) {
 		// go thru all vc indices and record those that are in SW_AB status;
		// then select a winer vc and record in top-level map;
		vector<long> vc_i_t;
		long input_vc_number = _input.vc_count( i);
		for ( long j = 0; j < input_vc_number; j++) {
			if ( _input.vc_state(i, j) == SW_AB) {
				// retrieve the selected routing as decided/calculated during
				// the prev. stage of VC_AB;
				VC_TYPE out_t = _input.selected_routing(i, j);
				if (( _output.counter_next_r(out_t.first, out_t.second) > 0) &&
					( _output.local_counter(out_t.first) > 0)) {
					vc_i_t.push_back(j);
				}
			}
		}
		long vc_size_t = vc_i_t.size();
		if ( vc_size_t > 1) {
			long win_t = _vnoc->topology()->rng().flat_l(0, vc_size_t);
			// r_t tells the output port index "r_t.first" and the
			// vc index "r_t.second" in the input buffer of the downstream
			// router, to which this flit will go;
			VC_TYPE r_t = _input.selected_routing(i, vc_i_t[win_t]);
			// record in map - with an entry for each output port id -
			// all the input ports that are competing for switch traversal;
			vc_o_map[r_t.first].push_back(VC_TYPE(i, vc_i_t[win_t]));
		} else if ( vc_size_t == 1) {
			VC_TYPE r_t = _input.selected_routing(i, vc_i_t[0]);
			vc_o_map[r_t.first].push_back(VC_TYPE(i, vc_i_t[0]));
		}
	}

	if ( vc_o_map.size() == 0) {
		return;
	}

	// (2) go thru each output port; if the map has recorded that this output
	// port is "desired" by more input ports, then arbitrate it and give it
	// randomly to an input port that competes for it;
	for ( long i = 0; i < _physical_ports_count; i++) {
		long vc_size_t = vc_o_map[i].size();
		if ( vc_size_t > 0) {
			VC_TYPE vc_win = vc_o_map[i][0];
			if ( vc_size_t > 1) {
				vc_win = vc_o_map[i][ _vnoc->topology()->rng().flat_l(0, vc_size_t) ];
			}
			// the vc that was in status SW_AB and won the arbitration is
			// updated to status SW_TR;
			_input.vc_state_update(vc_win.first, vc_win.second, SW_TR);
			FLIT &flit_t = _input.get_flit(vc_win.first, vc_win.second);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// vc arbitration algorithms: implementation of selection function in routing
//
////////////////////////////////////////////////////////////////////////////////

pair<long, long> ROUTER::vc_selection(long i, long j) 
{
	// the flit here at input port "i", vc index "j" of this upstream
	// router will have to be routed to the downstream routers' input 
	// buffer in some vc index; here we select that vc index randomly 
	// from among the free ones;
	// choose one of the candidate routing vc;

	// (1) first get all free potential vc candidates;
	vector<VC_TYPE > &vc_candidates = _input.routing(i, j);
	long r_size_t = vc_candidates.size();
	assert( r_size_t > 0);
	vector<VC_TYPE > vc_that_are_free;

	for ( long i = 0; i < r_size_t; i++) {
		VC_TYPE v_t = vc_candidates[i];
		if ( _vnoc->topology()->vc_sharing_mode() == SHARED) {
			// if shared mode, then the buffer of a vc can be used by 
			// multiple flits of different packets;
			if ( _output.vc_usage(v_t.first, v_t.second) == ROUTER_OUTPUT::FREE) {
				vc_that_are_free.push_back(vc_candidates[i]);
			}
		} else { // NOT_SHARED
			// if not in shared mode vc selection will go thru only if 
			// the downstream buffer of this vc is really empty; only in that
			// case it is included in the list from which the random pick will
			// be made;
			if ( _output.counter_next_r(v_t.first, v_t.second) == _buffer_size) {
				if (  _output.vc_usage(v_t.first, v_t.second) == ROUTER_OUTPUT::FREE) {
					vc_that_are_free.push_back(vc_candidates[i]);
				}
			}
		}
	}

	// (2) then, pick out one randomly;
	r_size_t = vc_that_are_free.size();
	long vc_t = 0;
	if ( r_size_t > 0) {
		if ( r_size_t > 1) {
			vc_t = _vnoc->topology()->rng().flat_l(0, r_size_t);
		}
		// return pair of output port id, and vc index;
		return (vc_that_are_free[vc_t]);
	} else {
		return VC_TYPE(-1, -1);
	}
}

void ROUTER::vc_arbitration_stage()
{
	// (1) go thru all input ports and thru all their vc indices;
	// if any vc state is VC_AB then process and record in map;
	map<VC_TYPE, vector<VC_TYPE> > vc_o_i_map;
	DATA_ATOMIC_UNIT vc_request = 0;

	for ( long i = 0; i < _physical_ports_count; i++) {
		long input_vc_number = _input.vc_count( i);
		for ( long j = 0; j < input_vc_number; j++) {
			VC_TYPE vc_t;
			if ( _input.vc_state(i, j) == VC_AB) {
				// pick out a vc index from among those that are
				// free of the output port where this flit has to go;
				// vc_t is a pair of output port id, and vc index (where
				// the vc index is the free vc index of input buffers of
				// downstream router);
				vc_t = vc_selection(i,j);
				if ((vc_t.first >= 0) && (vc_t.second >= 0)) {
					// output port id, vc index is "desired" also by
					// i,j; so record this in map that stores al i,j
					// that "want" this output port id, vc index;
					vc_o_i_map[vc_t].push_back(VC_TYPE(i,j));
					vc_request = vc_request | VC_MASK[i * input_vc_number + j];
				}
			}
		}
	}
	if ( vc_o_i_map.size() == 0) {
		return;
	}

	// (2) now, go and pick out randomly a winer among all
	// i,j that "wanted" a certain output port id, vc index;
	for ( long i = 1; i < _physical_ports_count; i++) {
		long output_vc_number = _output.vc_count( i);
		for ( long j = 0; j < output_vc_number; j++) {
			if ( _output.vc_usage(i, j) == ROUTER_OUTPUT::FREE) {
				long cont_temp = vc_o_i_map[VC_TYPE(i,j)].size();
				if ( cont_temp > 0) {
					VC_TYPE vc_win = vc_o_i_map[VC_TYPE(i,j)][0];
					if ( cont_temp > 1) {
						vc_win = vc_o_i_map[VC_TYPE(i,j)][
							_vnoc->topology()->rng().flat_l(0, cont_temp)];
					}
					// Note: the winer i,j gets its status changed to SW_AB;
					_input.vc_state_update(vc_win.first, vc_win.second, SW_AB);
					// () record the selected routing; this will be used during the
					// next stage of SA_AB;
					_input.assign_selected_routing(
						vc_win.first, vc_win.second, VC_TYPE(i,j));
					_output.acquire(i, j, vc_win);
					// power stuff here;
				}
			}
		}
	}	
}

////////////////////////////////////////////////////////////////////////////////
//
// routing_decision_stage
//
////////////////////////////////////////////////////////////////////////////////

void ROUTER::routing_decision_stage()
{
	// this is responsible with consuming the flits if their destination
	// is this router or "clearing and populating" the _routing and 
	// setting status to VC_AB for header flits; for body flits
	// status stays ROUTING if there are still flits in buffer;
	// for tail flits status become INIT (buffer is not used);
	// this is done by going thru all input port ids and all vc indices;

	// only two-dimension is supported for now;
	double event_time = _vnoc->event_queue()->current_time();

	// (1) process first the PE injection physical port 0;
	// Note: flits injected by the PE at this router and which are 
	// put in any of the vc channels of port "0", and whose destination
	// is the PE itself, are consumed directly and the vc status is HOME;
	// is this a real scenario of PE injecting packets to itself?
	long input_vc_number = _input.vc_count( 0);
	for ( long j = 0; j < input_vc_number; j++) {

		// HEADER flit;
		FLIT flit_t;
		if ( _input.vc_state(0,j) == ROUTING) {
			flit_t = _input.get_flit(0, j);
			int src_router_id = flit_t.src_router_id();
			int des_router_id = flit_t.des_router_id();
			if ( _id == des_router_id) {
				// (a) packets injected to itself;
				consume_flit( event_time, flit_t);
				_input.remove_flit(0, j);
				_input.vc_state_update(0, j, HOME);
			} else {
				// (b) packets injected for other destinations than itself;
				_input.clear_routing(0, j);
				_input.clear_selected_routing(0, j);

				// call the actual routing algo; this populates _routing of
				// _input that will be used during vc_arbitration_stage;
				call_current_routing_algorithm( des_router_id, src_router_id, 0, j);

				_input.vc_state_update(0, j, VC_AB);
			}

		// BODY or TAIL type flits;
		// these flits do not undergo the routing computation stage of
		// the router pipeline; they go thru the router in one cycle
		// as opposed to the header flits that should take 2 cycles;
		} else if ( _input.vc_state(0, j) == HOME) {
			if ( _input.input_buff(0, j).size() > 0) {
				flit_t = _input.get_flit(0, j);
				assert( flit_t.type() != FLIT::HEADER);
				consume_flit( event_time, flit_t);
				_input.remove_flit(0, j);
				if ( flit_t.type() == FLIT::TAIL) {
					if ( _input.input_buff(0, j).size() > 0) {
						_input.vc_state_update(0, j, ROUTING);
					} else {
						_input.vc_state_update(0, j, INIT);
					}
				}
			}
		}
	}

	// (2) then, process the other physical ports;
	for ( long i = 1; i < _physical_ports_count; i++) {
		long input_vc_number = _input.vc_count( i);
		for ( long j = 0; j < input_vc_number; j++) {

			// if flit is consumed here, then send right away back 
			// a CREDIT event/message to let upstream router that an 
			// empty slot just became available;
			FLIT flit_t;
			if ( _input.input_buff(i, j).size() > 0) {
				flit_t = _input.get_flit(i,j);
				int des_router_id = flit_t.des_router_id();
				if ( _id == des_router_id) {
					int cre_router_id = _neighbor_router_id_of_this_port[ i];
					long cre_port = _neighbor_port_of_this_port[ i];

					// physical link delay;
					//double delay = CREDIT_DELAY;
					double delay = _vnoc->h_credit_delay();
					_vnoc->event_queue()->add_event( EVENT(EVENT::CREDIT,
						event_time + delay,
						_id, // this router is src of credit;
						cre_router_id, // des of credit;
						cre_port, // pc;
						j)); // vc;
				}
			}

			// HEADER flit;
			if ( _input.vc_state(i, j) == ROUTING) {
				flit_t = _input.get_flit(i, j);
				assert(flit_t.type() == FLIT::HEADER);

				// ROUTING state only when the header passes through;
				int des_router_id = flit_t.des_router_id();
				int src_router_id = flit_t.src_router_id();
				if ( _id == des_router_id) {
					// (a) flit arrived at input port "i", vc index "j"
					// has as address this local router;
					consume_flit( event_time, flit_t);
					_input.remove_flit(i, j);
					_input.vc_state_update(i, j, HOME);
				} else {
					// (b) flits are just passing by toward other destination;
					_input.clear_routing(i, j);
					_input.clear_selected_routing(i, j);

					// call the actual routing algo; this populates _routing of
					// _input that will be used during vc_arbitration_stage;
					call_current_routing_algorithm( des_router_id, src_router_id, i, j);

					_input.vc_state_update(i, j, VC_AB);
				}

			// BODY or TAIL flits;
			} else if ( _input.vc_state(i, j) == HOME) {
				if ( _input.input_buff(i, j).size() > 0) {
					flit_t = _input.get_flit(i, j);
					assert( flit_t.type() != FLIT::HEADER);
					consume_flit( event_time, flit_t);
					_input.remove_flit(i, j);
					if ( flit_t.type() == FLIT::TAIL) {
						if ( _input.input_buff(i, j).size() > 0) {
							_input.vc_state_update(i, j, ROUTING);
						} else {
							_input.vc_state_update(i, j, INIT);
						}
					}
				}
			}
		}
	}
}

void ROUTER::call_current_routing_algorithm(
	const int des_r_id, const int src_r_id,
	long s_ph, long s_vc)
{
	// Note: src_r_id represents the origin of the routed flit and 
	// des_r_id represents its final destination, where will be consumed;

	// (1) LUT;
	if ( _routing_algo == ROUTING_LUT) {
		// based on the routing info stored in the local LUT we know
		// the direction (i.e., the output port) where this request 
		// should be routed;
		long port_to_route_to = _routing_LUT[src_r_id][des_r_id];

		_input.add_routing(s_ph, s_vc, VC_TYPE(port_to_route_to, 0));		

		// we can actually explore all virtual channels of the output port;
		// to do that later during arbitration we include them to the
		// "possible" routing here;
		long output_vc_number = _output.vc_count( port_to_route_to);
		for ( long k = 1; k < output_vc_number; k++) {
			_input.add_routing(s_ph, s_vc, VC_TYPE(port_to_route_to, k));
		}
	}
}

void ROUTER::sanity_check() const
{
	// sanity checks;
	for ( long i = 0; i < _physical_ports_count; i ++) {
		long input_vc_number = _input.vc_count( i);
		for ( long j = 0; j < input_vc_number; j ++) {
			if ( _input.input_buff(i,j).size() > 0) {
				cout << "Input is not empty" << endl;
			}
		}
		long output_vc_number = _output.vc_count( i);
		for ( long j = 0; j < output_vc_number; j ++) {		
			if ( _input.vc_state(i,j) != INIT) {
				cout << "Input state is wrong" << endl;
			} cout << _output.counter_next_r(i,j) << ":";
			if ( _output.counter_next_r(i,j) != _buffer_size) {
				cout << "Output vc counter is wrong" << endl;
			}
			if ( _output.vc_usage(i,j) != ROUTER_OUTPUT::FREE) {
				cout << "Output is not free" << endl;
			}
			if ( _output.assigned_to(i,j) != VC_NULL) {
				cout << "Output is not reset" << endl;
			}
		}
		if ( _output.out_buffer(i).size() > 0) {
			cout << "Output temp buffer is not empty" << endl;
		}
		if ( _output.out_addr(i).size() > 0) {
			cout << "Output temp buffer is not empty" << endl;
		}
		if( _output.local_counter(i) != _out_buffer_size) {
			cout << "Output local counter is not reset" << endl;
		}
	}
}

void ROUTER::print_router_stats() 
{
	printf("\nRouter: %d, ports#: %d", _id, _physical_ports_count);
	printf("\n\t Ports' vc_count in/out:");
	for ( int k = 0; k < _physical_ports_count; k++) {
		printf(" %d/%d", _input.vc_count( k), _output.vc_count( k));
	}
	printf("\n\t Neighbor routers:");
	for ( int k = 0; k < _physical_ports_count; k++) {
		printf(" %d", _neighbor_router_id_of_this_port[k]);
	}
	printf("\n\t Neighbor ports:");
	for ( int k = 0; k < _physical_ports_count; k++) {
		printf(" %d", _neighbor_port_of_this_port[k]);
	}
	printf("\n\t Routing LUT:");
	long routers_count = _vnoc->routers_count();
	for ( int i = 0; i < routers_count; i++) {
		printf("\n\t");
		for ( int j = 0; j < routers_count; j++) {
			printf(" %d", _routing_LUT[i][j]);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// VNOC
//
////////////////////////////////////////////////////////////////////////////////

VNOC::VNOC( TOPOLOGY *topology, EVENT_QUEUE *event_queue, 
			ROUTERS_GRAPH *noc_graph, MCF *mcf, bool verbose) :
	_routers(),
	_routers_count(0),
	_packets_count(0),
	_input_file_st(),
	_verbose(verbose)
{
	_topology = topology;
	_event_queue = event_queue;
	_noc_graph = noc_graph;

	_h_wire_delay = 1.0;
	_h_pipe_delay = 1.0;
	_h_credit_delay = 1.0;

	long vc_number = _topology->virtual_channel_number();
	long buffer_size = _topology->input_buffer_size();
	long output_buffer_size = _topology->output_buffer_size();
	long flit_size = _topology->flit_size();



	// (1) heterogeneous noc has a number of routers equal to the nodes 
	// count of noc_graph (same as the number of nodes inside mcf);
	_routers_count = _noc_graph->nodes_count();
	for ( int r_id = 0; r_id < _routers_count; r_id++) {
		ROUTERS_NODE *this_node = _noc_graph->get_node( r_id);

		// + 1 means, one for local IP/core injection/consumption;
		long phy_ports_t = this_node->adj().size() + 1;
		long assigned_core_id = this_node->assigned_core();

		// create custom router;
		// Note: here vc_count for each port is not known yet; we'll
		// allocate mem for input and output buffers later after vc_count
		// will have been computed;
		_routers.push_back(
			ROUTER( phy_ports_t, vc_number, buffer_size,
			output_buffer_size, flit_size,
			this, r_id, assigned_core_id));
	}
	// (2) go thru all routers again and calculate "downstream" and
	// "upstream" (i.e., neighbors) arrays;
	for ( int i = 0; i < _routers_count; i++) {
		ROUTERS_NODE *this_node = _noc_graph->get_node( i);
		for ( int k = 0; k < this_node->adj().size(); k++) {
			int adj_arc_id = this_node->adj()[k];
			ROUTERS_ARC *this_arc = _noc_graph->get_arc( adj_arc_id);
			int adj_i = (this_arc->src_id() == i) ? this_arc->des_id() : this_arc->src_id();

			// all adjacent routers id's (example: adj_i=0 1 5 6 7) stored
			// here at k=0,1,2,.. are associated with the ports k+1=1,2,3,.. 
			// (zero is reserved for local core);
			_routers[i].set_neighbor_router_id_of_this_port( k+1, adj_i);
			// neighbor_port is the port id of the adjacent router that
			// has this router as its neighbor;
			int k2 = _noc_graph->index_in_adj_of_router_id( adj_i, i); // this_r_id, r_id;

			_routers[i].set_neighbor_port_of_this_port( k+1, k2+1);
		}
	}


	// (3) use routing paths from mcf to create _routing_LUT tables 
	// inside each router;
	// Note: here we also bookkeep the number of paths that go thru a link
	// by incrementing the corresponding counter for both the input port
	// of the downstream router as well as for the output port of the upstream
	// router; the value of these counters will be used to set the number of
	// virtual channels in order to implement deadlock free routing;
	int no_commodities = mcf->no_commodities();
	for ( int i = 0; i < no_commodities; i++) {

		int src_r_id = mcf->get_src_of_commodity(i);
		int des_r_id = mcf->get_des_of_commodity(i);
		int r_id = src_r_id;

		// () increment also the port id 0 for each flow that starts at this
		// source; the number of virtual channels for the local injection port
		// will be the same as the number of flows leaving the node; initially
		// I had it fixed to 3 for each router;
		_routers[r_id].incr_flow_count_of_input_port( 0);

		while ( r_id != des_r_id) {
			// () get downstream node id along the preferred path of
			// this commodity;
			int next_r_id = mcf->get_preferred_path( r_id, i);

			// what is the output port id of this router that will go to the
			// neighbor router downstream?
			long port_id = _routers[r_id].get_port_id_of_adj_router( next_r_id);
			_routers[r_id].set_routing_LUT( src_r_id, des_r_id, port_id);

			// () record the going of one flow thru this link: update counters
			// at the input port of downstream router and output port of this
			// upstream router;
			_routers[r_id].incr_flow_count_of_output_port( port_id);
			long in_port_id = _routers[next_r_id].get_port_id_of_adj_router( r_id);
			_routers[next_r_id].incr_flow_count_of_input_port( in_port_id);

			r_id = next_r_id; // prepare for next iter;
		}
	}

	
	// (4) at this time vc_count for each port has been calculated;
	// so, we can allocate mem for all input and output buffers of 
	// each router;
	for ( int r_id = 0; r_id < _routers_count; r_id++) {
		ROUTERS_NODE *this_node = _noc_graph->get_node( r_id);
		long phy_ports_t = this_node->adj().size() + 1;

		_routers[r_id].input().initialize_input_buffers( phy_ports_t);
		_routers[r_id].output().initialize_output_buffers( phy_ports_t);
	}



	// () debug;
	if ( _verbose) {
		print_network_routers();
	}
	


	// (4) the master input file;
	init_input_file();
}

void VNOC::init_input_file() 
{
	// Note: _input_file_st is the file 'testcase.trs" that contains
	// all injections for all cores in one single file; all other
	// files are local to each core (hence router) and contain only parts
	// of this _input_file_st file;
	string filename;
	filename=_topology->trace_file().c_str();
	filename.append(".trs");
	_input_file_st.open( filename.c_str());
	if ( !_input_file_st) {
		printf("\nError: Cannot open source file: %s\n",filename.c_str());
		exit(1);
	}
	double event_time_t;
	_input_file_st >> event_time_t;

	_event_queue->add_event( EVENT(EVENT::PE, event_time_t));
}

bool VNOC::receive_EVENT_PE()
{
	// first, inject the flits;
	int src_router_id = -1, des_router_id = -1;
	int core_id = -1;
	long packet_size = -1;


	// read source;
	_input_file_st >> core_id;
	assert( !_input_file_st.eof());
	src_router_id = _noc_graph->core_id_to_router_id( core_id);
	assert( src_router_id >= 0 && src_router_id < _routers_count);


	// read destination;
	_input_file_st >> core_id;
	assert( !_input_file_st.eof());
	des_router_id = _noc_graph->core_id_to_router_id( core_id);
	assert( des_router_id >= 0 && des_router_id < _routers_count);


	// read packet size;
	_input_file_st >> packet_size;
	assert( !_input_file_st.eof());


	// LANDMARK HERE 1

	router( src_router_id)->receive_packet_from_trace_file();

    if ( !_input_file_st.eof()) {
        double event_time_t;
		_input_file_st >> event_time_t;
		if ( !_input_file_st.eof()) {
			_event_queue->add_event( EVENT(EVENT::PE, event_time_t));
		}
	}
}

bool VNOC::receive_EVENT_ROUTER( EVENT this_event)
{
	// router pipeline stage;
	//double delay = PIPE_DELAY;
	double delay = _h_pipe_delay;
	_event_queue->add_event(
		EVENT(EVENT::ROUTER, this_event.start_time() + delay));

	for ( long i = 0; i < _routers_count; i++) {
		_routers[i].simulate_one_router(); // including power;
	}
}

bool VNOC::receive_EVENT_LINK( EVENT this_event)
{
	int des_router_id = this_event.des_router_id();
	long pc_t = this_event.pc();
	long vc_t = this_event.vc();
	FLIT &flit = this_event.flit();
	router( des_router_id)->receive_flit_from_upstream( pc_t, vc_t, flit);
}

bool VNOC::receive_EVENT_CREDIT( EVENT this_event)
{
	int des_router_id = this_event.des_router_id();
	long pc_t = this_event.pc();
	long vc_t = this_event.vc();
	router( des_router_id)->receive_credit( pc_t, vc_t);
}

void VNOC::check_simulation() 
{
	// check if the network is back to the inital state?
	vector<ROUTER>::const_iterator first = _routers.begin();
	vector<ROUTER>::const_iterator last = _routers.end();
	for (; first != last; first++) {
		first->sanity_check(); // empty check;
	}
	printf("\nSuccess: check for routers cleaness is ok\n");
}

void VNOC::print_network_routers()
{
	// debug;
	printf("\nRouters of hNOC topology:");
	for ( int i = 0; i < _routers_count; i++) { 
		_routers[i].print_router_stats();
	}
	printf("\n");
}

void VNOC::print_simulation_results() 
{
	vector<ROUTER>::const_iterator first = _routers.begin();
	vector<ROUTER>::const_iterator last = _routers.end();
	double total_delay = 0;
	// calculate the total delay
	first = _routers.begin();
	for (; first != last; first++) {
		total_delay += first->total_delay();
	}
	long tot_f_t = _event_queue->finished_count();

	first = _routers.begin();
	last = _routers.end();
	for (; first != last; first++) {
		// accumulate power stuff here;
	}
	double curr_time = _event_queue->current_time();
	// compute averages here;

	// Note: because we deal with heterogeneous NOCs we can have different
	// maximum link lengths for different floorplans of the same application;
	// this basically means that T clock has to be different; because I
	// generated the trace files between times 1...10000.0 ns (under the
	// assumption that T clock = 1 ns, this was the delay used by popnet
	// for a technology 100 nm for a link of 1mm) I still keep the WIRE_DELAY
	// PIPE_DELAY, CREDIT_DELAY as the original ones - in order for the injection 
	// times from the trace files to make sense - but here I scale the final result
	// appropriately using the Elmore delay extrapolation;
	double delay_scale_factor = 1; //_h_wire_delay / WIRE_DELAY;
	
	// store here result; will be retrieved by any application
	// of the vNOC simulator; 
	_latency = total_delay / max(tot_f_t, long(1));
	_latency *= delay_scale_factor; // scale it as discussed in Note above;
	// in fact store it directly in its sketch counterpart of
	// the hetero host object;
	_hetero->set_sketch_latency( _latency);
	// record also the avg. number of packets per cycle; used for plots;
	_hetero->set_sketch_packets_per_cycle( double(_packets_count)/curr_time);

	cout.precision(6);
	cout << "--------------------------------------------------" << endl;
	cout << "total number of flits delivered:		" << tot_f_t << endl;
	cout << "average delay per flit:				" << _latency << endl;
	cout << "--------------------------------------------------" << endl;
}

bool VNOC::run_simulation()
{
	bool result;
	
	result = _event_queue->run_simulation();

	return result;
}

////////////////////////////////////////////////////////////////////////////////
//
// VNOC MISC
//
////////////////////////////////////////////////////////////////////////////////

double VNOC::Elmore_delay( double base_delay, double base_length, double actual_length)
{
	// delay is proportional to the square of the lengths ratio;
	// calculation is done based on empirical (transmission line simulation) data
	// for a given base wire length we know that we have a certain base delay;
	return base_delay * (actual_length/base_length)*(actual_length/base_length);
}

double VNOC::Linear_delay( double base_delay, double base_length, double actual_length)
{
	// delay is proportional to the length; this is true when the interconnects are
	// buffered; we assume that that is the case as in todays technology nodes;
	// otherwise, results - as flit latency - will have a large std. variation;
	return base_delay * (actual_length/base_length);
}

double VNOC::Via_delay( double base_delay, double layers) 
{
	// via delay is based on Elmore delay (base reference is one layer 
	// traversal); currently NOT used;
	//return Elmore_delay(VIA_DELAY, 1, layers);
	return 0.0;
}

void VNOC::compute_heterogeneous_t_clock( double max_link_length,
	bool square_proportionality)
{
	// compute the extrapolated values for: WIRE_DELAY, PIPE_DELAY, and
	// CREDIT_DELAY;
	if ( square_proportionality) { // delay is SQUARE with link length;
		_h_wire_delay = Elmore_delay( WIRE_DELAY, BASE_WIRE, // 0.9, 1000 um
			max_link_length);
		_h_pipe_delay = Elmore_delay( PIPE_DELAY, BASE_WIRE, // 1.0, 1000 um
			max_link_length);
		_h_credit_delay = Elmore_delay( CREDIT_DELAY, BASE_WIRE, // 1.0, 1000 um
			max_link_length);
		//printf("\nT clock: %.2f", _h_wire_delay);
		//printf("\nLink max length: %.2f\n", max_link_length);
	} else { // delay is LINEAR with link length;
		_h_wire_delay = Linear_delay( WIRE_DELAY, BASE_WIRE, // 0.9, 1000 um
			max_link_length);
		_h_pipe_delay = Linear_delay( PIPE_DELAY, BASE_WIRE, // 1.0, 1000 um
			max_link_length);
		_h_credit_delay = Linear_delay( CREDIT_DELAY, BASE_WIRE, // 1.0, 1000 um
			max_link_length);
	}
}

